/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.communication.jmx;

import javax.management.MBeanServer;

import org.apache.commons.pool.BaseKeyedPoolableObjectFactory;
import org.apache.commons.pool.KeyedObjectPool;
import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.jmx.api.api.JMXClient;
import org.ow2.petals.jmx.api.api.PetalsJmxApiFactory;
import org.ow2.petals.jmx.api.api.exception.ConnectionErrorException;
import org.ow2.petals.jmx.api.api.exception.DuplicatedServiceException;
import org.ow2.petals.jmx.api.api.exception.MissingServiceException;
import org.ow2.petals.jmx.api.api.exception.PetalsAdminDoesNotExistException;
import org.ow2.petals.jmx.api.api.exception.PetalsAdminServiceErrorException;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;

/**
 * A factory to create JMX client to put in a {@link KeyedObjectPool}
 * 
 * @author Christophe DENEUX - Linagora
 *
 */
public class JMXClientObjectFactory extends BaseKeyedPoolableObjectFactory<String, JMXClient> {

    private final String localContainerName;

    private final MBeanServer localJMXServer;

    private final ConfigurationService configurationService;

    public JMXClientObjectFactory(final String localContainerName, final MBeanServer localJMXServer,
            final ConfigurationService configurationService) {
        this.localContainerName = localContainerName;
        this.localJMXServer = localJMXServer;
        this.configurationService = configurationService;
    }

    @Override
    public JMXClient makeObject(final String containerName) throws Exception {
        final JMXClient petalsJmxClient;
        try {
            if (this.localContainerName.equals(containerName)) {
                // Create a JMX connection to the local JMX server
                petalsJmxClient = PetalsJmxApiFactory.getInstance().createJMXClient(this.localJMXServer);
            } else {
                // Create a JMX connection to a remote JMX server
                final ContainerConfiguration containerConfiguration = this.configurationService
                        .getContainerConfiguration(containerName);
                petalsJmxClient = PetalsJmxApiFactory.getInstance().createJMXClient(containerConfiguration.getHost(),
                        containerConfiguration.getJmxRMIConnectorPort(), containerConfiguration.getUser(),
                        containerConfiguration.getPassword());
            }

            return petalsJmxClient;
        } catch (final ConnectionErrorException e) {
            throw new PetalsException("Failed to connect the the remote JMX service of the container '" + containerName
                    + "'");
        } catch (final DuplicatedServiceException e) {
            throw new PetalsException(e);
        } catch (final MissingServiceException e) {
            throw new PetalsException(e);
        }
    }

    @Override
    public void destroyObject(final String key, final JMXClient petalsJmxClient) throws Exception {
        petalsJmxClient.disconnect();
    }

    @Override
    public boolean validateObject(final String key, final JMXClient petalsJmxClient) {
        try {
            petalsJmxClient.getPetalsAdminServiceClient().ping();
            return true;
        } catch (final PetalsAdminDoesNotExistException e) {
            return false;
        } catch (PetalsAdminServiceErrorException e) {
            return false;
        } catch (ConnectionErrorException e) {
            return false;
        } catch (PetalsException e) {
            return false;
        }
    }

}
