/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.communication.jmx.rmi;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.RMISocketFactory;
import java.util.HashMap;
import java.util.Properties;
import java.util.logging.Logger;

import javax.management.remote.JMXConnectorServer;
import javax.management.remote.JMXConnectorServerFactory;
import javax.management.remote.JMXServiceURL;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.objectweb.fractal.jmx.agent.AdminAttributes;
import org.ow2.petals.jmx.api.api.JMXClient;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * This is the Fractal JMX adaptor for use with MX4J tools.
 * @author ddesjardins - EBM WebSourcing
 */
@Component
public class RMIConnector {

    /**
     * AdminAttributes
     */
    @Requires(name = "adminAtt")
    protected AdminAttributes adminAttributes;

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration")
    private ConfigurationService configurationService;

    /**
     * Logger
     */
    protected final LoggingUtil log = new LoggingUtil(Logger.getLogger(Constants.FRACTAL_COMPONENT_LOGGER_NAME));

    /**
     * JMX server
     */
    protected JMXConnectorServer cs;

    /**
     * Build the JMX Service URL.
     * 
     * @param host
     *            the JMX/RMI host
     * @param jmxRmiPort
     *            the RMI port
     * @return the URL in String form
     */
    public static final String getJmxServiceURL(String host, int jmxRmiPort) {
        return "service:jmx:rmi:///jndi/rmi://" + host + ":" + jmxRmiPort + "/" + JMXClient.PETALS_JMX_RMI_STUB_NAME;
    }

    /**
     * Start the JMX RMI Server
     * 
     * @throws IOException
     */
    @Lifecycle(step = Step.START)
    public void start() throws IOException {
        this.log.start();

        try {
            // Retrieve the port and host
            ContainerConfiguration containerConfiguration = this.configurationService
                    .getContainerConfiguration();
            final String rmiHost = containerConfiguration.getHost();
            final int rmiPort = containerConfiguration.getJmxRMIConnectorPort();
            final long tcpConnectionTimeout = containerConfiguration.getTCPConnectionTimeout();

            System.setProperty("sun.rmi.transport.tcp.responseTimeout",
                    Long.toString(tcpConnectionTimeout));
            System.setProperty("java.rmi.server.disableHttp", "true");
            // needed so that the address used to register the JMX server in the RMI registry is correct
            System.setProperty("java.rmi.server.hostname", rmiHost);

            // Create a custom RMI Direct Socket Factory to handle the
            // connection timeout properly
            RMISocketFactory rmiSocketFactory = new TimedRMIDirectConnectorFactory(
                    (int) tcpConnectionTimeout, rmiHost);
            RMISocketFactory.setSocketFactory(rmiSocketFactory);

            // create an RMI connector
            Registry registry = LocateRegistry.getRegistry(rmiPort);
            try {
                registry.list();
            } catch (Exception e) {
                LocateRegistry.createRegistry(rmiPort);
            }

            // Note: the host is just here for information purpose, the RMI Registry (created above) always listen to
            // ALL available interfaces on the provided port.
            JMXServiceURL jmxUrl = new JMXServiceURL(getJmxServiceURL(rmiHost, rmiPort));

            HashMap<String, Object> env = this.setupJMXSecurity(containerConfiguration);

            this.cs = JMXConnectorServerFactory.newJMXConnectorServer(jmxUrl, env,
                    this.adminAttributes.getRawMBeanServer());
            this.cs.start();
        } catch (IOException e) {
            this.log.error("Failed to start JMX RMI connector service", e);
            throw e;
        }

        this.log.info("JMX RMI server started at : " + "service:jmx:rmi://"
                + this.cs.getAddress().getURLPath());

        this.log.end();
    }

    /**
     * Stop the RMI Server
     */
    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.start();
        if (this.cs != null) {
            try {
                this.cs.stop();
            } catch (Exception e) {
                this.log.warning("Failed to stop JMX RMI connector service", e);
            }
        }
        this.log.end();
    }

    /**
     * Setup JMX security : <br>
     * create a $PETALS_HOME/work/jmx/password.properties file containing jmx
     * user/pwd values <br>
     * create a $PETALS_HOME/work/jmx/access.properties file containing jmx user
     * and readwrite <br>
     * return a hashmap referencing these files
     * 
     * @param containerConfiguration
     * @return the env hashmap
     */
    private HashMap<String, Object> setupJMXSecurity(ContainerConfiguration containerConfiguration) {
        Properties pwdProperties = new Properties();
        pwdProperties.put(containerConfiguration.getUser(), containerConfiguration.getPassword());

        Properties accessProperties = new Properties();
        accessProperties.put(containerConfiguration.getUser(), "readwrite");

        File workDirectory = new File(containerConfiguration.getWorkDirectoryPath());

        File passwordFile = new File(workDirectory, "password.properties");
        File accessFile = new File(workDirectory, "access.properties");

        try (final FileOutputStream out = new FileOutputStream(passwordFile)) {
            pwdProperties.store(out, "role and password for JMX remote connection");
        } catch (IOException e) {
            this.log.error("Can not create " + passwordFile.getAbsolutePath()
                    + " file to store JMX password");
        }
        try (final FileOutputStream out = new FileOutputStream(accessFile)) {
            accessProperties.store(out,
                    "role and right for JMX remote connection");
        } catch (IOException e) {
            this.log.error("Can not create " + accessFile.getAbsolutePath()
                    + " file to store JMX rights");
        }

        HashMap<String, Object> env = new HashMap<String, Object>();
        env.put("jmx.remote.x.password.file", passwordFile.getAbsolutePath());
        env.put("jmx.remote.x.access.file", accessFile.getAbsolutePath());

        return env;

    }

}
