/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.communication.jndi.agent.util;

import static org.ow2.petals.microkernel.communication.jndi.agent.AbstractJNDIAgentServiceImpl.INITIAL_CONTEXT;

import java.io.Serializable;
import java.net.URI;
import java.util.Hashtable;
import java.util.Map;
import java.util.Map.Entry;

import javax.naming.Binding;
import javax.naming.InvalidNameException;
import javax.naming.LinkRef;
import javax.naming.NameAlreadyBoundException;
import javax.naming.NameClassPair;
import javax.naming.NameNotFoundException;
import javax.naming.NamingException;
import javax.naming.NotContextException;
import javax.naming.ServiceUnavailableException;

import org.ow2.petals.microkernel.communication.jndi.agent.AbstractJNDIAgentServiceImpl;
import org.ow2.petals.microkernel.communication.jndi.client.naming.JNDIConnection;
import org.ow2.petals.microkernel.communication.jndi.client.naming.JNDILocalConnection;
import org.ow2.petals.microkernel.communication.jndi.client.naming.NamingConnection;
import org.ow2.petals.microkernel.communication.jndi.client.naming.NamingContextImpl;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * Utility class to handle data.
 * @author ddesjardins - EBM WebSourcing
 */
public class RegistryUtil {

    private static final String SUBCONTEXT_KEYWORD = "subcontext";

    /**
     * The JNDI Agent implementation
     */
    private AbstractJNDIAgentServiceImpl jndiAgentServiceImpl;

    private LoggingUtil log;

    /**
     * Creates a new instance of {@link RegistryUtil}
     * 
     * @param server
     */
    public RegistryUtil(AbstractJNDIAgentServiceImpl jndiAgentServiceImpl) {
        super();
        this.jndiAgentServiceImpl = jndiAgentServiceImpl;
        this.log = this.jndiAgentServiceImpl.log();
    }

    /**
     * Bind an object with a key in the specified context
     * 
     * @param contextName
     *            name of the context
     * @param key
     *            key to bind with
     * @param value
     *            value to bind to the key
     * @throws NamingException
     *             if a problem occurs
     */
    public void bind(Object contextName, Object key, Object value) throws NamingException {
        this.log.start();

        String contextNameString;
        String keyString;

        if (key instanceof String) {
            keyString = (String) key;
            if (keyString.length() == 0) {
                throw new InvalidNameException("Naming key cannot be an empty String");
            }
        } else {
            throw new InvalidNameException("Naming key must be non null and a String");
        }

        if (contextName instanceof String) {
            contextNameString = (String) contextName;
        } else {
            throw new InvalidNameException("Naming context must be non null and a String");
        }

        if (!(value instanceof Serializable)) {
            throw new InvalidNameException("Naming value must be a Serializable");
        }

        // initialize variables
        String path = contextNameString + keyString;
        String fullContextName;
        if (keyString.contains("/")) {
            fullContextName = path.substring(0, path.lastIndexOf("/") + 1);
        } else {
            fullContextName = contextNameString;
        }

        String lastKey = path.substring(path.lastIndexOf("/") + 1);

        if (this.jndiAgentServiceImpl.containsContext(fullContextName)) {
            Map<String, Serializable> context = this.jndiAgentServiceImpl
                    .getContext(fullContextName);
            synchronized (context) {
                if (context.containsKey(lastKey)) {
                    throw new NameAlreadyBoundException(lastKey + " is already bound in context "
                            + fullContextName);
                }
                context.put(lastKey, (Serializable) value);
            }
        } else {
            throw new NotContextException("Context " + fullContextName + " does not exists");
        }

        this.log.end();
    }

    /**
     * Create and bind a new context in the specified context
     * 
     * @param contextName
     *            name of the context in which we want to create a new context
     * @param key
     *            name of the context that we want to create
     * @return newly created context
     * @throws NamingException
     *             if the context name does not exist or if the key is already
     *             bound
     */
    public String createSubcontext(Object contextName, Object key) throws NamingException {
        this.log.start();

        String contextNameString;
        String keyString;

        if (key instanceof String) {
            keyString = (String) key;
            if (keyString.length() == 0) {
                throw new InvalidNameException("Naming key cannot be an empty String");
            }
        } else {
            throw new InvalidNameException("Naming key must be non null and a String");
        }

        if (contextName instanceof String) {
            contextNameString = (String) contextName;
        } else {
            throw new InvalidNameException("Naming context must be non null and a String");
        }

        // initialize variables
        String path = contextNameString + keyString;
        String fullContextName;
        if (keyString.contains("/")) {
            fullContextName = path.substring(0, path.lastIndexOf("/") + 1);
        } else {
            fullContextName = contextNameString;
        }
        // the path is a Context
        path += "/";

        // Check the parent Context existence
        if (!this.jndiAgentServiceImpl.containsContext(fullContextName)) {
            throw new NotContextException("Context " + fullContextName + " does not exists");
        }

        Map<String, Serializable> parentContext = this.jndiAgentServiceImpl
                .getContext(fullContextName);
        synchronized (parentContext) {
            if (parentContext.containsKey(path)) {
                throw new NameAlreadyBoundException(path + " is already bound in context "
                        + fullContextName);
            }
            this.jndiAgentServiceImpl.getContext(path);
            // put only a reference to the parent
            parentContext.put(path, SUBCONTEXT_KEYWORD);
        }

        this.log.end();
        return path;
    }

    /**
     * Destroy a subcontext in the specified context
     * 
     * @param contextName
     *            name of the context in which we want to create a new context
     * @param key
     *            name of the context that we want to create
     * @return newly created context
     * @throws NamingException
     *             if the context name does not exist or if the key is already
     *             bound
     */
    public void destroySubcontext(Object contextName, Object key) throws NamingException {
        this.log.start();

        String contextNameString;
        String keyString;

        if (key instanceof String) {
            keyString = (String) key;
            if (keyString.length() == 0) {
                throw new InvalidNameException("Naming key cannot be an empty String");
            }
        } else {
            throw new InvalidNameException("Naming key must be non null and a String");
        }

        if (contextName instanceof String) {
            contextNameString = (String) contextName;
        } else {
            throw new InvalidNameException("Naming context must be non null and a String");
        }

        // initialize variables
        String path = contextNameString + keyString;
        String fullContextName;
        if (keyString.contains("/")) {
            fullContextName = path.substring(0, path.lastIndexOf("/") + 1);
        } else {
            fullContextName = contextNameString;
        }
        // the path is a Context
        path += "/";

        // check the parent Context existence
        if (!this.jndiAgentServiceImpl.containsContext(fullContextName)) {
            throw new NotContextException("Context " + fullContextName + " does not exists");
        }

        // check the Context existence
        if (!this.jndiAgentServiceImpl.containsContext(path)) {
            throw new NotContextException("Context " + path + " does not exists");
        }

        Map<String, Serializable> parentContext = this.jndiAgentServiceImpl
                .getContext(fullContextName);
        synchronized (parentContext) {
            parentContext.remove(path);

            // delete recursively subcontexts
            Map<String, Serializable> context = this.jndiAgentServiceImpl.getContext(path);
            for (String namingKey : context.keySet()) {
                if (namingKey.contains("/")) {
                    this.destroySubcontext(namingKey);
                }
            }

            this.jndiAgentServiceImpl.removeContext(path);
        }

        this.log.end();
    }

    /**
     * List the content of a context
     * 
     * @param contextName
     *            name of the context
     * @param contextToList
     *            name of the context to list the content
     * @return array of NameClassPair
     * @throws NamingException
     */
    public NameClassPair[] list(Object contextName, Object key) throws NamingException {
        this.log.start();

        String contextNameString;
        String keyString;

        if (key instanceof String) {
            keyString = (String) key;
            if (keyString.length() == 0) {
                throw new InvalidNameException("Naming key cannot be an empty String");
            }
        } else {
            throw new InvalidNameException("Naming key must be non null and a String");
        }

        if (contextName instanceof String) {
            contextNameString = (String) contextName;
        } else {
            throw new InvalidNameException("Naming context must be non null and a String");
        }

        // initialize variables
        NameClassPair[] names = null;
        String path;
        if (INITIAL_CONTEXT.equals(contextNameString) && INITIAL_CONTEXT.equals(keyString)) {
            path = "/";
        } else {
            path = contextNameString + keyString + "/";
        }

        if (!this.jndiAgentServiceImpl.containsContext(path)) {
            throw new NotContextException("Context " + path + " does not exists");
        }

        Map<String, Serializable> context = this.jndiAgentServiceImpl.getContext(path);
        synchronized (context) {
            names = new NameClassPair[context.keySet().size()];
            int i = 0;
            for (Entry<String, Serializable> entry : context.entrySet()) {

                String keyName = entry.getKey();
                String className = null;
                if (entry.getValue() != null) {
                    if (keyName.contains("/")) {
                        String[] splittedKeys = keyName.split("/");
                        keyName = splittedKeys[splittedKeys.length - 1];
                        className = NamingContextImpl.class.getName();
                    } else {
                        className = entry.getValue().getClass().getName();
                    }
                }

                names[i] = new NameClassPair(keyName, className);
                i++;
            }
        }

        this.log.end();
        return names;
    }

    /**
     * List the bindings of a context
     * 
     * @param contextName
     *            name of the context
     * @param contextToList
     *            name of the context to list the content
     * @param localConnection
     *            {@code true} if the connection is local
     * @return array of NameClassPair
     * @throws NamingException
     */
    public Binding[] listBindings(Object contextName, Object key, boolean localConnection)
            throws NamingException {
        this.log.start();

        String contextNameString;
        String keyString;

        if (key instanceof String) {
            keyString = (String) key;
            if (keyString.length() == 0) {
                throw new InvalidNameException("Naming key cannot be an empty String");
            }
        } else {
            throw new InvalidNameException("Naming key must be non null and a String");
        }

        if (contextName instanceof String) {
            contextNameString = (String) contextName;
        } else {
            throw new InvalidNameException("Naming context must be non null and a String");
        }

        // initialize variables
        Binding[] bindings = null;
        String path;
        if (INITIAL_CONTEXT.equals(contextNameString) && INITIAL_CONTEXT.equals(keyString)) {
            path = "/";
        } else {
            path = contextNameString + keyString + "/";
        }

        // check the Context existence
        if (!this.jndiAgentServiceImpl.containsContext(path)) {
            throw new NotContextException("Context " + path + " does not exists");
        }

        Map<String, Serializable> context = this.jndiAgentServiceImpl.getContext(path);
        synchronized (context) {
            bindings = new Binding[context.keySet().size()];
            int i = 0;
            for (Entry<String, Serializable> entry : context.entrySet()) {

                String bindingKey = entry.getKey();
                String className = null;
                Object bindingValue = entry.getValue();
                if (bindingValue != null) {
                    if (SUBCONTEXT_KEYWORD.equals(bindingValue)) {
                        String[] splittedKeys = bindingKey.split("/");
                        bindingKey = splittedKeys[splittedKeys.length - 1];
                        className = NamingContextImpl.class.getName();
                        NamingConnection namingConnection;
                        if (localConnection) {
                            namingConnection = new JNDILocalConnection(this.jndiAgentServiceImpl);
                        } else {
                            URI providerUrl = this.jndiAgentServiceImpl.getDomainConfiguration()
                                    .getJndiConfiguration().getJndiProviderUrl();
                            namingConnection = new JNDIConnection(providerUrl);
                        }

                        bindingValue = new NamingContextImpl(new Hashtable<String, Object>(),
                                namingConnection, entry.getKey(), localConnection);
                    } else {
                        className = entry.getValue().getClass().getName();
                    }
                }

                // If the object is a LinkRef, resolve the link
                if (bindingValue instanceof LinkRef) {
                    String linkedName = ((LinkRef) bindingValue).getLinkName();
                    try {
                        if (linkedName.startsWith(".")) {
                            // relative to the current context
                            bindingValue = this.lookup(contextName, linkedName.substring(1),
                                    localConnection);
                        } else {
                            // relative to the root context
                            bindingValue = this.lookup("/", linkedName, localConnection);
                        }
                    } catch (NamingException e) {
                        throw e;
                    }
                }

                bindings[i] = new Binding(bindingKey, className, bindingValue);
                i++;
            }
        }

        this.log.end();
        return bindings;
    }

    /**
     * Lookup for an object
     * 
     * @param contextName
     *            name of the context
     * @param key
     *            key to look for
     * @param localConnection
     *            {@code true} if the connection is local
     * @return value of the object bound to the key
     * @throws NamingException
     */
    public Serializable lookup(Object contextName, Object key, boolean localConnection)
            throws NamingException {
        this.log.start();

        String contextNameString;
        String keyString;

        if (key instanceof String) {
            keyString = (String) key;
            if (keyString.length() == 0) {
                throw new InvalidNameException("Naming key cannot be an empty String");
            }
        } else {
            throw new InvalidNameException("Naming key must be non null and a String");
        }

        if (contextName instanceof String) {
            contextNameString = (String) contextName;
        } else {
            throw new InvalidNameException("Naming context must be non null and a String");
        }

        // initialize variables
        Serializable out = null;
        String path = contextNameString + keyString;
        String lastKey = path.substring(path.lastIndexOf("/") + 1);
        String fullContextName;
        if (keyString.contains("/")) {
            fullContextName = path.substring(0, path.lastIndexOf("/") + 1);
        } else {
            fullContextName = contextNameString;
        }

        // check the Context existence
        if (!this.jndiAgentServiceImpl.containsContext(fullContextName)) {
            throw new NotContextException("Context " + fullContextName + " does not exists");
        }

        Map<String, Serializable> context = this.jndiAgentServiceImpl.getContext(fullContextName);
        synchronized (context) {
            if (context.containsKey(path + "/")) {
                // it is a context
                NamingConnection namingConnection = null;
                if (localConnection) {
                    namingConnection = new JNDILocalConnection(this.jndiAgentServiceImpl);
                } else {
                    URI providerUrl = this.jndiAgentServiceImpl.getDomainConfiguration()
                            .getJndiConfiguration().getJndiProviderUrl();
                    namingConnection = new JNDIConnection(providerUrl);
                }
                out = new NamingContextImpl(new Hashtable<String, Object>(), namingConnection, path
                        + "/", localConnection);
            } else if (context.containsKey(lastKey)) {
                out = context.get(lastKey);
            } else {
                throw new NameNotFoundException(lastKey + " is not bound in the context "
                        + fullContextName);
            }
        }

        // If the object is a LinkRef, resolve the link
        if (out instanceof LinkRef) {
            String linkedName = ((LinkRef) out).getLinkName();
            if (linkedName.startsWith(".")) {
                // relative to the current context
                out = this.lookup(contextName, linkedName.substring(1), localConnection);
            } else {
                // relative to the root context
                out = this.lookup("/", linkedName, localConnection);
            }
        }

        this.log.end();
        return out;
    }

    /**
     * Lookup for an object go through the links if the object found is a link
     * 
     * @param contextName
     *            name of the context
     * @param key
     *            key to look for
     * @param localConnection
     *            {@code true} if the connection is local
     * @return value of the object bound to the key
     * @throws NamingException
     */
    public Object lookupLink(Object contextName, Object key, boolean localConnection)
            throws NamingException {
        this.log.call();

        return this.lookup(contextName, key, localConnection);
    }

    /**
     * Rebind an object with a key in the specified context
     * 
     * @param data
     *            data
     * @param contextName
     *            name of the context
     * @param key
     *            key to bind with
     * @param value
     *            value to bind to the key
     * @throws NamingException
     *             if a problem occurs
     */
    public void rebind(Object contextName, Object key, Object value) throws NamingException {
        this.log.start();

        String contextNameString;
        String keyString;

        if (key instanceof String) {
            keyString = (String) key;
            if (keyString.length() == 0) {
                throw new InvalidNameException("Naming key cannot be an empty String");
            }
        } else {
            throw new InvalidNameException("Naming key must be non null and a String");
        }

        if (contextName instanceof String) {
            contextNameString = (String) contextName;
        } else {
            throw new InvalidNameException("Naming context must be non null and a String");
        }

        if (!(value instanceof Serializable)) {
            throw new InvalidNameException("Naming value must be a Serializable");
        }

        // initialize variables
        String path = contextNameString + keyString;
        String lastKey = path.substring(path.lastIndexOf("/") + 1);
        String fullContextName;
        if (keyString.contains("/")) {
            fullContextName = path.substring(0, path.lastIndexOf("/") + 1);
        } else {
            fullContextName = contextNameString;
        }

        // check the Context existence
        if (!this.jndiAgentServiceImpl.containsContext(fullContextName)) {
            throw new NotContextException("Context " + fullContextName + " does not exists");
        }

        Map<String, Serializable> context = this.jndiAgentServiceImpl.getContext(fullContextName);
        synchronized (context) {
            context.put(lastKey, (Serializable) value);
        }

        this.log.end();
    }

    /**
     * Rename an object with a key in the specified context
     * 
     * @param contextName
     *            name of the context
     * @param oldKey
     *            old key
     * @param newKey
     *            new key to bind to
     * @param localConnection
     *            {@code true} if the connection is local
     * @throws NamingException
     *             if a problem occurs
     */
    public void rename(Object contextName, Object oldKey, Object newKey, boolean localConnection)
            throws NamingException {
        this.log.start();

        if (!(oldKey instanceof String)) {
            throw new InvalidNameException("Old key must be non null and a String");
        }
        if (!(newKey instanceof String)) {
            throw new InvalidNameException("New key must be non null and a String");
        }
        if (!(contextName instanceof String)) {
            throw new InvalidNameException("Context name must be non null and a String");
        }
        if (((String) oldKey).length() == 0) {
            throw new InvalidNameException("Old key cannot be an empty String");
        }
        if (((String) newKey).length() == 0) {
            throw new InvalidNameException("New key cannot be an empty String");
        }

        Object value = this.lookup(contextName, oldKey, localConnection);
        this.unbind(contextName, oldKey);
        this.bind(contextName, newKey, value);

        this.log.end();
    }

    /**
     * Unbind an object from a Context
     * 
     * @param contextName
     *            name of the context in which we work
     * @param key
     *            key to unbind
     * @throws NamingException
     */
    public void unbind(Object contextName, Object key) throws NamingException {
        this.log.start();

        String contextNameString;
        String keyString;

        if (key instanceof String) {
            keyString = (String) key;
            if (keyString.length() == 0) {
                throw new InvalidNameException("Naming key cannot be an empty String");
            }
        } else {
            throw new InvalidNameException("Naming key must be non null and a String");
        }

        if (contextName instanceof String) {
            contextNameString = (String) contextName;
        } else {
            throw new InvalidNameException("Naming context must be non null and a String");
        }

        // initialize variables
        String path = contextNameString + keyString;
        String lastKey = path.substring(path.lastIndexOf("/") + 1);
        String fullContextName;
        if (keyString.contains("/")) {
            fullContextName = path.substring(0, path.lastIndexOf("/") + 1);
        } else {
            fullContextName = contextNameString;
        }

        // check the Context existence
        if (!this.jndiAgentServiceImpl.containsContext(fullContextName)) {
            throw new NotContextException("Context " + fullContextName + " does not exists");
        }

        Map<String, Serializable> context = this.jndiAgentServiceImpl.getContext(fullContextName);
        synchronized (context) {
            if (context.remove(lastKey) == null) {
                throw new NameNotFoundException(lastKey + " is not bound in the context "
                        + fullContextName);
            }
        }

        this.log.end();
    }

    private void destroySubcontext(String contextName) throws NotContextException,
            ServiceUnavailableException {

        // check the Context existence
        if (!this.jndiAgentServiceImpl.containsContext(contextName)) {
            throw new NotContextException("Context " + contextName + " does not exists");
        }

        // delete recursively subcontexts
        Map<String, Serializable> context = this.jndiAgentServiceImpl.getContext(contextName);
        for (String namingKey : context.keySet()) {
            if (namingKey.contains("/")) {
                this.destroySubcontext(namingKey);
            }
        }

        this.jndiAgentServiceImpl.removeContext(contextName);
    }
}
