/**
 * Copyright (c) 2016-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.communication.sharedarea;

import java.util.Collection;
import java.util.LinkedList;
import java.util.Properties;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.logging.Logger;

import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.transform.TransformerException;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.jbi.servicedesc.endpoint.Location;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint.EndpointType;
import org.ow2.petals.jbi.servicedesc.endpoint.impl.JBIServiceEndpointImpl;
import org.ow2.petals.microkernel.api.communication.sharedarea.SharedAreaService;
import org.ow2.petals.microkernel.api.communication.sharedarea.exception.SharedAreaException;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.implementation.exception.MutableImplException;
import org.w3c.dom.Document;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;
import com.ebmwebsourcing.easycommons.xml.DocumentBuilders;
import com.ebmwebsourcing.easycommons.xml.XMLHelper;
import com.google.common.base.Optional;
import com.google.common.collect.LinkedListMultimap;
import com.google.common.collect.Multimap;
import com.google.common.collect.Multimaps;

/**
 * @author Victor Noel - Linagora
 */
@Component(provides = @Interface(name = SharedAreaService.FRACTAL_SRV_ITF_NAME, signature = SharedAreaService.class))
public class StandaloneSharedAreaImpl implements SharedAreaService {

    protected final LoggingUtil log = new LoggingUtil(Logger.getLogger(COMPONENT_LOGGER_NAME));

    @Requires(name = SharedAreaService.CONFIGURATION_FRACTAL_ITF_NAME)
    protected ConfigurationService configurationService;

    private ContainerConfiguration containerConfiguration;

    private final ConcurrentMap<String, SavedEndpoint> internals = new ConcurrentHashMap<>();

    private final ConcurrentMap<String, SavedEndpoint> externals = new ConcurrentHashMap<>();

    private final Multimap<QName, PetalsServiceEndpoint> internalsI = Multimaps
            .synchronizedListMultimap(LinkedListMultimap.<QName, PetalsServiceEndpoint> create());

    private final Multimap<QName, PetalsServiceEndpoint> externalsI = Multimaps
            .synchronizedListMultimap(LinkedListMultimap.<QName, PetalsServiceEndpoint> create());

    private final Multimap<QName, PetalsServiceEndpoint> internalsS = Multimaps
            .synchronizedListMultimap(LinkedListMultimap.<QName, PetalsServiceEndpoint> create());

    private final Multimap<QName, PetalsServiceEndpoint> externalsS = Multimaps
            .synchronizedListMultimap(LinkedListMultimap.<QName, PetalsServiceEndpoint> create());

    @Lifecycle(step = Step.START)
    public void start() {
        this.log.call();

        this.containerConfiguration = this.configurationService.getContainerConfiguration();
    }

    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.call();
    }

    @Override
    public Collection<PetalsServiceEndpoint> getEndpoints(final EndpointType type) throws SharedAreaException {
        switch (type) {
            case INTERNAL:
                return getEndpoints(internals.values());
            case EXTERNAL:
                return getEndpoints(externals.values());
            default:
                throw new IllegalArgumentException("impossible");
        }
    }

    private static Collection<PetalsServiceEndpoint> getEndpoints(final Collection<SavedEndpoint> endpoints) {
        final Collection<PetalsServiceEndpoint> res = new LinkedList<>();
        for (final SavedEndpoint e : endpoints) {
            res.add(e.endpoint);
        }
        return res;
    }

    private static Collection<PetalsServiceEndpoint> getEndpointsPSE(
            final Collection<PetalsServiceEndpoint> endpoints) {
        return new LinkedList<>(endpoints);
    }

    private static String getEndpointKey(final PetalsServiceEndpoint serviceEndpoint) {
        return getEndpointKey(serviceEndpoint.getServiceName(), serviceEndpoint.getEndpointName(),
                serviceEndpoint.getType());
    }

    private static String getEndpointKey(final QName serviceName, final String endpointName, final EndpointType type) {
        return getEndpointKey(serviceName.toString(), endpointName, type);
    }

    private static String getEndpointKey(final String serviceName, final String endpointName, final EndpointType type) {
        return (type == EndpointType.INTERNAL ? "I" : "E") + "<" + serviceName + ">" + endpointName;
    }

    @Override
    public Collection<PetalsServiceEndpoint> getEndpointsForInterface(final QName interfaceName,
            final EndpointType type) throws SharedAreaException {
        switch (type) {
            case INTERNAL:
                return getEndpointsPSE(internalsI.get(interfaceName));
            case EXTERNAL:
                return getEndpointsPSE(externalsI.get(interfaceName));
            default:
                throw new IllegalArgumentException("impossible");
        }
    }

    @Override
    public Collection<PetalsServiceEndpoint> getEndpointsForService(final QName serviceName, final EndpointType type)
            throws SharedAreaException {
        switch (type) {
            case INTERNAL:
                return getEndpointsPSE(internalsS.get(serviceName));
            case EXTERNAL:
                return getEndpointsPSE(externalsS.get(serviceName));
            default:
                throw new IllegalArgumentException("impossible");
        }
    }

    @Override
    public PetalsServiceEndpoint registerInternalEndpoint(final String endpointName, final QName serviceName,
            final QName[] interfaces, final String componentName, final Document description)
            throws SharedAreaException {
        final PetalsServiceEndpoint e = new JBIServiceEndpointImpl(EndpointType.INTERNAL, endpointName, serviceName,
                interfaces, new Location(this.containerConfiguration.getName(), componentName));
        if (this.internals.putIfAbsent(getEndpointKey(e), new SavedEndpoint(e, description)) == null) {
            this.internalsS.put(serviceName, e);
            for (final QName i : interfaces) {
                this.internalsI.put(i, e);
            }
            return e;
        } else {
            return null;
        }
    }

    /**
     * TODO In contrary than with the registry overlay, here it is easy to implement external endpoints... see
     * PETALSESBCONT-393
     */
    @Override
    public PetalsServiceEndpoint registerExternalEndpoint(final ServiceEndpoint endpoint, final String componentName)
            throws SharedAreaException {
        // TODO should I store it to return it when queried?! normally I should always return it maybe...
        final PetalsServiceEndpoint e = new JBIServiceEndpointImpl(EndpointType.EXTERNAL, endpoint.getEndpointName(),
                endpoint.getServiceName(), endpoint.getInterfaces(),
                new Location(this.containerConfiguration.getName(), componentName));
        if (this.externals.putIfAbsent(getEndpointKey(e), new SavedEndpoint(e, null)) == null) {
            this.externalsS.put(endpoint.getServiceName(), e);
            for (final QName i : endpoint.getInterfaces()) {
                this.externalsI.put(i, e);
            }
            return e;
        } else {
            return null;
        }
    }

    @Override
    public PetalsServiceEndpoint deregisterEndpoint(final QName serviceName, final String endpointName,
            final String componentName, final EndpointType type) throws SharedAreaException {
        final String key = getEndpointKey(serviceName, endpointName, type);
        final SavedEndpoint e = this.internals.get(key);
        if (e != null) {
            if (e.endpoint.getLocation().getComponentName().equals(componentName)) {
                for (final QName i : e.endpoint.getInterfaces()) {
                    this.internalsI.remove(i, e.endpoint);
                }
                this.internalsS.remove(e.endpoint.getServiceName(), e.endpoint);
                this.internals.remove(key);
                return e.endpoint;

            } else {
                this.log.error("Component " + componentName + " tried to deactivate endpoint " + key
                        + " owned by component " + e.endpoint.getLocation().getComponentName() + " on container "
                        + e.endpoint.getLocation().getContainerName());
                return null;
            }
        } else {
            return null;
        }
    }

    @Override
    public PetalsServiceEndpoint getEndpoint(final QName service, final String name, final EndpointType type)
            throws SharedAreaException {
        final String key = getEndpointKey(service, name, type);
        final SavedEndpoint e = this.internals.get(key);
        return e == null ? null : e.endpoint;
    }

    @Override
    public Properties getConfiguration(final String providedTopologyPassphrase) {
        return new Properties();
    }

    @Override
    public Document getTopologyPartConfiguration(Properties properties) throws MutableImplException {
        return DocumentBuilders.newDocument();
    }

    @Override
    public String getImplementationName() {
        return "Petals Standalone Shared Memory";
    }

    @Override
    public String getImplementationVersion() {
        return this.getClass().getPackage().getImplementationVersion();
    }

    @Override
    public Optional<Document> getDescription(final ServiceEndpoint endpoint) throws SharedAreaException {
        if (!(endpoint instanceof PetalsServiceEndpoint)
                || ((PetalsServiceEndpoint) endpoint).getType() != EndpointType.INTERNAL) {
            return null;
        } else {
            final String key = getEndpointKey((PetalsServiceEndpoint) endpoint);
            final SavedEndpoint e = this.internals.get(key);
            return e == null ? null : Optional.fromNullable(e.description);
        }
    }

    @Override
    public String getDescription(final QName serviceName, final String endpointName) throws SharedAreaException {
        final String key = getEndpointKey(serviceName, endpointName, EndpointType.INTERNAL);
        final SavedEndpoint e = this.internals.get(key);
        try {
            return e == null ? null : XMLHelper.createStringFromDOMNode(e.description, true);
        } catch (final TransformerException e1) {
            throw new SharedAreaException(e1);
        }
    }

    @Override
    public Collection<SavedEndpoint> leaveTopology() throws SharedAreaException {
        this.internalsI.clear();
        this.internalsS.clear();
        this.externalsI.clear();
        this.externalsS.clear();

        final Collection<SavedEndpoint> res = new LinkedList<>(this.internals.values());
        this.internals.clear();

        res.addAll(this.externals.values());
        this.externals.clear();

        return res;
    }

    @Override
    public void restoreEndpoints(final Collection<SavedEndpoint> endpoints) throws SharedAreaException {
        for (final SavedEndpoint e : endpoints) {
            switch (e.endpoint.getType()) {
                case EXTERNAL: {
                    this.externals.put(getEndpointKey(e.endpoint), e);
                    for (final QName i : e.endpoint.getInterfaces()) {
                        this.externalsI.put(i, e.endpoint);
                    }
                    this.externalsS.put(e.endpoint.getServiceName(), e.endpoint);
                    return;
                }
                case INTERNAL: {
                    this.internals.put(getEndpointKey(e.endpoint), e);
                    for (final QName i : e.endpoint.getInterfaces()) {
                        this.internalsI.put(i, e.endpoint);
                    }
                    this.internalsS.put(e.endpoint.getServiceName(), e.endpoint);
                    return;
                }
                default:
                    throw new IllegalArgumentException("impossible");
            }
        }

    }

    @Override
    public void joinTopology(final boolean addMyself) throws SharedAreaException {
        // nothing to do for this implementation!
    }
}
