/**
 * Copyright (c) 2009-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.communication.topology;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.ServiceLoader;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Logger;

import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.clientserverapi.topology.RemoteContainer;
import org.ow2.petals.clientserverapi.topology.exception.AttachContainerException;
import org.ow2.petals.clientserverapi.topology.exception.DetachContainerException;
import org.ow2.petals.clientserverapi.topology.exception.DomainIsNotDynamicException;
import org.ow2.petals.clientserverapi.topology.exception.InconsistencyDomainException;
import org.ow2.petals.clientserverapi.topology.exception.InconsistencyException;
import org.ow2.petals.clientserverapi.topology.exception.InconsistentDomainNameException;
import org.ow2.petals.clientserverapi.topology.exception.InvalidPassphraseException;
import org.ow2.petals.clientserverapi.topology.exception.NoDomainConfigurationProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.NoRemoteContainerProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.NoSecurityPassPhraseProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.TopologyServiceException;
import org.ow2.petals.jmx.api.api.JMXClient;
import org.ow2.petals.jmx.api.api.PetalsJmxApiFactory;
import org.ow2.petals.jmx.api.api.TopologyServiceClient;
import org.ow2.petals.jmx.api.api.exception.ConnectionErrorException;
import org.ow2.petals.jmx.api.api.exception.DuplicatedServiceException;
import org.ow2.petals.jmx.api.api.exception.MissingServiceException;
import org.ow2.petals.microkernel.api.communication.sharedarea.SharedAreaService;
import org.ow2.petals.microkernel.api.communication.sharedarea.SharedAreaService.SavedEndpoint;
import org.ow2.petals.microkernel.api.communication.topology.TopologyService;
import org.ow2.petals.microkernel.api.communication.topology.exception.TopologyException;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration.ContainerState;
import org.ow2.petals.microkernel.api.configuration.DomainConfiguration;
import org.ow2.petals.microkernel.api.configuration.DomainConfiguration.DomainMode;
import org.ow2.petals.microkernel.api.jbi.messaging.RouterService;
import org.ow2.petals.microkernel.api.server.FractalHelper;
import org.ow2.petals.microkernel.api.util.ConfigurationServiceUtils;
import org.ow2.petals.microkernel.communication.RemoteCheckerClient;
import org.ow2.petals.microkernel.communication.sharedarea.StandaloneSharedAreaImpl;
import org.ow2.petals.microkernel.server.PetalsCompositeController;
import org.ow2.petals.microkernel.server.exception.PetalsCompositeCtrlException;
import org.ow2.petals.topology.TopologyHelper;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.Domain;
import org.ow2.petals.topology.generated.Topology;

import com.ebmwebsourcing.easycommons.lang.StringHelper;
import com.ebmwebsourcing.easycommons.log.LoggingUtil;
import com.ebmwebsourcing.easycommons.uuid.SimpleUUIDGenerator;

/**
 * The default implementation of the topology service.
 *
 * TODO for now, it does not handle well problems during attachment: if an error happens, the best thing to do is to
 * stop the container and restart it...
 * 
 * For it to work we would need: to remember endpoints that were removed in a previous detachment and taking care of
 * well differentiating detached/attached state w.r.t. to that.
 *
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Christophe Deneux
 * @author Victor Noel
 */
@Component(provides = @Interface(name = "service", signature = TopologyService.class))
public class TopologyServiceImpl implements TopologyService {

    /**
     * Pattern defining a description to put in the domain definition about the new domain of a container after
     * detachment.
     */
    private static final String DOMAIN_DESCR_DETACHMENT_PATTERN = "Domain created during container detachment from domain '%s'";

    /**
     * Inner class to ping the remote container.
     * 
     * @author Roland Naudin - EBM WebSourcing
     */
    private class TopologyPinger extends TimerTask {

        @Override
        public void run() {
            final String myName = TopologyServiceImpl.this.configurationService.getContainerConfiguration().getName();
            final Set<ContainerConfiguration> containers = TopologyServiceImpl.this.configurationService
                    .getContainersConfiguration();
            for (ContainerConfiguration cc : containers) {
                if (TopologyServiceImpl.this.stopped) {
                    break;
                }
                final String containerName = cc.getName();
                if (!containerName.equals(myName)) {
                    if (TopologyServiceImpl.this.remoteChecker.ping(containerName)) {
                        cc.setState(ContainerState.REACHABLE);
                        TopologyServiceImpl.this.log.debug("Container '" + containerName + "' is reachable");
                    } else {
                        cc.setState(ContainerState.UNREACHABLE);
                        TopologyServiceImpl.this.log.debug("Container '" + containerName + "' is unreachable");
                    }
                }
            }
        }

    }

    /**
     * Logger wrapper.
     */
    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(COMPONENT_LOGGER_NAME));
    
    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration")
    private ConfigurationService configurationService;

    /**
     * The service 'SharedArea'
     */
    @Requires(name = "sharedArea")
    private SharedAreaService sharedArea;

    /**
     * The service 'Router'
     */
    @Requires(name = "router")
    private RouterService router;

    /**
     * The service 'PetalsCompositeController'
     */
    @Requires(name = "petalsCompositeController")
    private PetalsCompositeController petalsCompositeController;

    /**
     * Flag to know if the fractal component is stopping
     */
    private volatile boolean stopped;

    /**
     * The Ping service<br>
     * TODO: It must be a ping via the transporter instead of JMX
     */
    @Requires(name = "remote")
    private RemoteCheckerClient remoteChecker;

    /**
     * The timer to ping the remote containers
     */
    private Timer topologyTimer;

    private boolean isCurrentDomainDynamic() {
        return this.configurationService.getDomainConfiguration().getMode() == DomainMode.DYNAMIC;
    }

    @Override
    public boolean isContainerReachable(String containerName) {
        return ContainerState.REACHABLE
                .equals(this.configurationService.getContainerConfiguration(containerName).getState());
    }

    @Lifecycle(step = Step.START)
    public void start() throws PetalsException {
        this.log.start();

        this.stopped = false;
        this.topologyTimer = new Timer("Topology Pinger");
        if (this.remoteChecker != null) {
            final ContainerConfiguration containerConfig = this.configurationService.getContainerConfiguration();
            this.topologyTimer.schedule(new TopologyPinger(), containerConfig.getTopologyPingerStartDelay(),
                    containerConfig.getTopologyPingerPeriodDelay());
        } else {
            this.log.warning("No remote checker is bound. The topology pinger is not scheduled !!");
        }

        // Retrieve the shared area implementation from configuration
        // this has been populated from the topology.xml file, so it can be null or empty!
        final String sharedAreaImplProperty = this.configurationService.getDomainConfiguration()
                .getSharedAreaImplementation();

        if (!StringHelper.isNullOrEmpty(sharedAreaImplProperty)) {
            this.log.info("Shared area implementation property found: " + sharedAreaImplProperty);
        }

        final Class<?> sharedAreaImplDef = validateSharedAreaImplementation(sharedAreaImplProperty);

        this.petalsCompositeController.replaceSharedArea(sharedAreaImplDef);

        // let's overwrite it with the actually used shared area implementation!
        this.configurationService.getDomainConfiguration().setSharedAreaImplementation(sharedAreaImplDef.getName());

        this.log.info("Shared area loaded");

        this.log.end();
    }

    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.start();

        this.stopped = true;
        this.topologyTimer.cancel();

        this.log.end();
    }

    @Override
    public Set<Map<String, String>> retrieveTopology(final String securityPassPhrase,
            final boolean isSensibleInformationExpected) throws TopologyServiceException {
        this.log.start();
        
        // Check security
        final boolean isSecurityOk = this.configurationService.isSecurityTopologyPassphraseOk(securityPassPhrase);
        if (isSensibleInformationExpected && !isSecurityOk) {
            throw new InvalidPassphraseException();
        }
        if (this.log.isDebugEnabled()) {
            if (isSecurityOk) {
                this.log.debug("Security validated. Sensible information will be returned.");
            } else {
                this.log.debug("Security not validated. Sensible information will NOT be returned.");
            }
        }

        final Set<Map<String, String>> topologyMap = new HashSet<>();
        topologyMap.add(this.convertDomainConfiguration(this.configurationService.getDomainConfiguration(),
                securityPassPhrase));
        for (final ContainerConfiguration cc : this.configurationService.getContainersConfiguration()) {
            topologyMap.add(this.convertContainerConfiguration(cc, isSecurityOk));
        }
        this.log.end();
        return topologyMap;
    }

    @Override
    public Object retrieveTopologyObject(String securityPassPhrase, boolean isSensibleInformationExpected)
            throws PetalsException {
        this.log.start();

        // Check security
        final boolean isSecurityOk = this.configurationService.isSecurityTopologyPassphraseOk(securityPassPhrase);
        if (isSensibleInformationExpected && !isSecurityOk) {
            throw new InvalidPassphraseException();
        }

        if (this.log.isDebugEnabled()) {
            if (isSecurityOk) {
                this.log.debug("Security validated. Sensible information will be returned.");
            } else {
                this.log.debug("Security not validated. Sensible information will NOT be returned.");
            }
        }

        // let's get a copy
        final Topology topology = (Topology) this.configurationService.getLocalTopology().clone();

        // in case it is not set
        topology.getDomain().setRegistryImplementation(
                this.configurationService.getDomainConfiguration().getSharedAreaImplementation());

        if (!isSecurityOk) {
            for (final Container container : topology.getDomain().getContainer()) {
                container.setUser(null);
                container.setPassword(null);
            }
        }

        return topology;
    }

    /**
     * Convert a domain configuration into a key/value map form.
     * 
     * @param domainConfiguration
     *            The domain configuration to convert.
     * @return The domain configuration as key/value map form.
     * @param securityPassPhrase
     *            Passphrase needed to return sensible information. If <code>null</code>, empty or not matching the one
     *            of the container on which the connection is established, no sensible information will be returned.
     * @throws TopologyException
     *             An error occurs getting configuration properties
     */
    private Map<String, String> convertDomainConfiguration(final DomainConfiguration domainConfiguration,
            final String securityPassPhrase) throws TopologyException {
        this.log.start();

        try {
            final Map<String, String> domainMap = domainConfiguration.toMap();
            domainMap.put(TopologyItem.TOPOLOGY_TYPE, TopologyItem.DOMAIN);

            try {
                final Properties registryProperties = this.petalsCompositeController
                        .getSharedAreaCltConfiguration(securityPassPhrase);
                for (final String propertyName : registryProperties.stringPropertyNames()) {
                    domainMap.put(propertyName, registryProperties.getProperty(propertyName));
                }
            } catch (final PetalsCompositeCtrlException e) {
                throw new TopologyException(e);
            }

            return domainMap;
        } finally {
            this.log.end();
        }
    }

    /**
     * Convert a container configuration into a key/value map form.
     * 
     * @param containerConfiguration
     *            The container configuration to convert.
     * @param isSecurityOk
     *            if <code>true</code>, sensible information are returned. Otherwise, sensible information are skipped.
     * @return The container configuration as key/value map form.
     */
    private Map<String, String> convertContainerConfiguration(final ContainerConfiguration containerConfiguration,
            final boolean isSecurityOk) {
        this.log.start();

        try {
            final Map<String, String> containerMap = containerConfiguration.toMap(isSecurityOk);
            containerMap.put(TopologyItem.TOPOLOGY_TYPE, TopologyItem.CONTAINER);

            return containerMap;
        } finally {
            this.log.end();
        }
    }

    @Override
    public synchronized void detachContainer() throws DomainIsNotDynamicException, DetachContainerException {

        final ContainerConfiguration currentContainerConfig = this.configurationService.getContainerConfiguration();
        final DomainConfiguration currentDomainConfig = this.configurationService.getDomainConfiguration();

        final ClassLoader oldClassloader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());
        try {
            this.checkIfDetachmentIsPossible();

            try {
                // A stand-alone topology is persisted in the working area to be able to restart the
                // container in its own topology. The topology is persisted before to be stored in the
                // shared area, in this way, if an error occurs writing in the registry, a restart is
                // sufficient to restore state and to not join the initial topology.
                // The stand-alone topology is created re-using some configuration information:
                // - the container name to avoid to change the content of the server.properties
                // - the container hostname that has not to be changed
                // - the JMX configuration
                // - the port of the fast remote transporter
                // - the configuration of the web-service service
                // - the shared area client configuration is re-used, and as the domain name has been changed, the
                // shared areas (the shared area leaved, and the new shared area created) are distinct.
                final DomainConfiguration newDomainCfg = new DomainConfiguration();
                newDomainCfg.setName(new SimpleUUIDGenerator().getNewID());
                newDomainCfg
                        .setDescription(String.format(DOMAIN_DESCR_DETACHMENT_PATTERN, currentDomainConfig.getName()));
                newDomainCfg.setMode(DomainMode.STATIC);
                newDomainCfg.setSharedAreaImplementation(StandaloneSharedAreaImpl.class.getName());

                final Topology standaloneTopology = ConfigurationServiceUtils.toTopology(newDomainCfg,
                        Collections.singletonList(currentContainerConfig));

                replaceSharedArea(standaloneTopology, false);

                this.log.info("Current container detached from its inital domain");
            } catch (final PetalsException e) {
                // these are PetalsException (and other) thrown by other functions we rely on
                throw new DetachContainerException(currentDomainConfig.getName(), e);
            }
        } catch (final Exception e) {
            this.log.warning("Can't detach the current container from its current domain", e);
            throw e;
        } finally {
            Thread.currentThread().setContextClassLoader(oldClassloader);
        }
    }

    /**
     * <p>
     * Only one attachment of the current container can be done in the same time. That's why this method is
     * synchronized. We use a distributed lock to synchronize the attachments of several containers to the same target
     * topology.
     * </p>
     */
    @Override
    public synchronized void attachContainerTo(final RemoteContainer targetContainer, final String securityPassPhrase)
            throws NoRemoteContainerProvidedException, NoDomainConfigurationProvidedException,
            DomainIsNotDynamicException, InconsistencyException, NoSecurityPassPhraseProvidedException,
            AttachContainerException {

        final ClassLoader oldClassloader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());
        try {
            if (targetContainer == null) {
                throw new NoRemoteContainerProvidedException();
            }

            if (securityPassPhrase == null || securityPassPhrase.isEmpty()) {
                throw new NoSecurityPassPhraseProvidedException();
            }

            final String currentContainerName = this.configurationService.getContainerConfiguration().getName();

            final String domainName = targetContainer.getDomainName();
            if (domainName == null) {
                throw new NoDomainConfigurationProvidedException();
            }

            this.log.info(
                    String.format("Attaching the current container '%s' to the domain '%s' using the container '%s:%s'",
                            currentContainerName, domainName, targetContainer.getHost(), targetContainer.getPort()));

            // this contains everything needed to connect to another topology including shared area specificities
            final Topology targetTopology = checkIfAttachmentIsPossible(targetContainer, securityPassPhrase);

            try {
                replaceSharedArea(targetTopology, true);

                // We notify the Petals composite controller that the configuration has changed to adjust
                // states of Fractal components (for example, the Embedded Registry must be stopped if the
                // container is not running in a stand-alone topology).
                this.petalsCompositeController.onConfigurationUpdated();

                this.log.info("Current container attached to the domain " + domainName);
            } catch (final PetalsException e) {
                // these are PetalsException (and other) thrown by other functions we rely on
                throw new AttachContainerException(currentContainerName, targetContainer,
                        "An error occurs attaching the container to another domain", e);
            }
        } catch (final Exception e) {
            this.log.warning("Can't attach the current container to another domain", e);
            throw e;
        } finally {
            Thread.currentThread().setContextClassLoader(oldClassloader);
        }
    }

    /**
     * <ul>
     * <li>Stop the endpoint directory (this will block endpoint resolution and thus new exchange being sent)</li>
     * <li>Leave the topology and get our endpoints</li>
     * <li>Stop the messaging sub-system (i.e. wait for router to process exchanges left and then stop the transporters)
     * </li>
     * <li>Stop the shared area composite</li>
     * <li>replace the shared area component (inside the composite)</li>
     * <li>restart shared area composite</li>
     * <li>restart the messaging subsystem (this restart transporters)</li>
     * <li>join the new topology (potentially new exchanges will thus arrive for us)</li>
     * <li>restart the endpoint directory (new exchange will be sent from us)</li>
     * </ul>
     */
    private void replaceSharedArea(final Topology topology, final boolean addMyself) throws PetalsException {

        // we stop the endpoint directory: this means no new exchange will be sent from us
        // because endpoint resolution won't work
        try (final AutoCloseable ed = this.petalsCompositeController.pauseEndpointDirectory()) {

            // we leave the current topology (which is dynamic)
            // this will prevent other containers to send us new exchange
            final Collection<SavedEndpoint> endpoints = this.sharedArea.leaveTopology();

            // then let's wait for exchanges to have been processed if possible
            this.log.debug("Waiting for pending message exchanges ...");
            if (this.router.waitEndOfPendingExchange(
                    this.configurationService.getContainerConfiguration().getRouterPauseTrafficDelay())) {
                this.log.debug("All pending message exchanges are terminated.");
            } else {
                this.log.warning(
                        "Some pending message exchanges were not terminated and could be lost or terminated in error.");
            }

            // we stop the transporters:
            // 1) stopping the local one means that even if we send (non-new) exchanges locally, they will be blocked
            // (TODO what's the point, since we simply don't want new exchanges to be sent locally, because of endpoint
            // resolution, no?)
            // 2) stopping the tcp one means that
            // 2a) other containers will fail sending exchanges to us
            // (TODO btw when we restart, we should not accept their exchanges then!)
            // 2b) and that sending (non-new) exchanges to other containers will be blocked
            // (TODO don't we want for these exchanges to simply fail instead? after restart they will still be sent...)
            try (final AutoCloseable ms = this.petalsCompositeController.pauseMessagingSubsystem()) {

                // Note: it is the composite that is paused, and the component inside that is replaced
                try (final AutoCloseable sa = this.petalsCompositeController.pauseSharedArea()) {
                    final String sharedAreaImplName = topology.getDomain().getRegistryImplementation();

                    this.log.info("Replacing current shared area implementation with: " + sharedAreaImplName);

                    // the validity of this has already been tested before
                    final Class<?> sharedAreaImpl = getClass().getClassLoader().loadClass(sharedAreaImplName);

                    // we replace it first, so that if it fails, we don't reconfigure the domain!
                    this.petalsCompositeController.replaceSharedArea(sharedAreaImpl);

                    this.log.info("Shared area replaced");

                    // Note: when restarting the shared area, this information will be used
                    // to join the target topology (which is dynamic).
                    this.configurationService.reconfigureDomain(topology);
                }

                // Now the new shared area is restarted, but we still need to join the topology and restore endpoints
                // Note: if restarting the shared area fails, we should not try to run the following

                // and now we can explicitly join the topology:
                // new exchange will thus arrive from other containers
                // TODO but the tcp transporter is not up, so it means we may loose some of them?!?!
                // shouldn't we joint it AFTER restarting the topology?!?!
                this.sharedArea.joinTopology(addMyself);

                this.sharedArea.restoreEndpoints(endpoints);
            }
        } catch (final PetalsException e) {
            // Note: this also include the exception from the various close, because they are all, in this case,
            // of the type PetalsCompositeCtrlException
            throw e;
        } catch (final Exception e) {
            // these exceptions are not expected at all!
            throw new PetalsException("Unexpected exceptions...", e);
        }

        // Now the endpoint directory is restarted and our endpoint resolution resumes and we stard sending new
        // exchanges
    }

    /**
     * <p>
     * Check that the container attachment can be realized. During checks, the target domain configuration is retrieved.
     * Checks applied are:
     * <ul>
     * <li>the domain of the current container is 'dynamic'</li>
     * <li>the given container exists in the target topology</li>
     * <li>the target domain mode is 'dynamic'</li>
     * <li>the registry implementation of the target domain must be the same than the one of the current domain</li>
     * </ul>
     * </p>
     * 
     * @param petalsJmxClient
     *            A JMX client to the given target container
     * @param targetContainer
     *            The given container member of the target domain. Not <code>null</code>, and all required fields have
     *            been previously checked
     * @param securityPassPhrase
     *            The security pass-phrase of the given container member of the target domain
     * @return The target domain configuration filled during checks
     * @throws PetalsException
     */
    private Topology checkIfAttachmentIsPossible(final RemoteContainer targetContainer,
            final String securityPassPhrase)
            throws DomainIsNotDynamicException, InconsistencyException, AttachContainerException {

        // check if the domain of the current container is dynamic (or we are in a standalone topology)
        if (!this.isCurrentDomainDynamic() && !this.configurationService.getDomainConfiguration()
                .getSharedAreaImplementation().equals(StandaloneSharedAreaImpl.class.getName())) {
            throw new DomainIsNotDynamicException();
        }

        final ContainerConfiguration containerConfiguration = this.configurationService.getContainerConfiguration();
        final String currentContainerName = containerConfiguration.getName();
        final String targetDomainName = targetContainer.getDomainName();

        final Topology targetTopology;
        try {
            // Create a JMX connection to the container of the domain in which the current container must be
            // attached. The service 'JMXService' cannot be used because the container of the domain to join is not a
            // member of the initial topology of the current container.
            final JMXClient petalsJmxClient = PetalsJmxApiFactory.getInstance().createJMXClient(
                    targetContainer.getHost(),
                    targetContainer.getPort() == 0 ? JMXClient.DEFAULT_PORT : targetContainer.getPort(),
                    targetContainer.getUser(), targetContainer.getPassword());
            try {
                final TopologyServiceClient topologyServiceClient = petalsJmxClient.getTopologyServiceClient();
                targetTopology = (Topology) topologyServiceClient.retrieveTopologyObject(securityPassPhrase, true);
            } catch (final PetalsException e) {
                throw new AttachContainerException(currentContainerName, targetContainer,
                        "Can't retrieve topology informations from container used to move", e);
            } finally {
                try {
                    petalsJmxClient.disconnect();
                } catch (final ConnectionErrorException e) {
                    this.log.warning("Error while disconnecting from target container", e);
                }
            }
        } catch (final ConnectionErrorException | DuplicatedServiceException | MissingServiceException e) {
            throw new AttachContainerException(currentContainerName, targetContainer,
                    "An error occurs connecting to the container used to move using JMX", e);
        }

        if (this.log.isDebugEnabled()) {
            this.log.debug("Target topology: " + targetTopology);
        }

        final Domain domain = targetTopology.getDomain();
        if (!targetDomainName.equals(domain.getName())) {
            throw new InconsistentDomainNameException(domain.getName(), targetDomainName);
        }
        
        if (domain.getMode() != org.ow2.petals.topology.generated.DomainMode.DYNAMIC) {
            throw new DomainIsNotDynamicException(domain.getName());
        }

        if (TopologyHelper.findContainer(currentContainerName, domain) != null) {
            throw new InconsistencyDomainException(
                    currentContainerName + " is already present in the domain " + targetDomainName);
        }

        // Check the registry implementation consistency
        final String registryImplementation = domain.getRegistryImplementation();

        if (StringHelper.isNullOrEmpty(registryImplementation)) {
            // we expect that this information is explicitly present!
            throw new InconsistencyDomainException(
                    "No shared area implementation defined for the domain " + targetDomainName);
        } else if (StandaloneSharedAreaImpl.class.getName().equals(registryImplementation.trim())) {
            throw new InconsistencyDomainException("Can't attach to a standalone domain (" + targetDomainName + ")");
        }

        validateSharedAreaImplementation(registryImplementation);

        // normally, the configuration service must have already checked that the host and the listen interface are
        // compatible so now we just need to make sure that other containers will be able to contact us.
        final String containerHost = containerConfiguration.getHost();
        boolean myHostIsLoopback = false;
        try {
            myHostIsLoopback = InetAddress.getByName(containerHost).isLoopbackAddress();
        } catch (final UnknownHostException e) {
            this.log.warning("The current container's hostname is not resolvable (" + containerHost
                    + "), other containers may not be able to contact it!", e);
        }
        boolean targetHostIsLoopback = false;
        try {
            targetHostIsLoopback = InetAddress.getByName(targetContainer.getHost()).isLoopbackAddress();
        } catch (final UnknownHostException e) {
            // normally this shouldn't happen since we already created the jmx client...
            this.log.warning("The target container's hostname (" + targetContainer.getHost()
                    + ") is not resolvable, we may have problems to attach to the domain through it.", e);
        }

        if (myHostIsLoopback && !targetHostIsLoopback) {
            this.log.warning("The current container is contactable through the loopback address (" + containerHost
                    + ") but the target container used for attachment seems to be on an external network address ("
                    + targetContainer.getHost() + "), other containes may have problems to contact us!");
        }

        return targetTopology;
    }

    /**
     * <p>
     * Check that the container detachment can be realized. Checks applied are:
     * <ul>
     * <li>the domain of the current container is 'dynamic'.</li>
     * </ul>
     * </p>
     * 
     * @throws DomainIsNotDynamicException
     *             The current domain is not <b>dynamic</b>.
     */
    private void checkIfDetachmentIsPossible() throws DomainIsNotDynamicException {
        // check if the domain of the current container is dynamic
        if (!this.isCurrentDomainDynamic()) {
            throw new DomainIsNotDynamicException();
        }
    }

    private Class<? extends SharedAreaService> validateSharedAreaImplementation(final String sharedAreaImplName)
            throws InconsistencyDomainException {
        // first let's be sure we can create a shared area component
        final Class<? extends SharedAreaService> sharedAreaImplDef;
        if (!StringHelper.isNullOrEmpty(sharedAreaImplName)) {
            try {
                final Class<?> clazz = getClass().getClassLoader().loadClass(sharedAreaImplName);
                if (!SharedAreaService.class.isAssignableFrom(clazz)) {
                    throw new InconsistencyDomainException(
                            sharedAreaImplName + " is not an subclass of " + SharedAreaService.class.getName());
                }
                @SuppressWarnings("unchecked")
                final Class<? extends SharedAreaService> clazz2 = (Class<? extends SharedAreaService>) clazz;
                sharedAreaImplDef = clazz2;
            } catch (final SecurityException | ClassNotFoundException e) {
                final String message = "Can't find " + sharedAreaImplName + " in the classpath";
                this.log.error(message);
                throw new InconsistencyDomainException(message, e);
            }
        } else {
            sharedAreaImplDef = loadSharedAreaImplementation();
        }

        try {
            FractalHelper.createPrimitiveComponent(sharedAreaImplDef);
        } catch (final InstantiationException e) {
            throw new InconsistencyDomainException(
                    sharedAreaImplDef.getName() + " is not an instantiable Fractal component", e);
        }

        return sharedAreaImplDef;
    }

    private Class<? extends SharedAreaService> loadSharedAreaImplementation() throws InconsistencyDomainException {
        if (!this.isCurrentDomainDynamic() && this.configurationService.getContainersConfiguration().size() == 1) {
            this.log.info("Using standalone shared area implementation for static one-container domain: "
                    + StandaloneSharedAreaImpl.class.getName());
            return StandaloneSharedAreaImpl.class;
        } else {
            final ServiceLoader<SharedAreaService> serviceLoader = ServiceLoader.load(SharedAreaService.class);
            final Iterator<SharedAreaService> servicesIterator = serviceLoader.iterator();
            if (servicesIterator.hasNext()) {
                final Class<? extends SharedAreaService> sharedAreaImplDef = servicesIterator.next().getClass();
                if (servicesIterator.hasNext()) {
                    throw new InconsistencyDomainException(
                            "Too many SharedAreaService implementations available in the classpath");
                }
                this.log.info("Shared area implementation found in the classpath: " + sharedAreaImplDef.getName());
                return sharedAreaImplDef;
            } else {
                throw new InconsistencyDomainException(
                        "No SharedAreaService implementation available in the classpath");
            }
        }
    }
}
