/**
 * Copyright (c) 2014-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.extension;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.objectweb.fractal.api.Component;
import org.ow2.petals.microkernel.api.extension.PetalsExtensionController;
import org.ow2.petals.microkernel.api.extension.exception.NotSatisfiedDependenciesException;

/**
 * Utility class about extension manager
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ExtensionManagerUtils {

    private ExtensionManagerUtils() {
        // An utility class has a private constructor
    }

    /**
     * Check that all required dependencies of an extension are activated
     * 
     * @param extensionCtrl
     *            The extension controller of the extension to check
     * @param extensionControllers
     *            All extension controller already activated
     * @return <code>true</code> if all dependencies are activated, <code>false</code> if one dependency is not
     *         activated
     */
    public static boolean isAllDependenciesLoaded(final PetalsExtensionController extensionCtrl,
            final Map<PetalsExtensionController, Component> extensionControllers) {

        final String[] dependencies = extensionCtrl.getDependencies();
        if (dependencies == null || dependencies.length == 0) {
            return true;
        } else {
            final List<String> extensionAlreadyLoaded = new ArrayList<String>();
            for (final PetalsExtensionController pec : extensionControllers.keySet()) {
                extensionAlreadyLoaded.add(pec.getExtensionName());
            }
            for (final String dependency : dependencies) {
                if (!extensionAlreadyLoaded.contains(dependency)) {
                    return false;
                }
            }
            return true;
        }
    }

    /**
     * Order the extension controllers according to their dependencies on other extension
     * 
     * @param extensionControllers
     *            The extension controllers to sort
     * @return The sorted extension controllers
     * @throws NotSatisfiedDependenciesException
     *             Some extension dependencies are not satisfied
     */
    public static List<PetalsExtensionController> sortExtensionControllers(
            final List<PetalsExtensionController> extensionControllers) throws NotSatisfiedDependenciesException {

        final List<PetalsExtensionController> orderedExtensionController = new ArrayList<PetalsExtensionController>();
        final List<String> orderedExtensionNames = new ArrayList<String>();

        // First, we add extension controllers that have no dependency
        {
            final Iterator<PetalsExtensionController> itExtensionControllers = extensionControllers.iterator();
            while (itExtensionControllers.hasNext()) {
                final PetalsExtensionController extensionController = itExtensionControllers.next();
                final String[] extensionDependencies = extensionController.getDependencies();
                if (extensionDependencies == null || extensionDependencies.length == 0) {
                    orderedExtensionController.add(extensionController);
                    orderedExtensionNames.add(extensionController.getExtensionName());
                    itExtensionControllers.remove();
                }
            }
        }

        // Second, we add extension controllers that have dependencies
        {
            boolean extensionAdded = false;
            do {
                final Iterator<PetalsExtensionController> itExtensionControllers = extensionControllers.iterator();
                while (itExtensionControllers.hasNext()) {
                    final PetalsExtensionController extensionController = itExtensionControllers.next();
                    final String[] extensionDependencies = extensionController.getDependencies();
                    boolean allDependenciesFound = true;
                    for (final String extensionDependency : extensionDependencies) {
                        if (!orderedExtensionNames.contains(extensionDependency)) {
                            allDependenciesFound = false;
                            break;
                        }
                    }

                    if (allDependenciesFound) {
                        orderedExtensionController.add(extensionController);
                        orderedExtensionNames.add(extensionController.getExtensionName());
                        itExtensionControllers.remove();
                        extensionAdded = true;
                    } else {
                        extensionAdded = false;
                    }
                }
            } while (extensionAdded && extensionControllers.size() > 0);

            if (extensionControllers.size() != 0) {
                // Some dependencies are not satisfied
                throw new NotSatisfiedDependenciesException(extensionControllers);
            }
        }

        return orderedExtensionController;

    }

}
