/**
 * Copyright (c) 2013-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.extension;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.ServiceLoader;
import java.util.logging.Logger;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.extensions.Controller;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.extension.InstallationExtension;
import org.ow2.petals.microkernel.api.extension.PetalsExtensionController;
import org.ow2.petals.microkernel.api.extension.PetalsPostExtensionController;
import org.ow2.petals.microkernel.api.extension.exception.CreationExtensionException;
import org.ow2.petals.microkernel.api.extension.exception.NotSatisfiedDependenciesException;
import org.ow2.petals.microkernel.api.extension.exception.PetalsExtensionException;
import org.ow2.petals.microkernel.api.server.FractalHelper;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * The implementation of post-extension manager
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
@org.objectweb.fractal.fraclet.annotations.Component(provides = @Interface(name = "service", signature = PostExtensionsManager.class))
public class PostExtensionsManagerImpl implements PostExtensionsManager {

    /**
     * The configuration service
     */
    @Requires(name = "configuration")
    private ConfigurationService configurationService;

    /**
     * The installation extension manager
     */
    @Requires(name = "installationExtMgr")
    private InstallationExtensionsManager installationExtMgr;

    @Controller
    private Component comp;

    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(PostExtensionsManager.COMPONENT_LOGGER_NAME));

    /**
     * All extension controller activated
     */
    private final Map<PetalsExtensionController, Component> extensionControllers = new HashMap<>();

    @Override
    public Map<PetalsExtensionController, Component> getExtensionControllers() {
        return this.extensionControllers;
    }

    /**
     * Starts the post-extensions manager.
     */
    @Lifecycle(step = Step.START)
    public void start() throws PetalsExtensionException {
        this.log.start();

        // Retrieve the extension composite
        try {
            final Component extensionComposite = FractalHelper.getParentComponent(this.comp);

            // Load extensions looking up extension controller available on the
            // classloader, and instantiating associated Fractal component
            final ServiceLoader<PetalsPostExtensionController> serviceLoader = ServiceLoader
                    .load(PetalsPostExtensionController.class);
            final Iterator<PetalsPostExtensionController> servicesIterator = serviceLoader.iterator();
            if (servicesIterator.hasNext()) {
                final List<PetalsExtensionController> extensionControllers = new ArrayList<>();
                while (servicesIterator.hasNext()) {
                    final PetalsExtensionController extensionController = servicesIterator.next();
                    extensionControllers.add(extensionController);
                }

                final List<PetalsExtensionController> orderedExtensionControllers = ExtensionManagerUtils
                        .sortExtensionControllers(extensionControllers);

                for (final PetalsExtensionController extensionController : orderedExtensionControllers) {
                    this.log.debug("Loading post-extension: " + extensionController.getExtensionName());
                    if (extensionController.isActivated(this.configurationService)) {
                        // Check that all dependencies are activated
                        if (ExtensionManagerUtils.isAllDependenciesLoaded(extensionController,
                                this.extensionControllers)) {
                            this.log.debug("Post-extension: " + extensionController.getExtensionName()
                                    + " activated.");

                            try {
                                final Component extensionImpl = extensionController
                                        .createFractalComponent(extensionComposite);

                                this.log.debug("Post-extension: "
                                        + extensionController.getExtensionName()
                                        + ", Fractal component created.");

                                FractalHelper.startComponent(extensionImpl, this.log);
                                this.log.debug("Post-extension: "
                                        + extensionController.getExtensionName()
                                        + ", Fractal component started.");

                                this.extensionControllers.put(extensionController, extensionImpl);

                                final InstallationExtension extensionItf = getInstallationExtension(
                                        (PetalsPostExtensionController) extensionController, extensionImpl);
                                if (extensionItf != null) {
                                    this.installationExtMgr.register(extensionItf);
                                }
                                this.log.info("Post-extension loaded: "
                                        + extensionController.getExtensionName());
                            } catch (final NoSuchInterfaceException | IllegalLifeCycleException
                                    | CreationExtensionException e) {
                                this.log.warning(
                                        "Error starting the post-extension '"
                                        + extensionController.getExtensionName() + "'", e);
                            }
                        } else {
                            this.log.warning("Unable to activate the post-extension '"
                                    + extensionController.getExtensionName()
                                    + "' because a dependency is not activated.");
                        }
                    } else {
                        this.log.info("Post-extension found but not activated: "
                                + extensionController.getExtensionName());
                    }
                }
            } else {
                this.log.info("No Petals post-extension found.");
            }
        } catch (final NoSuchInterfaceException e) {
            throw new PetalsExtensionException("Unable to retrieve the post-extension composite", e);
        } catch (final NotSatisfiedDependenciesException e) {
            this.log.warning(e.getMessage());
            for (final PetalsExtensionController pec : e.getExtensionControllers()) {
                this.log.warning(String.format("\tThe post-extension '%s' requires %s",
                        pec.getExtensionName(), Arrays.toString(pec.getDependencies())));
            }
            throw e;
        } catch (final Exception e) {
            throw new PetalsExtensionException("An unexpected error occurs", e);
        } finally {
            this.log.end();
        }
    }

    private InstallationExtension getInstallationExtension(final PetalsPostExtensionController extensionController,
            final Component extensionImpl) throws NoSuchInterfaceException {
        if (extensionController.isInstallationExtension()) {
            final Object extensionItf = extensionImpl
                    .getFcInterface(PetalsPostExtensionController.INST_EXT_SRV_ITF_NAME);
            if (extensionItf instanceof InstallationExtension) {
                return (InstallationExtension) extensionItf;
            } else {
                this.log.warning("Post-extension is marked as an installation extension, but it's "
                        + PetalsPostExtensionController.INST_EXT_SRV_ITF_NAME + " interface does not implement "
                        + InstallationExtension.class.getName());
                return null;
            }
        } else {
            return null;
        }
    }

    /**
     * Stops the post-extensions manager.
     */
    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.call();

        // Unload extensions
        final Iterator<Map.Entry<PetalsExtensionController, Component>> itExtensionEntries = this.extensionControllers
                .entrySet().iterator();
        while (itExtensionEntries.hasNext()) {
            final Map.Entry<PetalsExtensionController, Component> entry = itExtensionEntries.next();
            final PetalsExtensionController extensionController = entry.getKey();
            final Component extensionImpl = entry.getValue();
            try {
                FractalHelper.stopComponent(extensionImpl, this.log);
                final InstallationExtension extensionItf = getInstallationExtension(
                        (PetalsPostExtensionController) extensionController, extensionImpl);
                if (extensionItf != null) {
                    this.installationExtMgr.deregister(extensionItf);
                }
            } catch (final NoSuchInterfaceException | IllegalLifeCycleException e) {
                this.log.warning(
                        "Error stopping or removing the post-extension '"
                                + extensionController.getExtensionName()
                                + "'", e);
            }
            extensionController.removeFractalComponent();

            itExtensionEntries.remove();
        }
    }

}
