/**
 * Copyright (c) 2014-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.extension;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.ServiceLoader;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.extensions.Controller;
import org.objectweb.fractal.fraclet.types.Step;
import org.objectweb.fractal.util.Fractal;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.extension.PetalsExtensionController;
import org.ow2.petals.microkernel.api.extension.PetalsPreExtensionController;
import org.ow2.petals.microkernel.api.extension.exception.CreationExtensionException;
import org.ow2.petals.microkernel.api.extension.exception.NotSatisfiedDependenciesException;
import org.ow2.petals.microkernel.api.extension.exception.PetalsExtensionException;
import org.ow2.petals.microkernel.api.server.FractalHelper;
import org.ow2.petals.microkernel.server.exception.PetalsCompositeCtrlException;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * The implementation of pre-extension manager
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
@org.objectweb.fractal.fraclet.annotations.Component(provides = @Interface(name = PreExtensionsManager.FRACTAL_SRV_ITF_NAME, signature = PreExtensionsManager.class))
public class PreExtensionsManagerImpl implements PreExtensionsManager {

    /**
     * The configuration service
     */
    @Requires(name = "configuration")
    private ConfigurationService configurationService;

    @Controller
    private Component comp;

    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(PreExtensionsManager.COMPONENT_LOGGER_NAME));

    /**
     * All extension controller activated
     */
    private final Map<PetalsExtensionController, Component> extensionControllers = new HashMap<>();

    @Override
    public Map<PetalsExtensionController, Component> getExtensionControllers() {
        return this.extensionControllers;
    }

    /**
     * Starts the pre-extensions manager.
     */
    @Lifecycle(step = Step.START)
    public void start() throws PetalsExtensionException {
        this.log.start();

        // Retrieve the extension composite
        try {
            final Component extensionComposite = FractalHelper.getParentComponent(this.comp);

            // Load extensions looking up extension controller available on the
            // classloader, and instantiating associated Fractal component
            final ServiceLoader<PetalsPreExtensionController> serviceLoader = ServiceLoader
                    .load(PetalsPreExtensionController.class);
            final Iterator<PetalsPreExtensionController> servicesIterator = serviceLoader.iterator();
            if (servicesIterator.hasNext()) {
                final List<PetalsExtensionController> extensionControllers = new ArrayList<>();
                while (servicesIterator.hasNext()) {
                    final PetalsExtensionController extensionController = servicesIterator.next();
                    extensionControllers.add(extensionController);
                }

                final List<PetalsExtensionController> orderedExtensionControllers = ExtensionManagerUtils
                        .sortExtensionControllers(extensionControllers);

                for (final PetalsExtensionController extensionController : orderedExtensionControllers) {
                    this.log.debug("Loading pre-extension: " + extensionController.getExtensionName());
                    if (extensionController.isActivated(this.configurationService)) {
                        final boolean failsPetalsStartup = extensionController.failsPetalsStartup();

                        // Check that all dependencies are activated
                        if (ExtensionManagerUtils.isAllDependenciesLoaded(extensionController,
                                this.extensionControllers)) {
                            this.log.debug("Pre-extension: " + extensionController.getExtensionName()
                                    + " activated.");

                            try {
                                final Component extensionImpl = extensionController
                                        .createFractalComponent(extensionComposite);

                                this.log.debug("Pre-extension: "
                                        + extensionController.getExtensionName()
                                        + ", Fractal component created.");

                                FractalHelper.startComponent(extensionImpl, this.log);

                                this.log.debug("Pre-extension: "
                                        + extensionController.getExtensionName()
                                        + ", Fractal component started.");

                                this.extensionControllers.put(extensionController, extensionImpl);

                                this.log.info("Pre-extension loaded: "
                                        + extensionController.getExtensionName());
                            } catch (final IllegalLifeCycleException | CreationExtensionException e) {
                                final String errorMsg = "Error starting the pre-extension '"
                                        + extensionController.getExtensionName() + "'";
                                if (failsPetalsStartup) {
                                    throw new PetalsExtensionException(errorMsg, e);
                                } else {
                                    this.log.getLogger().log(Level.WARNING, errorMsg, e);
                                }
                            }
                        } else {
                            final String errorMsg = "Unable to activate the pre-extension '"
                                    + extensionController.getExtensionName()
                                    + "' because a dependency is not activated.";
                            if (failsPetalsStartup) {
                                throw new PetalsExtensionException(errorMsg);
                            } else {
                                this.log.getLogger().log(Level.WARNING, errorMsg);
                            }
                        }
                    } else {
                        this.log.info("Pre-extension found but not activated: "
                                + extensionController.getExtensionName());
                    }
                }
            } else {
                this.log.info("No Petals pre-extension found.");
            }
        } catch (final NoSuchInterfaceException e) {
            throw new PetalsExtensionException("Unable to retrieve the pre-extension composite", e);
        } catch (final NotSatisfiedDependenciesException e) {
            this.log.warning(e.getMessage());
            for (final PetalsExtensionController pec : e.getExtensionControllers()) {
                this.log.warning(String.format("\tThe pre-extension '%s' requires %s",
                        pec.getExtensionName(), Arrays.toString(pec.getDependencies())));
            }
            throw e;
        } catch (final PetalsExtensionException e) {
            throw e;
        } catch (final Exception e) {
            throw new PetalsExtensionException("An unexpected error occurs", e);
        } finally {
            this.log.end();
        }
    }

    /**
     * Stops the pre-extensions manager.
     */
    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.call();

        // Unload all extensions
        final Iterator<Map.Entry<PetalsExtensionController, Component>> itExtensionEntries = this.extensionControllers
                .entrySet().iterator();
        while (itExtensionEntries.hasNext()) {
            final Map.Entry<PetalsExtensionController, Component> entry = itExtensionEntries.next();
            final PetalsExtensionController extensionController = entry.getKey();
            final Component extensionImpl = entry.getValue();
            try {
                FractalHelper.stopComponent(extensionImpl, this.log);
            } catch (final IllegalLifeCycleException e) {
                this.log.warning(
                        "Error stopping or removing the pre-extension '" + extensionController.getExtensionName() + "'",
                        e);
            }

            extensionController.removeFractalComponent();

            itExtensionEntries.remove();
        }
    }

    @Override
    public void onConfigurationUpdated() throws PetalsCompositeCtrlException {

        this.log.call();

        // Stop or start extensions according to the current configuration
        final Iterator<Map.Entry<PetalsExtensionController, Component>> itExtensionEntries = this.extensionControllers
                .entrySet().iterator();
        while (itExtensionEntries.hasNext()) {
            final Map.Entry<PetalsExtensionController, Component> entry = itExtensionEntries.next();
            final PetalsExtensionController extensionController = entry.getKey();
            final Component extensionImpl = entry.getValue();
            try {
                final LifeCycleController extensionImplLC = Fractal.getLifeCycleController(extensionImpl);
                if (extensionController.isActivated(this.configurationService)) {
                    if (LifeCycleController.STOPPED.equals(extensionImplLC.getFcState())) {
                        this.log.debug("The pre-extension '" + extensionController.getExtensionName()
                                + "' is activated but the associated Fractal component is stopped. We start it !!");
                        FractalHelper.startComponent(extensionImpl, this.log);
                    } else {
                        // NOP: The Fractal component is already running
                        this.log.debug("The pre-extension '" + extensionController.getExtensionName()
                                + "' is activated and the associated Fractal component is already started !!");
                    }
                } else {
                    if (LifeCycleController.STARTED.equals(extensionImplLC.getFcState())) {
                        this.log.debug("The pre-extension '" + extensionController.getExtensionName()
                                + "' is not activated but the associated Fractal component is started. We stop it !!");
                        FractalHelper.stopComponent(extensionImpl, this.log);
                    } else {
                        // NOP: The Fractal component is already stopped
                        this.log.debug("The pre-extension '" + extensionController.getExtensionName()
                                + "' is not activated and the associated Fractal component is already stopped !!");
                    }
                }
            } catch (final NoSuchInterfaceException e) {
                this.log.warning(
                        "Error starting or stopping the pre-extension '" + extensionController.getExtensionName() + "'",
                        e);
            } catch (final IllegalLifeCycleException e) {
                this.log.warning(
                        "Error starting or stopping the pre-extension '" + extensionController.getExtensionName() + "'",
                        e);
            } catch (final PetalsExtensionException e) {
                this.log.warning(
                        "Error starting or stopping the pre-extension '" + extensionController.getExtensionName() + "'",
                        e);
            }
        }
    }

}
