/**
 * Copyright (c) 2005-2012 Fossil E-Commerce, 2005-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.installation;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.AbstractMap.SimpleImmutableEntry;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.management.ComponentLifeCycleMBean;
import javax.jbi.management.LifeCycleMBean;
import javax.management.ObjectName;

import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.container.ContainerService;
import org.ow2.petals.microkernel.api.container.Installer;
import org.ow2.petals.microkernel.api.container.SharedLibraryBean;
import org.ow2.petals.microkernel.api.container.SharedLibraryLifeCycle;
import org.ow2.petals.microkernel.api.jbi.management.Context;
import org.ow2.petals.microkernel.api.jbi.management.DeploymentService;
import org.ow2.petals.microkernel.api.jbi.management.InstallationService;
import org.ow2.petals.microkernel.api.jbi.management.TaskProcessor;
import org.ow2.petals.microkernel.api.jbi.messaging.RouterService;
import org.ow2.petals.microkernel.api.system.SystemStateService;
import org.ow2.petals.microkernel.api.system.repository.RepositoryService;
import org.ow2.petals.microkernel.container.lifecycle.ExtendedComponentLifeCycle;
import org.ow2.petals.microkernel.extension.InstallationExtensionsManager;
import org.ow2.petals.microkernel.jbi.management.MBeanNameBuilder;
import org.ow2.petals.microkernel.jbi.management.task.installation.InstallationTaskFactory;
import org.ow2.petals.microkernel.util.ParameterCheckHelper;
import org.ow2.petals.systemstate.generated.Component;
import org.ow2.petals.systemstate.generated.SharedLibrary;

import com.ebmwebsourcing.easycommons.lang.ExceptionHelper;
import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * JBI installation service implementation.
 * @author Rafael Marins - Fossil E-Commerce
 * @author Olivier Fabre - EBM WebSourcing
 */
@org.objectweb.fractal.fraclet.annotations.Component(provides = @Interface(name = "service", signature = InstallationService.class))
public class InstallationServiceImpl implements InstallationService {

    /**
     * Container :: Container Service.
     */
    @Requires(name = "container")
    private ContainerService containerService;

    private InstallationTaskFactory installationTaskFactory;

    /**
     * logger wrapper
     */
    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(InstallationService.COMPONENT_LOGGER_NAME));

    /**
     * Platform :: Repository Service.
     */
    @Requires(name = "repository")
    private RepositoryService repositoryService;

    /**
     * JBI messaging :: Router Service.
     */
    @Requires(name = "router")
    private RouterService router;

    /**
     * JBI management :: Deployment Service.
     */
    @Requires(name = "deployment")
    private DeploymentService deploymentService;

    /**
     * Platform :: SystemState Service.
     */
    @Requires(name = "systemstate")
    private SystemStateService systemStateService;

    /**
     * Configuration Service
     */
    @Requires(name = "configuration")
    private ConfigurationService configurationService;

    @Requires(name = "installationExtensionManager")
    private InstallationExtensionsManager installationExtensionManager;

    @Requires(name = "mbean-name-builder")
    private MBeanNameBuilder mbeanNameBuilder;

    @Override
    public boolean forceUnloadInstaller(String componentName) {
        this.log.start("Component Name : " + componentName);

        boolean removed = false;
        try {
            this.forceShutdownComponent(componentName);
        } catch (final JBIException e) {
            this.log.warning("Can't shutdown component " + componentName, e);
            // do nothing
        }
        try {
            this.uninstallComponent(componentName);
        } catch (final JBIException e) {
            this.log.warning("Can't uninstall component " + componentName, e);
            // do nothing
        }

        try {
            removed = this.unloadInstallerInternal(componentName, true);
        } catch (final JBIException e) {
            this.log.warning("Can't uninstall component " + componentName, e);
            // do nothing
        }

        this.log.end();
        return removed;
    }

    @Override
    public String[] getInstalledComponentsForSharedLibrary(final String slName, final String slVersion)
            throws Exception {
        this.log.call();

        final SharedLibraryLifeCycle sharedLibraryLifeCycle = this.containerService.getSharedLibraryByName(slName,
                slVersion);
        if (sharedLibraryLifeCycle == null) {
            throw new JBIException("Shared Library '" + slName + "' is not installed");
        } else {
            List<String> registeredComponents = sharedLibraryLifeCycle.getRegisteredComponents();
            return registeredComponents.toArray(new String[registeredComponents.size()]);
        }
    }

    @Override
    public String[][] getInstalledSharedLibraries() {
        this.log.call();

        final Map<SharedLibraryBean, SharedLibraryLifeCycle> sls = this.containerService.getSharedLibraries();
        final String[][] result = new String[sls.size()][2];
        int i = 0;
        for (final SharedLibraryBean sharedLibrary : sls.keySet()) {
            result[i][0] = sharedLibrary.getName();
            result[i++][1] = sharedLibrary.getVersion();
        }
        return result;
    }

    @Override
    public String[] getInstallers() {
        this.log.call();
        
        final List<String> result = new ArrayList<String>();
        for (Installer installer : this.containerService.getInstallers().values()) {
            result.add(installer.getComponentName());
        }

        return result.toArray(new String[result.size()]);
    }

    public Logger getLogger() {
        return this.log.getLogger();
    }

    @Override
    public String installSharedLibrary(String slZipURL) {
        this.log.start();

        ParameterCheckHelper.isNullOrEmptyParameterWithLog(slZipURL, "slZipURL", this.log);

        Context installationContext = null;
        try {
            URL archiveURL;
            try {
                archiveURL = new URL(slZipURL);
            } catch (final MalformedURLException e) {
                this.log.warning("Non legal URL used to install SL, parsing it as a path: '" + slZipURL + "'");
                archiveURL = new File(slZipURL).toURI().toURL();
            }

            // Create installation execution context
            installationContext = this.createInstallationContext(archiveURL);

            // Create installation processor and fill it with tasks
            final TaskProcessor processor = this.createSLInstallationProcessor(installationContext,
                    archiveURL);

            // Launch the process
            final Throwable processResult = processor.process();
            if (processResult != null) {
                throw new Exception("The installation of the Shared Library '" + slZipURL
                        + "' failed.", processResult);
            }
        } catch (Exception e) {
            this.log
                    .error("Unable to install the shared library at location '" + slZipURL + "'", e);
            // Redefine the exception to prevent exotic exception cause type,
            // and then unserialization problem for external class-loaders (JMX
            // client)
            throw new RuntimeException(ExceptionHelper.getStackTrace(e));
        }
        this.log.info("Shared Library '" + installationContext.getEntityName() + "' with version '"
                + installationContext.getEntityVersion() + "' installed");

        this.log.end();
        return installationContext.getEntityName();
    }

    @Override
    public ObjectName loadInstaller(String componentName) {
        this.log.start();

        ParameterCheckHelper.isNullOrEmptyParameterWithLog(componentName, "componentName", this.log);

        try {
            return doLoadInstaller(componentName).getKey();
        } catch (final JBIException e) {
            this.log.error("Failed to initialize Component Bootstrap for '" + componentName + "'", e);
            // Redefine the exception to prevent exotic exception cause type,
            // and then unserialization problem for external class-loaders (JMX client)
            throw new RuntimeException(ExceptionHelper.getStackTrace(e));
        } finally {
            this.log.end();
        }
    }

    @Override
    public Installer loadInstallerInternal(String componentName) throws JBIException {
        return doLoadInstaller(componentName).getValue();
    }

    private Map.Entry<ObjectName, Installer> doLoadInstaller(final String componentName) throws JBIException {
        // Call init on bootstrap
        final Installer installer = this.containerService.getInstallerByName(componentName);
        if (installer != null) {
            installer.initBootstrap();

            this.log.info("Installer of the component '" + componentName + "' loaded");

            return new SimpleImmutableEntry<>(this.mbeanNameBuilder.createInstallerMBeanName(componentName), installer);
        } else {
            return new SimpleImmutableEntry<>(null, null);
        }
    }

    /**
     * @throws RuntimeException
     *             if an error occurs during installation process. We have
     *             chosen this type of exception because, in the specification,
     *             this method doesn't return an exception
     */
    @Override
    public ObjectName loadNewInstaller(final String installZipURL) {
        this.log.start();

        ParameterCheckHelper.isNullOrEmptyParameterWithLog(installZipURL, "installZipURL", this.log);

        try {
            return doLoadNewInstaller(installZipURL).getKey();
        } catch (Exception e) {
            this.log.error("Unable to load a new installer at location '" + installZipURL + "'", e);
            // Redefine the exception to prevent exotic exception cause type,
            // and then unserialization problem for external class-loaders (JMX
            // client)
            throw new RuntimeException(ExceptionHelper.getStackTrace(e));
        } finally {
            this.log.end();
        }
    }

    @Override
    public Installer loadNewInstallerInternal(final String installZipURL) throws JBIException, MalformedURLException {
        return doLoadNewInstaller(installZipURL).getValue();
    }

    private Map.Entry<ObjectName, Installer> doLoadNewInstaller(final String installZipURL)
            throws JBIException, MalformedURLException {
        // if it's invalid, then it will fail but it shouldn't be invalid!
        final URL archiveURL = new URL(installZipURL);

        // Create installation execution context
        final Context installationContext = this.createInstallationContext(archiveURL);

        // Create installation processor and fill it with tasks
        final TaskProcessor processor = this.createComponentInstallationProcessor(installationContext, archiveURL);

        // Launch the installation processor
        final Throwable processResult = processor.process();
        if (processResult != null) {
            throw new JBIException(processResult);
        }

        this.log.info("A new installer for the component '" + installationContext.getEntityName() + "' is loaded");

        return new SimpleImmutableEntry<>(installationContext.getInstallerObjectName(),
                installationContext.getInstaller());
    }

    @Override
    public void shutdown() throws Exception {
        this.log.call();

        this.unloadAllInstallers(true);
        this.uninstallAllSharedLibraries();
    }

    @Override
    public String[] shutdownAllComponents() {
        this.log.call();

        final List<String> shutdownComponents = new ArrayList<String>();
        for (final Component componentStateHolder : this.systemStateService.getComponentStateHolders()) {
            final String name = componentStateHolder.getName();
            try {
                this.shutdownComponent(name);
                shutdownComponents.add(name);
            } catch (final JBIException e) {
                this.log.warning("Can't shutdown component '" + name + "'", e);
                // continue
            }
        }
        return shutdownComponents.toArray(new String[shutdownComponents.size()]);
    }

    @Override
    public String[] shutdownAllNoMoreUsedComponents() {
        this.log.call();
        
        final List<String> shutdownComponents = new ArrayList<String>();
        for (final Installer installer : this.containerService.getInstallers().values()) {
            final String componentName = installer.getComponentName();
            try {
                final String[] deployedServiceAssemblies = this.deploymentService
                        .getDeployedServiceAssembliesForComponent(componentName);
                if (deployedServiceAssemblies.length == 0) {
                    this.shutdownComponent(componentName);
                    shutdownComponents.add(componentName);
                }
            } catch (final Exception e) {
                this.log.warning("Can't shutdown component '" + componentName + "'", e);
                // continue
            }
        }

        return shutdownComponents.toArray(new String[shutdownComponents.size()]);
    }

    @Override
    public boolean shutdownComponent(String componentName) throws JBIException {
        return this.shutdownComponent(componentName, false);
    }

    @Override
    public String[] startAllComponents() {
        this.log.call();

        final List<String> startedComponents = new ArrayList<String>();
        for (final Component componentStateHolder : this.systemStateService.getComponentStateHolders()) {
            final String name = componentStateHolder.getName();
            try {
                this.startComponent(name);
                startedComponents.add(name);
            } catch (final JBIException e) {
                this.log.warning("Can't start component '" + name + "'", e);
                // continue
            }
        }

        return startedComponents.toArray(new String[startedComponents.size()]);
    }

    @Override
    public boolean startComponent(final String componentName) throws JBIException {
        this.log.start("Component Name : " + componentName);
        ParameterCheckHelper
                .isNullOrEmptyParameterWithLog(componentName, "componentName", this.log);

        final ComponentLifeCycleMBean componentLifeCycle = this.containerService
                .getComponentLifecycleByName(componentName);

        if (componentLifeCycle != null) {
            componentLifeCycle.start();
        } else {
            throw new JBIException("The component '" + componentName + "' is not installed");
        }

        this.log.info("Component '" + componentName + "' started");

        this.log.end();
        return true;
    }

    @Override
    public String[] stopAllComponents() {
        this.log.call();

        final List<String> stoppedComponents = new ArrayList<String>();
        for (final Component componentStateHolder : this.systemStateService.getComponentStateHolders()) {
            final String name = componentStateHolder.getName();
            try {
                this.stopComponent(name);
                stoppedComponents.add(name);
            } catch (final JBIException e) {
                this.log.warning("Can't stop component '" + name + "'", e);
                // continue
            }
        }
        return stoppedComponents.toArray(new String[stoppedComponents.size()]);
    }

    @Override
    public boolean stopComponent(String componentName) throws JBIException {
        return this.stopComponent(componentName, false);
    }

    @Override
    public String[] uninstallAllComponents() {
        this.log.call();

        final List<String> uninstalledComponents = new ArrayList<String>();
        for (final Component componentStateHolder : this.systemStateService.getComponentStateHolders()) {
            final String name = componentStateHolder.getName();
            try {
                this.uninstallComponent(name);
                uninstalledComponents.add(name);
            } catch (final JBIException e) {
                this.log.warning("Can't uninstall component '" + name + "'", e);
                // continue
            }
        }
        return uninstalledComponents.toArray(new String[uninstalledComponents.size()]);
    }

    @Override
    public String[] uninstallAllNoMoreUsedComponents() {
        this.log.call();

        final List<String> uninstalledComponents = new ArrayList<String>();
        for (final Installer installer : this.containerService.getInstallers().values()) {
            final String componentName = installer.getComponentName();
            try {
                final String[] deployedServiceAssemblies = this.deploymentService
                        .getDeployedServiceAssembliesForComponent(componentName);
                if (deployedServiceAssemblies.length == 0) {
                    this.uninstallComponent(componentName);
                    uninstalledComponents.add(componentName);
                }
            } catch (final Exception e) {
                this.log.warning("Can't uninstall component '" + componentName + "'", e);
                // continue
            }
        }

        return uninstalledComponents.toArray(new String[uninstalledComponents.size()]);
    }

    @Override
    public String[] uninstallAllSharedLibraries() {
        this.log.call();

        final List<String> removedSharedLibraries = new ArrayList<String>();
        for (final SharedLibrary sharedLibraryStateHolder : this.systemStateService.getSharedLibraryStateHolders()) {
            final String name = sharedLibraryStateHolder.getName();
            try {
                if (this.uninstallSharedLibraryInternal(name)) {
                    removedSharedLibraries.add(name);
                }
            } catch (final JBIException e) {
                this.log.warning("Can't uninstall SL '" + name + "'", e);
                // continue
            }
        }
        return removedSharedLibraries.toArray(new String[removedSharedLibraries.size()]);
    }

    @Override
    public String[] uninstallAllNoMoreUsedSharedLibraries() {
        this.log.call();

        final List<String> removedSharedLibraries = new ArrayList<String>();

        final Map<SharedLibraryBean, SharedLibraryLifeCycle> sharedLibraries = this.containerService
                .getSharedLibraries();
        for (final Map.Entry<SharedLibraryBean, SharedLibraryLifeCycle> entry : sharedLibraries.entrySet()) {
            final SharedLibraryLifeCycle sharedLibrary = entry.getValue();
            final List<String> registeredComponents = sharedLibrary.getRegisteredComponents();
            if (registeredComponents != null && registeredComponents.isEmpty()) {
                final String slName = entry.getKey().getName();

                try {
                    if (this.uninstallSharedLibraryInternal(slName)) {
                        removedSharedLibraries.add(slName);
                    }
                } catch (final JBIException e) {
                    this.log.warning("Can't uninstall SL '" + slName + "'", e);
                    // continue
                }
            }
        }

        return removedSharedLibraries.toArray(new String[removedSharedLibraries.size()]);
    }

    @Override
    public boolean uninstallComponent(String componentName) throws JBIException {
        this.log.start("Component Name : " + componentName);
        ParameterCheckHelper
                .isNullOrEmptyParameterWithLog(componentName, "componentName", this.log);

        if (!this.systemStateService.isComponentInstalled(componentName)) {
            throw new JBIException("The component '" + componentName + "' is not installed");
        }

        this.doLoadInstaller(componentName);

        final Installer installer = this.containerService.getInstallerByName(componentName);
        if (installer != null) {
            installer.uninstall();
        } else {
            throw new JBIException("The installer of the component '" + componentName + "' is not loaded");
        }

        this.log.info("Component '" + componentName + "' uninstalled");

        this.log.end();
        return true;
    }

    @Override
    public boolean uninstallSharedLibrary(String slName) {
        try {
            return uninstallSharedLibraryInternal(slName);
        } catch (JBIException e) {
            this.log.error("Unable to uninstall the SL(s) '" + slName + "'", e);
            // Redefine the exception to prevent exotic exception cause type,
            // and then unserialization problem for external class-loaders (JMX
            // client)
            throw new RuntimeException(ExceptionHelper.getStackTrace(e));
        }
    }

    private boolean uninstallSharedLibraryInternal(String slName) throws JBIException {
        this.log.start();

        ParameterCheckHelper.isNullOrEmptyParameterWithLog(slName, "slName", this.log);

        final List<SharedLibrary> sharedLibraryStateHolders = this.systemStateService
                .getSharedLibraryStateHolders(slName);

        if (sharedLibraryStateHolders.isEmpty()) {
            // maybe it is the name-version that is passed as slName
            // this is due to the trick to handle SL version that is not
            // supported by JBI APIs
            for (final SharedLibrary sharedLibraryStateHolder : this.systemStateService
                    .getSharedLibraryStateHolders()) {
                if (slName.startsWith(sharedLibraryStateHolder.getName())
                        && slName.length() > sharedLibraryStateHolder.getName().length() + 1) {
                    final String version = slName.substring(sharedLibraryStateHolder.getName()
                            .length() + 1);
                    if (sharedLibraryStateHolder.getVersion().equals(version)) {
                        sharedLibraryStateHolders.add(sharedLibraryStateHolder);
                        break;
                    }
                }
            }
        }
        if (sharedLibraryStateHolders.isEmpty()) {
            throw new JBIException("No shared library with identification '" + slName
                    + "' is installed");
        }

        boolean result = true;
        final List<JBIException> exs = new ArrayList<>();
        for (final SharedLibrary sharedLibraryStateHolder : sharedLibraryStateHolders) {
            try {
                if (!uninstallSharedLibraryInternal(sharedLibraryStateHolder.getName(),
                        sharedLibraryStateHolder.getVersion())) {
                    result = false;
                }
            } catch (final JBIException e) {
                exs.add(e);
            }
        }
        
        if (!exs.isEmpty()) {
            final JBIException ex = new JBIException(
                    "Couldn't uninstall all the SLs denoted by the name '" + slName + "'");
            for (final JBIException e : exs) {
                ex.addSuppressed(e);
            }
            throw ex;
        }

        this.log.end();
        return result;
    }

    @Override
    public boolean uninstallSharedLibrary(String slName, String version) {
        try {
            return uninstallSharedLibraryInternal(slName, version);
        } catch (final JBIException e) {
            this.log.error("Unable to uninstall the SL(s) '" + slName + "'", e);
            // Redefine the exception to prevent exotic exception cause type,
            // and then unserialization problem for external class-loaders (JMX
            // client)
            throw new RuntimeException(ExceptionHelper.getStackTrace(e));
        }
    }

    private boolean uninstallSharedLibraryInternal(String slName, String version) throws JBIException {
        this.log.start();

        ParameterCheckHelper.isNullOrEmptyParameterWithLog(slName, "slName", this.log);
        ParameterCheckHelper.isNullOrEmptyParameterWithLog(slName, "version", this.log);

        boolean result = true;

        final SharedLibrary sharedLibraryStateHolder = this.systemStateService
                .getSharedLibraryStateHolder(slName, version);
        if (sharedLibraryStateHolder == null) {
            throw new JBIException(
                    "No shared library with identification '" + slName
                    + "' and version " + version + " is installed");
        }

        if (this.containerService.getSharedLibraryByName(slName, version)
                .getRegisteredComponents().size() != 0) {
            throw new JBIException(
                    "At least one component is using the shared library '"
                    + sharedLibraryStateHolder.getName() + "'");
        }

        // Create uninstallation execution context
        final Context uninstallationContext = this.createSLUnInstallationContext(
                sharedLibraryStateHolder.getName(), sharedLibraryStateHolder.getVersion());
        // Create uninstallation processor and fill with tasks
        final TaskProcessor processor = this.createSLUnInstallationProcessor(uninstallationContext);
        // Launch the process

        final Throwable processResult = processor.process();
        if (processResult != null) {
            result = false;
            this.log.error("Unable to uninstall the shared library '"
                    + sharedLibraryStateHolder.getName() + "'", processResult);
        } else {
            this.log.info("Shared library '" + slName + "' with version '"
                    + sharedLibraryStateHolder.getVersion() + "' uninstalled");
        }

        this.log.end();
        return result;
    }

    @Override
    public String[] unloadAllInstallers(boolean force) {
        this.log.call();
        final List<String> removedComponents = new ArrayList<String>();
        final List<Component> componentStateHolders = this.systemStateService
                .getComponentStateHolders();
        Collections.reverse(componentStateHolders);
        for (final Component componentStateHolder : componentStateHolders) {
            final String name = componentStateHolder.getName();

            if (force) {
                if (this.forceUnloadInstaller(name)) {
                    removedComponents.add(name);
                }
            } else {
                try {
                    if (this.unloadInstallerInternal(name, true)) {
                        removedComponents.add(name);
                    }
                } catch (JBIException e) {
                    this.log.warning("Can't unload the installer for '" + name + "'", e);
                    // continue
                }
            }
        }

        return removedComponents.toArray(new String[removedComponents.size()]);
    }


    /**
     * TODO unloadInstaller isToBeDeleted parameter isn't used at the moment: this method completely deletes the
     * component from the container
     */
    @Override
    public boolean unloadInstaller(String componentName, boolean isToBeDeleted) {
        try {
            return unloadInstallerInternal(componentName, isToBeDeleted);
        } catch (JBIException e) {
            this.log.error("Unable to unload the installer for the component '" + componentName + "'", e);
            // Redefine the exception to prevent exotic exception cause type,
            // and then unserialization problem for external class-loaders (JMX
            // client)
            throw new RuntimeException(ExceptionHelper.getStackTrace(e));
        }
    }

    private boolean unloadInstallerInternal(String componentName, boolean isToBeDeleted) throws JBIException {
        this.log.start();

        ParameterCheckHelper
                .isNullOrEmptyParameterWithLog(componentName, "componentName", this.log);

        boolean result = true;
        try {
            if (this.systemStateService.getComponentStateHolder(componentName) == null) {
                throw new JBIException("The installer of the component '" + componentName + "' does not existing");
            }

            // Create uninstallation execution context
            final Context uninstallationContext = this
                    .createComponentUninstallationContext(componentName);

            // Create uninstallation processor and fill it with tasks
            final TaskProcessor processor = this
                    .createComponentUnInstallationProcessor(uninstallationContext);

            // Launch the process
            final Throwable processResult = processor.process();
            if (processResult != null) {
                throw new Exception(processResult);
            }
            this.log.info("Installer of the component '" + componentName + "' succesfully unloaded");
        } catch (final Exception e) {
            this.log.error("Unable to unload the installer of the component '" + componentName
                    + "'", e);
            result = false;
        }

        this.log.end();
        return result;
    }

    /**
     * Initialize the InstallationServiceImpl
     */
    @Lifecycle(step = Step.START)
    public void start() {
        this.log.start();

        final ContainerConfiguration containerConfiguration = this.configurationService
                .getContainerConfiguration();

        // Initialize installation task factory
        this.installationTaskFactory = new InstallationTaskFactory(this.systemStateService, this.repositoryService,
                this.log, this.router, containerConfiguration, this.containerService,
                this.installationExtensionManager.getExtensions());

        // Clean the success file in the work directory
        final File workFile = new File(containerConfiguration.getWorkDirectoryPath());
        final File[] filesOfWorkDir = workFile.listFiles();
        if (filesOfWorkDir != null) {
            for (final File file : filesOfWorkDir) {
                if (!file.isDirectory() && file.getName().endsWith(".success")) {
                    file.delete();
                }
            }
        }

        this.log.end();
    }

    /**
     * Stop the InstallationServiceImpl
     */
    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.call();
    }

    /**
     * Create the installation processor and add all required task for the
     * complete installation of a component
     * 
     * @param installationContext
     *            the installation context
     * @param archiveURI
     *            the component archive URI
     * @return the filled installation processor
     */
    private TaskProcessor createComponentInstallationProcessor(final Context installationContext,
            final URL archiveURL) {
        final TaskProcessor processor = new TaskProcessor(installationContext, this.log);
        if (this.isRemoteURL(archiveURL)) {
            processor.addTask(this.installationTaskFactory.getDownloadTask());
        } else {
            processor.addTask(this.installationTaskFactory.getCopyArchiveToWorkDirTask());
        }
        processor.addTask(this.installationTaskFactory.getExtractTask());
        processor.addTask(this.installationTaskFactory.getComponentCheckPackageTask());
        processor.addTask(this.installationTaskFactory.getRegisterComponentInSLTask());
        processor.addTask(this.installationTaskFactory.getCreateComponentFolderTask());
        processor.addTask(this.installationTaskFactory.getCreateComponentStateHolderTask());
        processor.addTask(this.installationTaskFactory.getCreateAndRegisterInstallerMBeanTask());
        processor.addTask(this.installationTaskFactory.getCopyArchiveToInstalledDirTask());
        return processor;
    }

    /**
     * Create the component uninstallation context
     * 
     * @param name
     *            name of the component to uninstall
     * @return the Uninstallation Context HashMap
     */
    private Context createComponentUninstallationContext(String name) {
        final Context uninstallationContext = new Context();
        uninstallationContext.setEntityName(name);
        return uninstallationContext;
    }

    /**
     * Create the uninstallation processor and add all required task for the
     * complete uninstallation of a component
     * 
     * @param uninstallationContext
     *            the uninstallation context
     * @return the filled uninstallation processor
     */
    private TaskProcessor createComponentUnInstallationProcessor(Context uninstallationContext) {
        final TaskProcessor processor = new TaskProcessor(uninstallationContext, this.log);
        processor.addTask(this.installationTaskFactory.getComponentCheckInstalledTask());
        processor.addTask(this.installationTaskFactory.getUnregisterInstallerTask());
        processor.addTask(this.installationTaskFactory.getUnregisterComponentFromSLTask());
        processor.addTask(this.installationTaskFactory.getRemoveComponentStateHolderTask());
        processor.addTask(this.installationTaskFactory.getRemoveComponentFolderTask());
        processor.addTask(this.installationTaskFactory.getRemoveComponentArchiveTask());
        return processor;
    }

    /**
     * Create the installation context
     * 
     * @param zipUrl
     *            String containing the location URL of the Component or Shared
     *            lib ZIP file
     * @return the Installation Context HashMap
     */
    private Context createInstallationContext(URL archiveUrl) {
        final Context installationContext = new Context();
        installationContext.setArchiveUrl(archiveUrl);
        return installationContext;
    }

    /**
     * Create the installation processor and add all required task for the
     * complete installation of a shared lib
     * 
     * @param installationContext
     *            the installation context
     * @param archiveURI
     *            the sl archive URI
     * @return the filled installation processor
     */
    private TaskProcessor createSLInstallationProcessor(final Context installationContext, final URL archiveURL) {
        final TaskProcessor processor = new TaskProcessor(installationContext, this.log);
        if (this.isRemoteURL(archiveURL)) {
            processor.addTask(this.installationTaskFactory.getDownloadTask());
        } else {
            processor.addTask(this.installationTaskFactory.getCopyArchiveToWorkDirTask());
        }
        processor.addTask(this.installationTaskFactory.getExtractTask());
        processor.addTask(this.installationTaskFactory.getSLCheckPackageTask());
        processor.addTask(this.installationTaskFactory.getCreateSLRepositoryTask());
        processor.addTask(this.installationTaskFactory.getCreateSLStateHolderTask());
        processor.addTask(this.installationTaskFactory.getCreateSLClassloaderTask());
        processor.addTask(this.installationTaskFactory.getCopyArchiveToInstalledDirTask());
        return processor;
    }

    private boolean isRemoteURL(final URL archiveURL) {
        final String protocol = archiveURL.getProtocol();
        if (protocol.equalsIgnoreCase("http") || protocol.equalsIgnoreCase("https") || protocol.equalsIgnoreCase("ftp")
                || protocol.equalsIgnoreCase("mvn")) {
            return true;
        } else if (protocol.equalsIgnoreCase("file")) {
            return false;
        } else {
            throw new IllegalArgumentException("URL protocol not supported:" + protocol);
        }
    }

    /**
     * Create the shared lib uninstallation context
     * 
     * @param name
     *            name of the shared lib to uninstall
     * @return the Uninstallation Context HashMap
     */
    private Context createSLUnInstallationContext(String name, String version) {
        final Context uninstallationContext = new Context();
        uninstallationContext.setEntityName(name);
        uninstallationContext.setEntityVersion(version);
        return uninstallationContext;
    }

    /**
     * Create the uninstallation processor and add all required task for the
     * complete uninstallation of a shared lib
     * 
     * @param uninstallationContext
     *            the uninstallation context
     * @return the filled uninstallation processor
     */
    private TaskProcessor createSLUnInstallationProcessor(Context uninstallationContext) {
        final TaskProcessor processor = new TaskProcessor(uninstallationContext, this.log);
        processor.addTask(this.installationTaskFactory.getRemoveSLClassloaderTask());
        processor.addTask(this.installationTaskFactory.getRemoveSLStateHolderTask());
        processor.addTask(this.installationTaskFactory.getRemoveSLFolderTask());
        processor.addTask(this.installationTaskFactory.getRemoveSLArchiveTask());
        return processor;
    }

    private void forceShutdownComponent(String componentName) throws JBIException {
        this.log.call();

        final ComponentLifeCycleMBean componentLifeCycle = this.containerService
                .getComponentLifecycleByName(componentName);
        if (componentLifeCycle != null) {
            final String compState = componentLifeCycle.getCurrentState();
            if (LifeCycleMBean.STARTED.equals(compState)) {
                this.stopComponent(componentName, true);
                this.shutdownComponent(componentName, true);
            } else if (LifeCycleMBean.STOPPED.equals(compState)) {
                this.shutdownComponent(componentName, true);
            }
        }
    }

    /**
     * Shutdown the component
     * 
     * @param componentName
     * @param force
     * @return
     * @throws JBIException
     */
    private boolean shutdownComponent(final String componentName, boolean force)
            throws JBIException {
        this.log.start("Component Name : " + componentName);
        ParameterCheckHelper
                .isNullOrEmptyParameterWithLog(componentName, "componentName", this.log);

        final LifeCycleMBean componentLifeCycle = this.containerService.getComponentLifecycleByName(componentName);

        if (componentLifeCycle != null) {
            // TODO : dirty, find a better way to do that
            ((ExtendedComponentLifeCycle) componentLifeCycle).setForcedState(force);
            try {
                componentLifeCycle.shutDown();
            } finally {
                ((ExtendedComponentLifeCycle) componentLifeCycle).setForcedState(false);
            }
        } else {
            throw new JBIException("The component '" + componentName + "' is not installed");
        }

        this.log.info("Component '" + componentName + "' shut down");

        this.log.end();
        return true;
    }

    private boolean stopComponent(final String componentName, boolean force) throws JBIException {
        this.log.start("Component Name : " + componentName);
        ParameterCheckHelper
                .isNullOrEmptyParameterWithLog(componentName, "componentName", this.log);

        final LifeCycleMBean componentLifeCycle = this.containerService.getComponentLifecycleByName(componentName);

        if (componentLifeCycle != null) {
            // TODO : dirty, find a better way to do that
            ((ExtendedComponentLifeCycle) componentLifeCycle).setForcedState(force);
            try {
                componentLifeCycle.stop();
            } finally {
                ((ExtendedComponentLifeCycle) componentLifeCycle).setForcedState(false);
            }
        } else {
            throw new JBIException("The component '" + componentName + "' is not installed");
        }

        this.log.info("Component '" + componentName + "' stopped");

        this.log.end();
        return true;
    }

}
