/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.recovery.task;

import java.io.File;
import java.net.URISyntaxException;
import java.util.List;

import javax.jbi.JBIException;
import javax.jbi.management.DeploymentServiceMBean;
import javax.jbi.management.LifeCycleMBean;

import org.ow2.petals.jbi.descriptor.original.JBIDescriptorBuilder;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.microkernel.api.container.ContainerService;
import org.ow2.petals.microkernel.api.container.ServiceAssemblyLifeCycle;
import org.ow2.petals.microkernel.api.jbi.management.Context;
import org.ow2.petals.microkernel.api.jbi.management.ManagementException;
import org.ow2.petals.microkernel.api.jbi.management.TaskProcessor;
import org.ow2.petals.microkernel.api.system.SystemStateService;
import org.ow2.petals.microkernel.api.system.repository.RepositoryService;
import org.ow2.petals.microkernel.jbi.management.task.deployment.DeploymentTaskFactory;
import org.ow2.petals.systemstate.generated.ServiceAssembly;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * The task processing the recovery of a service assembly
 * 
 * @author Olivier Fabre - EBM WebSourcing
 * @author Christophe DENEUX - LIlnagora
 * 
 */
public class RecoverSaTask extends RecoverAbstractTask {

    private final ServiceAssembly saStateHolder;

    private DeploymentTaskFactory deploymentTaskFactory;

    public RecoverSaTask(final List<File> unknownDirs, final ServiceAssembly saStateHolder,
            final RepositoryService repositoryService, final SystemStateService systemStateService,
            final ContainerService containerService, final DeploymentTaskFactory deploymentTaskFactory,
            final LoggingUtil log) {
        super(unknownDirs, repositoryService, systemStateService, containerService, log);
        this.saStateHolder = saStateHolder;
        this.deploymentTaskFactory = deploymentTaskFactory;
    }

    @Override
    public RecoveryResult call() throws ManagementException {

        try {
            final File installFile = this.repositoryService
                    .getServiceAssemblyInstallDirectory(this.saStateHolder.getName());
            final Jbi descriptor = JBIDescriptorBuilder.getInstance().buildJavaJBIDescriptorFromFolder(installFile);

            this.recoverServiceAssembly(installFile, descriptor, this.saStateHolder.getLifecycleState());

            // Delete SA from the unknownDirs list because this
            // SA is known now...
            if (this.unknownDirs != null) {
                final File knownComp = installFile.getParentFile();
                if (knownComp.exists()) {
                    this.unknownDirs.remove(knownComp);
                }
            }
            return new RecoveryResult(this.saStateHolder.getName());

        } catch (final Exception e) {
            final StringBuilder msgBuffer = new StringBuilder("A service assembly consistency is corrupted: ");
            msgBuffer.append(this.saStateHolder.getName());
            // try to set the states of the SA to UNKNOWN
            try {
                this.systemStateService
                        .updateServiceAssemblyState(this.saStateHolder.getName(), LifeCycleMBean.UNKNOWN);
                msgBuffer.append("\nIts state changes to UNKNOWN");
            } catch (final Exception e1) {
                // do nothing
            }
            msgBuffer.append("\nPlease check the repository consistency");
            throw new ManagementException(msgBuffer.toString(), e);
        }
    }

    /**
     * Recover the given Service Assembly. Recovering a Service Assembly consist in redoing deployment process of a
     * Service Assembly (see <code>DeploymentServiceMBean</code>) and putting the Service Assembly in its previous state
     * by calling "start" or/and "stop" on the created component lifecycle.
     * 
     * @param installationRoot
     * @param lifecycleState
     * @throws ManagementException
     */
    private void recoverServiceAssembly(final File installRoot, final Jbi descriptor, final String lifecycleState)
            throws ManagementException {
        this.log.start();

        final String saName = descriptor.getServiceAssembly().getIdentification().getName();
        try {

            // Create deployment execution context
            Context deploymentContext = null;

            try {
                deploymentContext = this.createDeploymentContext(installRoot, descriptor);
            } catch (URISyntaxException e) {
                throw new ManagementException("Failed to build recovery context for Service Assembly '"
                        + descriptor.getServiceAssembly().getIdentification().getName() + "'", e);
            }
            
            // Recover the given service assembly
            this.recoverSA(deploymentContext);

            // TODO log the returned values!
            if (DeploymentServiceMBean.STOPPED.equals(lifecycleState)) {
                // Restore the STOPPED state
                final ServiceAssemblyLifeCycle salc = this.containerService.getServiceAssemblyByName(saName);
                salc.init();
            } else if (DeploymentServiceMBean.STARTED.equals(lifecycleState)) {
                // Restore the STARTED state
                final ServiceAssemblyLifeCycle salc = this.containerService.getServiceAssemblyByName(saName);
                salc.start();
            } else if (DeploymentServiceMBean.SHUTDOWN.equals(lifecycleState)) {
                // Restore the SHUTDOWN state
                final ServiceAssemblyLifeCycle salc = this.containerService.getServiceAssemblyByName(saName);
                salc.init();
                salc.shutDown();
            } else {
                this.log.warning("The Service Assembly '" + saName + "' was in an unknown state "
                        + ", it is recovered in state " + DeploymentServiceMBean.SHUTDOWN);
                try {
                    this.systemStateService.updateServiceAssemblyState(saName, DeploymentServiceMBean.SHUTDOWN);
                } catch (final Exception e) {
                    this.log.error("The Service Assembly '" + saName + "' failed to be recovered at state "
                            + DeploymentServiceMBean.SHUTDOWN, e);
                }
                final ServiceAssemblyLifeCycle salc = this.containerService.getServiceAssemblyByName(saName);
                salc.init();
                salc.shutDown();
            }
        } catch (final JBIException e) {
            throw new ManagementException("Failed to recover Service Assembly '" + saName + "'", e);
        }

        this.log.end();
    }

    /**
     * Process the deployment of a given Service assembly
     * 
     * @param saDeploymentContext
     *            the deployment context of the given SA
     * @throws Exception
     *             if an error occurred during deployment process
     */
    private void recoverSA(final Context deploymentContext) throws ManagementException {
        final Jbi descriptor = deploymentContext.getDescriptor();

        // Create deployment processor and fill it with tasks
        final TaskProcessor processor = this.createSARecoveryProcessor(deploymentContext);

        // Launch the recover processor
        final Throwable processResult = processor.process();
        if (processResult != null) {
            throw new ManagementException("Failed to recover Service Assembly '"
                    + descriptor.getServiceAssembly().getIdentification().getName() + "'", processResult);
        }
    }

    /**
     * Create the deployment processor and add all required task for the complete recovery of a service assembly
     * 
     * @param deploymentContext
     *            the deployment context
     * @return the filled deployment processor
     */
    private TaskProcessor createSARecoveryProcessor(final Context deploymentContext) {
        final TaskProcessor processor = new TaskProcessor(deploymentContext, this.log);
        processor.addTask(this.deploymentTaskFactory.getSaCheckPackageTask());
        processor.addTask(this.deploymentTaskFactory.getCreateXMLDeploymentResultTask());
        processor.addTask(this.deploymentTaskFactory.getRetrieveSAAndSUInstallRootsTask());
        processor.addTask(this.deploymentTaskFactory.getRegisterSALifeCycleTask());
        processor.addTask(this.deploymentTaskFactory.getDeployAllSUTask());
        processor.addTask(this.deploymentTaskFactory.getRegisterAllConnectionTask());
        return processor;
    }

    /**
     * Create the deployment context
     * 
     * @param descriptor
     *            the jbi descriptor
     * @return the Deployment Context HashMap
     */
    protected Context createDeploymentContext(final File installFile, final Jbi descriptor) throws URISyntaxException {
        final Context deploymentContext = new Context();
        deploymentContext.setUnzipRoot(installFile);
        deploymentContext.setDescriptor(descriptor);
        return deploymentContext;
    }
}
