/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.recovery.task;

import java.io.File;
import java.io.IOException;
import java.util.List;

import org.ow2.petals.jbi.descriptor.JBIDescriptorException;
import org.ow2.petals.jbi.descriptor.original.JBIDescriptorBuilder;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.microkernel.api.container.ContainerService;
import org.ow2.petals.microkernel.api.jbi.management.Context;
import org.ow2.petals.microkernel.api.jbi.management.ManagementException;
import org.ow2.petals.microkernel.api.jbi.management.TaskProcessor;
import org.ow2.petals.microkernel.api.system.SystemStateService;
import org.ow2.petals.microkernel.api.system.repository.RepositoryService;
import org.ow2.petals.microkernel.jbi.management.task.installation.InstallationTaskFactory;
import org.ow2.petals.systemstate.generated.SharedLibrary;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * The task processing the recovery of a shared library
 * 
 * @author Olivier Fabre - EBM WebSourcing
 * @author Christophe DENEUX - LIlnagora
 * 
 */
public class RecoverSlTask extends RecoverAbstractInstallationTask {

    private final SharedLibrary sharedLibraryState;

    private final InstallationTaskFactory installationTaskFactory;

    public RecoverSlTask(final List<File> unknownDirs, final SharedLibrary sharedLibraryState,
            final RepositoryService repositoryService, final SystemStateService systemStateService,
            final ContainerService containerService, final InstallationTaskFactory installationTaskFactory,
            final LoggingUtil log) {
        super(unknownDirs, repositoryService, systemStateService, containerService, log);
        this.sharedLibraryState = sharedLibraryState;
        this.installationTaskFactory = installationTaskFactory;
    }

    @Override
    public RecoveryResult call() throws ManagementException {

        try {
            final File installFile = this.repositoryService
                    .getSharedLibraryInstallDirectory(sharedLibraryState.getName(), sharedLibraryState.getVersion());

            final Jbi descriptor = JBIDescriptorBuilder.getInstance().buildJavaJBIDescriptorFromFolder(installFile);

            this.recoverSharedLib(installFile, descriptor);

            // Delete SL from the unknownDirs list because this
            // SL is known now...
            if (this.unknownDirs != null) {
                final File knownComp = installFile.getParentFile();
                if (knownComp.exists()) {
                    this.unknownDirs.remove(knownComp);
                }
            }

            return new RecoveryResult(sharedLibraryState.getName());
        } catch (final IOException | JBIDescriptorException e) {
            final String msg = "A Shared Library can't be recovered : " + sharedLibraryState.getName()
                    + "\nPlease check the repository consistency";
            throw new ManagementException(msg, e);
        }
    }

    /**
     * Recover the given shared library. Recovering a shared library consist in redoing installation process of a shared
     * library (see <code>InstallationServiceMBean</code>).
     * 
     * @param installationRoot
     * @throws ManagementException
     */
    protected void recoverSharedLib(final File installFile, final Jbi descriptor) throws ManagementException {
        this.log.start();

        // Create installation execution context
        Context installationContext = null;

        installationContext = this.createRecoveryContext(descriptor, installFile);
        
        // Re-process installation of the shared libraries
        this.recoverSharedLib(installationContext);

        this.log.end();
    }

    /**
     * Reload the shared lib classloader for the given shared lib
     * 
     * @param installationContext
     *            the installation context for the recovery process : handles descriptor, archive URI and installRoot
     *            information
     * @throws ManagementException
     *             if the shared library classloader can't be recovered
     */
    private void recoverSharedLib(final Context installationContext) throws ManagementException {
        final Jbi descriptor = installationContext.getDescriptor();

        // Create installation processor and fill it with tasks
        final TaskProcessor processor = this.createSLRecoveryProcessor(installationContext);

        // Launch the recover processor
        final Throwable processResult = processor.process();
        if (processResult != null) {
            throw new ManagementException("Failed to recover Shared Library '"
                    + descriptor.getSharedLibrary().getIdentification().getName() + "'", processResult);
        }
    }

    /**
     * Create the installation processor and add all required task for the complete recovery of a shared lib
     * 
     * @param installationContext
     *            the installation context
     * @return the filled installation processor
     */
    private TaskProcessor createSLRecoveryProcessor(final Context installationContext) {
        final TaskProcessor processor = new TaskProcessor(installationContext, this.log);
        processor.addTask(this.installationTaskFactory.getSLCheckPackageTask());
        processor.addTask(this.installationTaskFactory.getCreateSLClassloaderTask());
        return processor;
    }
}
