/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.task.deployment.deploy;

import java.io.File;
import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.jbi.JBIException;
import javax.jbi.management.DeploymentException;

import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.jbi.descriptor.original.generated.ServiceUnit;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.container.ServiceAssemblyLifeCycle;
import org.ow2.petals.microkernel.api.jbi.management.Context;
import org.ow2.petals.microkernel.api.jbi.management.ManagementException;
import org.ow2.petals.microkernel.api.jbi.management.TaskProcessor;
import org.ow2.petals.microkernel.api.jbi.management.XMLResult;
import org.ow2.petals.microkernel.jbi.management.task.AbstractLoggableTask;
import org.ow2.petals.microkernel.jbi.management.task.deployment.DeploymentTaskFactory;
import org.ow2.petals.microkernel.jbi.management.util.ManagementMessageUtil;

import com.ebmwebsourcing.easycommons.lang.ExceptionHelper;
import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * This task creates service unit lifecycle for each Service Unit included into the service assembly, register it to the
 * Service Assembly then executes the deployment of the service units.
 * 
 * <p>
 * Inputs : {@link Jbi}, sa install root {@link File}, {@link XMLResult}, SUInstallRoots {@link Map}, SU task processors
 * {@link List} <br/>
 * Outputs : {@link ServiceAssemblyLifeCycle}<br/>
 * The task has a robust approach: it stops at the first subtask failed and has undo capability.
 * </p>
 * 
 * @author Olivier Fabre - EBM WebSourcing
 * @author Frederic Gardes - EBM WebSourcing
 */
public class DeployAllSUTask extends AbstractLoggableTask {

    protected DeploymentTaskFactory deploymentTaskFactory;

    public DeployAllSUTask(final LoggingUtil log,
            final DeploymentTaskFactory deploymentTaskFactory,
            final ContainerConfiguration containerConfiguration) {
        super(log, containerConfiguration);
        this.deploymentTaskFactory = deploymentTaskFactory;
    }

    @Override
    public void execute(final Context context) throws Exception {
        Jbi descriptor = context.getDescriptor();
        XMLResult xmlResult = context.getXmlResult();
        Map<String, File> suInstallFiles = context.getSuInstallFiles();
        ServiceAssemblyLifeCycle saLifeCycle = context.getSaLifeCycle();

        // For each Service Unit included into the service assembly : execute
        // the deployment process of a service unit
        List<TaskProcessor> suTaskProcessors = this.deploySUS(saLifeCycle, descriptor, suInstallFiles, xmlResult);

        if (suTaskProcessors.size() == 0) {
            throw new DeploymentException("All the Service Unit deployment has failed");
        }

        context.setSuTaskProcessors(suTaskProcessors);
    }

    /**
     * Deploy all Service Units to the JBI container
     * 
     * @param installationRoot
     *            the root {@link URI} of the service assembly
     * @param saLifeCycle
     *            the {@link ServiceAssemblyLifeCycle} of the service assembly
     * @param descriptor
     *            the service assembly jbi descriptor
     * @param suInstallRoots
     *            a map of all SU install roots
     * @param xmlResult
     *            xml task result
     * @throws ManagementException
     *             if an sus deployment fails caused by a severe error of the
     *             container
     * 
     */
    protected List<TaskProcessor> deploySUS(final ServiceAssemblyLifeCycle saLifeCycle, final Jbi descriptor,
            final Map<String, File> suInstallFiles, final XMLResult xmlResult) throws JBIException {
        final List<ServiceUnit> serviceUnits = descriptor.getServiceAssembly().getServiceUnit();
        final List<TaskProcessor> suTaskProcessors = new ArrayList<TaskProcessor>();

        if (serviceUnits != null) {
            final List<Context> suDeploymentContexts = new ArrayList<Context>(serviceUnits.size());
            final StringBuilder componentOutput = new StringBuilder();
            for (ServiceUnit serviceUnit : serviceUnits) {
                // Create Deployment context
                final Context suDeploymentContext = this.createSUDeploymentContext(saLifeCycle, suInstallFiles,
                        serviceUnit);
                suDeploymentContexts.add(suDeploymentContext);
                // Process SU deployment
                this.registerAndDeploySU(suDeploymentContext, suTaskProcessors);
                componentOutput.append(suDeploymentContext.getComponentResult());
            }

            xmlResult.addComponentTaskResult(componentOutput);
        }

        return suTaskProcessors;
    }

    private Context createSUDeploymentContext(ServiceAssemblyLifeCycle saLifeCycle, Map<String, File> suInstallFiles,
            ServiceUnit serviceUnitInSa) {
        Context suDeploymentContext = new Context();
        suDeploymentContext.setSaLifeCycle(saLifeCycle);
        suDeploymentContext.setSuDescriptorInSa(serviceUnitInSa);
        File suInstallFile = suInstallFiles.get(serviceUnitInSa.getIdentification().getName());
        suDeploymentContext.setSuInstallFile(suInstallFile);

        return suDeploymentContext;
    }

    /**
     * Process the deployment of a given SU
     * 
     * @param suDeploymentContext
     *            the deployment context of the given SU
     * @param suTaskProcessors
     *            a storage that handles all previously executed SU deployment
     *            processor
     */
    private void registerAndDeploySU(Context suDeploymentContext, List<TaskProcessor> suTaskProcessors) {
        // Create processor and add tasks
        TaskProcessor processor = new TaskProcessor(suDeploymentContext, this.log);
        processor.addTask(this.deploymentTaskFactory.getCreateAndRegisterSULifeCycleTask());
        processor.addTask(this.deploymentTaskFactory.getDeploySUToComponentTask());

        // Launch processor
        Throwable result = processor.process();
        if (result != null) {
            if (suDeploymentContext.getComponentResult() != null) {
                // it is an error from the component' Service Unit
                this.log.warning("A component Service Unit Manager failed to deploy a Service Unit :\n"
                        + suDeploymentContext.getComponentResult());
            } else {
                final String xmlResult = ManagementMessageUtil.getComponentTaskResult(
                        suDeploymentContext.getSuDescriptorInSa().getTarget().getComponentName(),
                        "deploy", ManagementMessageUtil.TASK_RESULT_FAILED,
                        ManagementMessageUtil.MESSAGE_TYPE_ERROR, "2", "Failed to deploy SU : {1}",
                        new String[] { suDeploymentContext.getSuDescriptorInSa()
                                .getIdentification().getName() },
                        ExceptionHelper.getStackTrace(result));
                suDeploymentContext.setComponentResult(xmlResult);
                this.log.warning("Failed to deployment the Service Unit '"
                        + suDeploymentContext.getSuDescriptorInSa().getIdentification().getName()
                        + "' :\n" + xmlResult);
            }
        } else {
            suTaskProcessors.add(processor);
        }
    }

    @Override
    public void undo(Context context) throws Exception {
        final List<TaskProcessor> suTaskProcessors = context.getSuTaskProcessors();
        if (suTaskProcessors != null) {
            for (TaskProcessor processor : suTaskProcessors) {
                processor.unprocess();
            }
        }
    }

}
