/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.task.deployment.deploy;

import java.io.File;
import java.net.URL;

import javax.jbi.component.ServiceUnitManager;

import org.ow2.petals.jbi.descriptor.original.generated.ServiceUnit;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.container.ContainerService;
import org.ow2.petals.microkernel.api.container.Installer;
import org.ow2.petals.microkernel.api.jbi.management.Context;
import org.ow2.petals.microkernel.api.jbi.management.ManagementException;
import org.ow2.petals.microkernel.jbi.management.task.AbstractLoggableTask;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * This task deploys service unit to the targeted component.
 * <p>
 * This task MUST be executed with a {@link Context} containing:
 * <ul>
 * <li>the service unit part of the service-assembly JBI descriptor ( {@link ServiceUnit}),</li>
 * <li>the root directory of the service-unit installation as a {@link File},</li>
 * </ul>
 * and returns following data in the {@link Context}:
 * <ul>
 * <li>the XML result for of the deployment as {@link String}.</li>
 * </ul>
 * </p>
 * <p>
 * The task has got a robust approach : it stops at the first subtask failed and has undo capability.
 * </p>
 * 
 * @author Frederic Gardes - EBM WebSourcing
 * @author Olivier Fabre - EBM WebSourcing
 */
public class DeploySUToComponentTask extends AbstractLoggableTask {

    /**
     * JBI Container :: Container service
     */
    protected final ContainerService containerService;

    public DeploySUToComponentTask(final LoggingUtil log, final ContainerService containerService,
            final ContainerConfiguration containerConfiguration) {
        super(log, containerConfiguration);
        this.containerService = containerService;
    }

    @Override
    public void execute(Context context) throws Exception {
        final File installFile = context.getSuInstallFile();
        final ServiceUnit serviceUnit = context.getSuDescriptorInSa();
        String xmlResult = null;
        xmlResult = this.deploySUtoComponent(installFile, serviceUnit);
        context.setComponentResult(xmlResult);
        if (xmlResult.contains("<task-result>FAILED</task-result>")) {
            // throw an exception to stop and trigger the reversion of the processor
            throw new Exception("Task failed");
        }
    }

    /**
     * Deploy the given service unit to the targeted component. The component
     * must be in the STARTED state to allow deployment of service units.
     * 
     * @param installationRoot
     *            the SU installation {@link URL}
     * @param serviceUnit
     *            the service unit production element
     * @return a {@link String} xml result of the service unit deployment on the
     *         targeted component (see section 6.9 of the JSR 208)
     * @throws ManagementException
     *             if component and related object are missing, if the call of
     *             the service unit manager method "deploy" fails or if the
     *             component isn't in the STARTED state
     */
    protected String deploySUtoComponent(File installFile, ServiceUnit serviceUnit) throws Exception {
        String xmlReturn = null;
        // Retrieve the target component life cycle
        String targetComponent = serviceUnit.getTarget().getComponentName();
        Installer installer = this.containerService.getInstallerByName(targetComponent);
        if (installer == null) {
            throw new ManagementException(
                    "The Service Unit can not be deployed, as the component '" + targetComponent
                            + "' is not installed");
        }

        // Retrieve the service unit manager of the component that allow to
        // process deployment
        ServiceUnitManager sum = installer.getComponent().getServiceUnitManager();
        if (sum == null) {
            throw new ManagementException("Component '" + targetComponent
                    + "' doesn't support deployment : service unit manager is null");
        }
        // Set the context class loader to the component's one just for the
        // method invocation
        final String serviceUnitName = serviceUnit.getIdentification().getName();
        Thread.currentThread().setContextClassLoader(sum.getClass().getClassLoader());
        try {
            xmlReturn = sum.deploy(serviceUnitName, installFile.getAbsolutePath());
        } finally {
            Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());
        }

        return xmlReturn;
    }

    @Override
    public void undo(Context context) throws Exception {
        // nothing to do, it is the responsibility of the component to handle problems of deployment, we should expect
        // it to revert to a reasonable state!
        this.log.debug("Revert DeploySUToComponent (nothing to do)");
    }
}
