/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.task.deployment.deploy;

import java.util.ArrayList;
import java.util.List;

import javax.xml.namespace.QName;

import org.ow2.petals.clientserverapi.jbi.messaging.endpoint.exception.EndpointDirectoryException;
import org.ow2.petals.jbi.descriptor.original.generated.Connection;
import org.ow2.petals.jbi.descriptor.original.generated.Connections;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.jbi.management.Context;
import org.ow2.petals.microkernel.api.jbi.management.ManagementException;
import org.ow2.petals.microkernel.api.jbi.messaging.endpoint.EndpointDirectoryService;
import org.ow2.petals.microkernel.jbi.management.task.AbstractLoggableTask;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * This task registers connections if they exist.
 * <p>
 * Inputs : {@link Jbi}<br/>
 * The task has a robust approach: it stops at the first subtask failed
 * and has undo capability.
 * </p>
 * 
 * @author Olivier Fabre - EBM WebSourcing
 * @author Frederic Gardes - EBM WebSourcing
 */
public class RegisterAllConnectionTask extends AbstractLoggableTask {

    protected final EndpointDirectoryService endpointDirectory;

    public RegisterAllConnectionTask(final LoggingUtil log, final EndpointDirectoryService endpointDirectory,
            final ContainerConfiguration containerConfiguration) {
        super(log, containerConfiguration);
        this.endpointDirectory = endpointDirectory;
    }

    @Override
    public void execute(final Context context) throws Exception {
        final Jbi descriptor = context.getDescriptor();
        final Connections connections = descriptor.getServiceAssembly().getConnections();
        final List<Connection> successfulyRegisteredConnections = new ArrayList<Connection>();
        if (connections != null) {
            for (final Connection connection : connections.getConnection()) {
                try {
                    this.registerConnection(connection, descriptor);
                } catch (final ManagementException e) {
                    // Unregister previously registered connection
                    for (final Connection successConnection : successfulyRegisteredConnections) {
                        try {
                            this.removeConnection(successConnection);
                        } catch (final Exception e1) {
                            this.log.error("Can't delete an already registered connection");
                        }
                    }
                    throw e;
                }
                successfulyRegisteredConnections.add(connection);
            }
        }
    }

    /**
     * Register a connection into the endpoint registry of the JBI container
     * 
     * @param connection
     *            The {@link Connection} to register
     * @param descriptor
     *            the service assembly descriptor
     * @throws ManagementException
     *             when container fails to register a connection into the
     *             endpoint registry
     */
    protected void registerConnection(final Connection connection, final Jbi descriptor)
            throws ManagementException {
        final String saName = descriptor.getServiceAssembly().getIdentification().getName();
        final QName consInterface = connection.getConsumer().getInterfaceName();
        final QName consService = connection.getConsumer().getServiceName();
        final String consEndpoint = connection.getConsumer().getEndpointName();
        final QName provService = connection.getProvider().getServiceName();
        final String provEndpoint = connection.getProvider().getEndpointName();
        try {
            // TODO service connection they must be restricted to a given SA, see PETALSESBCONT-391
            if (consInterface != null) {
                this.endpointDirectory.registerConnection(consInterface, provService, provEndpoint);
            } else {
                this.endpointDirectory.registerConnection(consService, consEndpoint, provService,
                        provEndpoint);
            }
        } catch (final EndpointDirectoryException e) {
            throw new ManagementException(
                    "Error while registering a service unit connection for the service assembly '"
                            + saName + "'", e);
        }

    }

    @Override
    public void undo(final Context context) throws Exception {
        final Jbi descriptor = context.getDescriptor();
        if (descriptor != null) {
            final Connections connections = descriptor.getServiceAssembly().getConnections();
            if (connections != null) {
                for (final Connection connection : connections.getConnection()) {
                    try {
                        this.removeConnection(connection);
                    } catch (Exception e) {
                        this.log.error("Failed to revert a RegisterAllConnectionTask", e);
                    }
                }
            }
        }
    }

    /**
     * Remove a connection registered during the SA deployment process
     * 
     * @param connection
     *            a Connection to remove
     * @throws Exception
     *             if problem occurs during the deletion of a specific
     *             connection
     */
    protected void removeConnection(final Connection connection) throws Exception {
        if ((connection != null) && (connection.getConsumer() != null)) {
            final QName consInterface = connection.getConsumer().getInterfaceName();
            final QName consService = connection.getConsumer().getServiceName();
            final String consEndpoint = connection.getConsumer().getEndpointName();
            final QName provService = connection.getProvider().getServiceName();
            final String provEndpoint = connection.getProvider().getEndpointName();
            if ((provEndpoint != null) && (provService != null)) {
                // TODO service connection they must be restricted to a given SA, see PETALSESBCONT-391
                if (consInterface != null) {
                    this.endpointDirectory.deregisterConnection(consInterface, provService,
                            provEndpoint);
                } else if ((consService != null) && (consEndpoint != null)) {
                    this.endpointDirectory.deregisterConnection(consService, consEndpoint,
                            provService, provEndpoint);
                }
            }
        }
    }

}
