/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.task.installation;

import java.util.List;

import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.container.ContainerService;
import org.ow2.petals.microkernel.api.extension.InstallationExtension;
import org.ow2.petals.microkernel.api.jbi.messaging.RouterService;
import org.ow2.petals.microkernel.api.system.SystemStateService;
import org.ow2.petals.microkernel.api.system.repository.RepositoryService;
import org.ow2.petals.microkernel.jbi.management.task.CopyArchiveToInstalledDirTask;
import org.ow2.petals.microkernel.jbi.management.task.CopyArchiveToWorkDirTask;
import org.ow2.petals.microkernel.jbi.management.task.DownloadTask;
import org.ow2.petals.microkernel.jbi.management.task.ExtractTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.install.component.ComponentCheckPackageTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.install.component.CreateAndRegisterInstallerMBeanTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.install.component.CreateComponentRepositoryTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.install.component.CreateComponentStateHolderTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.install.component.RegisterComponentInSLTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.install.sl.CreateSLClassloaderTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.install.sl.CreateSLRepositoryTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.install.sl.CreateSLStateHolderTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.install.sl.SLCheckPackageTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.uninstall.component.ComponentCheckInstalledTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.uninstall.component.RemoveComponentArchiveTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.uninstall.component.RemoveComponentRepositoryTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.uninstall.component.RemoveComponentStateHolderTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.uninstall.component.UnregisterComponentFromSLTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.uninstall.component.UnregisterInstallerTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.uninstall.sl.RemoveSLArchiveTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.uninstall.sl.RemoveSLClassloaderTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.uninstall.sl.RemoveSLRepositoryTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.uninstall.sl.RemoveSLStateHolderTask;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * This Factory creates Installation Tasks.
 * @author Olivier Fabre - EBM WebSourcing
 */
public class InstallationTaskFactory {
    private final CopyArchiveToWorkDirTask copyArchiveToWorkDirTask;

    private final RegisterComponentInSLTask registerComponentInSLTask;

    private final CreateComponentRepositoryTask createComponentFolderTask;

    private final ComponentCheckPackageTask componentCheckPackageTask;

    private final ComponentCheckInstalledTask componentCheckInstalledTask;

    private final CreateComponentStateHolderTask createComponentStateHolderTask;

    private final CreateAndRegisterInstallerMBeanTask createAndRegisterInstallerMBeanTask;

    private final CreateSLClassloaderTask createSLClassloaderTask;

    private final CreateSLRepositoryTask createSLRepositoryTask;

    private final SLCheckPackageTask slCheckPackageTask;

    private final CreateSLStateHolderTask createSLStateHolderTask;

    private final CopyArchiveToInstalledDirTask copyArchiveToInstalledDirTask;

    private final RemoveComponentRepositoryTask removeComponentFolderTask;

    private final RemoveComponentStateHolderTask removeComponentStateHolderTask;

    private final RemoveComponentArchiveTask removeComponentArchiveTask;

    private final UnregisterInstallerTask unregisterInstallerTask;

    private final UnregisterComponentFromSLTask unregisterComponentFromSLTask;

    private final RemoveSLArchiveTask removeSLArchiveTask;

    private final RemoveSLClassloaderTask removeSLClassloaderTask;

    private final RemoveSLRepositoryTask removeSLFolderTask;

    private final RemoveSLStateHolderTask removeSLStateHolderTask;

    private final ExtractTask extractTask;

    private final DownloadTask downloadTask;

    /**
     * Creates a new instance of {@link InstallationTaskFactory}
     * 
     * @param packageHandler
     * @param systemStateService
     * @param repositoryService
     * @param log
     * @param adminService
     * @param router
     * @param loaderService
     */
    public InstallationTaskFactory(SystemStateService systemStateService, RepositoryService repositoryService,
            LoggingUtil log, RouterService router, final ContainerConfiguration containerConfiguration,
            final ContainerService containerService, final List<InstallationExtension> installationExtensions) {
        super();

        /*
         * Create Tasks
         */
        // Component creation tasks
        this.registerComponentInSLTask = new RegisterComponentInSLTask(containerService);

        this.createComponentFolderTask = new CreateComponentRepositoryTask(log, repositoryService,
                containerConfiguration);

        this.componentCheckPackageTask = new ComponentCheckPackageTask(log, containerService,
                containerConfiguration);

        this.createComponentStateHolderTask = new CreateComponentStateHolderTask(log,
                systemStateService, containerConfiguration);

        this.createAndRegisterInstallerMBeanTask = new CreateAndRegisterInstallerMBeanTask(log, containerService,
                containerConfiguration);

        // SL creation tasks
        this.createSLClassloaderTask = new CreateSLClassloaderTask(log, containerService,
                containerConfiguration);

        this.createSLRepositoryTask = new CreateSLRepositoryTask(log, repositoryService,
                containerConfiguration);

        this.slCheckPackageTask = new SLCheckPackageTask(log, containerService, containerConfiguration);

        this.createSLStateHolderTask = new CreateSLStateHolderTask(log, systemStateService,
                containerConfiguration);

        this.copyArchiveToInstalledDirTask = new CopyArchiveToInstalledDirTask(log,
                containerConfiguration, installationExtensions);

        // Component removal tasks
        this.componentCheckInstalledTask = new ComponentCheckInstalledTask(log, containerService,
                containerConfiguration);

        this.removeComponentFolderTask = new RemoveComponentRepositoryTask(log, repositoryService,
                containerConfiguration);

        this.removeComponentStateHolderTask = new RemoveComponentStateHolderTask(log,
                systemStateService, containerConfiguration);

        this.removeComponentArchiveTask = new RemoveComponentArchiveTask(log,
                containerConfiguration, installationExtensions);

        this.unregisterInstallerTask = new UnregisterInstallerTask(log, containerService, containerConfiguration);

        this.unregisterComponentFromSLTask = new UnregisterComponentFromSLTask(log, containerService,
                containerConfiguration);

        // SL removal tasks
        this.removeSLArchiveTask = new RemoveSLArchiveTask(log, containerConfiguration,
                installationExtensions);

        this.removeSLClassloaderTask = new RemoveSLClassloaderTask(log, containerService,
                containerConfiguration);

        this.removeSLFolderTask = new RemoveSLRepositoryTask(log, repositoryService,
                containerConfiguration);

        this.removeSLStateHolderTask = new RemoveSLStateHolderTask(log, systemStateService,
                containerConfiguration);

        // shared tasks
        this.extractTask = new ExtractTask(log, containerConfiguration);

        this.downloadTask = new DownloadTask(log, containerConfiguration);

        this.copyArchiveToWorkDirTask = new CopyArchiveToWorkDirTask(log, containerConfiguration);
    }

    public DownloadTask getDownloadTask() {
        return this.downloadTask;
    }

    public CopyArchiveToWorkDirTask getCopyArchiveToWorkDirTask() {
        return this.copyArchiveToWorkDirTask;
    }

    public RegisterComponentInSLTask getRegisterComponentInSLTask() {
        return this.registerComponentInSLTask;
    }

    public CreateComponentRepositoryTask getCreateComponentFolderTask() {
        return this.createComponentFolderTask;
    }

    public ComponentCheckPackageTask getComponentCheckPackageTask() {
        return this.componentCheckPackageTask;
    }

    public ComponentCheckInstalledTask getComponentCheckInstalledTask() {
        return this.componentCheckInstalledTask;
    }

    public CreateComponentStateHolderTask getCreateComponentStateHolderTask() {
        return this.createComponentStateHolderTask;
    }

    public CopyArchiveToInstalledDirTask getCopyArchiveToInstalledDirTask() {
        return this.copyArchiveToInstalledDirTask;
    }

    public CreateAndRegisterInstallerMBeanTask getCreateAndRegisterInstallerMBeanTask() {
        return this.createAndRegisterInstallerMBeanTask;
    }

    public CreateSLClassloaderTask getCreateSLClassloaderTask() {
        return this.createSLClassloaderTask;
    }

    public CreateSLRepositoryTask getCreateSLRepositoryTask() {
        return this.createSLRepositoryTask;
    }

    public SLCheckPackageTask getSLCheckPackageTask() {
        return this.slCheckPackageTask;
    }

    public CreateSLStateHolderTask getCreateSLStateHolderTask() {
        return this.createSLStateHolderTask;
    }

    public RemoveComponentArchiveTask getRemoveComponentArchiveTask() {
        return this.removeComponentArchiveTask;
    }

    public RemoveSLArchiveTask getRemoveSLArchiveTask() {
        return this.removeSLArchiveTask;
    }

    public UnregisterInstallerTask getUnregisterInstallerTask() {
        return this.unregisterInstallerTask;
    }

    public UnregisterComponentFromSLTask getUnregisterComponentFromSLTask() {
        return this.unregisterComponentFromSLTask;
    }

    public RemoveSLClassloaderTask getRemoveSLClassloaderTask() {
        return this.removeSLClassloaderTask;
    }

    public RemoveSLRepositoryTask getRemoveSLFolderTask() {
        return this.removeSLFolderTask;
    }

    public RemoveSLStateHolderTask getRemoveSLStateHolderTask() {
        return this.removeSLStateHolderTask;
    }

    public RemoveComponentRepositoryTask getRemoveComponentFolderTask() {
        return this.removeComponentFolderTask;
    }

    public RemoveComponentStateHolderTask getRemoveComponentStateHolderTask() {
        return this.removeComponentStateHolderTask;
    }

    public ExtractTask getExtractTask() {
        return this.extractTask;
    }

}
