/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.task.installation.install.component;

import java.io.File;
import java.io.IOException;

import javax.jbi.JBIException;
import javax.management.ObjectName;
import javax.naming.NamingException;

import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.container.ContainerService;
import org.ow2.petals.microkernel.api.container.Installer;
import org.ow2.petals.microkernel.api.container.exception.ContainerServiceException;
import org.ow2.petals.microkernel.api.jbi.management.Context;
import org.ow2.petals.microkernel.jbi.management.task.AbstractLoggableTask;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * This task creates the InstallerMBean for a component and register it in the JMX server.
 * <p>
 * The task has got a non-robust approach : it stops at the first subtask failed <br />
 * and have the undo capability.<br />
 * Inputs : {@link Jbi}, component install root {@link File}, ComponentContext<br/>
 * Outputs : component installer {@link ObjectName}
 * </p>
 * 
 * @author Olivier Fabre - EBM WebSourcing
 * @author Frederic Gardes - EBM WebSourcing
 */
public class CreateAndRegisterInstallerMBeanTask extends AbstractLoggableTask {

    private ContainerService containerService;

    /**
     * Creates a new instance of {@link CreateAndRegisterInstallerMBeanTask}
     * 
     * @param log
     * @param containerService
     */
    public CreateAndRegisterInstallerMBeanTask(final LoggingUtil log, final ContainerService container,
            final ContainerConfiguration containerConfiguration) {
        super(log, containerConfiguration);
        this.containerService = container;
    }

    @Override
    public void execute(Context context) throws Exception {
        this.createAndRegisterInstallerMBean(context);
    }

    /**
     * Create and register a new InstallerMBean for the loaded component
     * 
     * @param componentContext
     *            the new Component execution context
     * @param descriptor
     *            the component Jbi
     * @param installationRoot
     *            the URI representing the component installation folder
     * @return the new InstallerMBean name
     * @throws PetalsException
     */
    private void createAndRegisterInstallerMBean(Context ctxt) throws PetalsException {
        try {
            final Jbi descriptor = ctxt.getDescriptor();
            final org.ow2.petals.jbi.descriptor.original.generated.Component componentDescription = descriptor
                    .getComponent();
            // Create the JBI component installer
            final Installer installer = this.containerService.createInstaller(componentDescription);
            // Initialize the JBI component installer
            installer.init(ctxt);
            // Register the InstallerMBean
            final ObjectName on = this.containerService
                    .registerInstallerMBean(componentDescription.getIdentification().getName());

            ctxt.setInstaller(installer);
            ctxt.setInstallerObjectName(on);
        } catch (final JBIException e) {
            throw new PetalsException("Error creating or registring installer MBean", e);
        } catch (IOException e) {
            throw new PetalsException("Error creating or registring installer MBean", e);
        } catch (NamingException e) {
            throw new PetalsException("Error creating or registring installer MBean", e);
        }
    }

    @Override
    public void undo(Context context) throws Exception {
        final Jbi descriptor = context.getDescriptor();
        final String componentName = descriptor.getComponent().getIdentification().getName();

        try {
            // Unregister the InstallerMBean
            this.containerService.unregisterInstallerMBean(componentName);
        } catch (final ContainerServiceException e) {
            this.log.error("Component installer can't be unregistered or doesn't exist", e);
        }

        try {
            this.containerService.removeInstaller(componentName);
        } catch (final ContainerServiceException e) {
            this.log.error("Fractal installer can't be removed or doesn't exist", e);
        }
    }
}
