/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.task.installation.install.sl;

import java.io.File;

import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi.SharedLibrary;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.container.ContainerService;
import org.ow2.petals.microkernel.jbi.management.task.AbstractCheckPackageTask;
import org.ow2.petals.microkernel.jbi.management.task.installation.InstallationUtils;

import com.ebmwebsourcing.easycommons.lang.StringHelper;
import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * This task checks the validity of the Shared Library package.
 * @author Olivier Fabre - EBM WebSourcing
 */
public class SLCheckPackageTask extends AbstractCheckPackageTask {

    /**
     * JBI Container Service :: Container Service
     */
    private final ContainerService containerService;

    public SLCheckPackageTask(final LoggingUtil log, final ContainerService containerService,
            final ContainerConfiguration containerConfiguration) {
        super(log, containerConfiguration);
        this.containerService = containerService;
    }

    /**
     * Check the validity of the Shared Library archive
     * 
     * @param descriptor
     *            the shared lib <code>Jbi</code>
     * @param filePackage
     *            the service assembly package root
     * @return {@code true} if checking is successful and {@code false}
     *         otherwise
     * @throws PetalsException
     * @throws ComponentCheckingException
     */
    @Override
    protected void checkPackage(Jbi descriptor, File filePackage) throws PetalsException {
        // check if JBI descriptor has a shared-library element
        SharedLibrary sl = descriptor.getSharedLibrary();
        if (sl == null) {
            throw new PetalsException("The shared-library element was not found in "
                    + "JBI Descriptor within installation package: "
                    + filePackage.getAbsolutePath());
        } else {
            // check if it is already loaded by package uuid
            final String slName = sl.getIdentification().getName();
            final String slVersion = sl.getVersion();
            if (StringHelper.isNullOrEmpty(slName)) {
                throw new PetalsException(
                        "The Shared Library name must be non null and non empty :"
                                + filePackage.getAbsolutePath());
            }
            if (StringHelper.isNullOrEmpty(slVersion)) {
                throw new PetalsException(
                        "The Shared Library version must be non null and non empty :"
                                + filePackage.getAbsolutePath());
            }

            if (this.containerService.getSharedLibraryByName(slName, slVersion) != null) {
                throw new PetalsException("Other Shared Library package already exists loaded in "
                        + "the component framework: " + slName);
            } else {
                // Check the classpath element of the shared library
                InstallationUtils.checkClassPathElements(filePackage, sl
                        .getSharedLibraryClassPath().getPathElement());
            }

        }
    }

    @Override
    protected String extractEntityName(Jbi descriptor) {
        return descriptor.getSharedLibrary().getIdentification().getName();
    }

    @Override
    protected String extractEntityVersion(Jbi descriptor) {
        return descriptor.getSharedLibrary().getVersion();
    }

}
