/**
 * Copyright (c) 2009-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.messaging.control;

import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.ow2.petals.clientserverapi.jbi.management.admin.AdminService;
import org.ow2.petals.jbi.servicedesc.endpoint.Location;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint;
import org.ow2.petals.jmx.api.api.JMXClient;
import org.ow2.petals.microkernel.api.communication.JMXService;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.util.ParameterCheckHelper;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * The JMX implementation of the echange checker client.
 * 
 * TODO: The exchange checker must be implemented via the transporter.
 * 
 * TODO if the component is local, then do not contact ourselves via JMX
 * 
 * @author Christophe Hamerling - EBM WebSourcing
 */
@Component(provides = @Interface(name = "service", signature = ExchangeCheckerClient.class))
public class JMXExchangeCheckerClientImpl implements ExchangeCheckerClient {

    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(Constants.FRACTAL_COMPONENT_LOGGER_NAME));

    @Requires(name = "jmx")
    private JMXService jmxService;

    @Requires(name = "admin")
    private AdminService adminService;

    @Requires(name = "configuration")
    private ConfigurationService configurationService;

    @Override
    public boolean isExchangeWithConsumerOkayForComponent(final Location providerLocation,
            final PetalsServiceEndpoint providerEndpoint, final MessageExchange exchange) {
        this.log.start();
        ParameterCheckHelper.isNullParameterWithLog(providerEndpoint, "consumerEndpoint", this.log);

        boolean test = true;
        try {
            if (this.configurationService.getContainerConfiguration().getName()
                    .equals(providerLocation.getContainerName())) {
                // we are in local, no need to use JMX
                try {
                    test = this.adminService.isExchangeWithConsumerOkayForComponent(providerLocation.getComponentName(),
                            providerEndpoint, exchange);
                } catch (final JBIException e) {
                    // there is this exception if the component does not exists on the container, so we simply return
                    // false
                    test = false;
                    this.log.error("Problem in isExchangeWithConsumerOkayForComponent, returning " + test, e);
                }
            } else {
                // Retrieve a connection to the JMX server of the container were the
                // component is deployed
                final String containerName = providerLocation.getContainerName();
                final JMXClient server = this.jmxService.getJMXClient(containerName);
                try {
                    // Ask the component if it is ok to accept the exchange
                    test = server.getAdminServiceClient().isExchangeWithConsumerOkayForComponent(
                            providerLocation.getComponentName(), providerEndpoint, exchange);
                } finally {
                    this.jmxService.returnJMXClient(containerName, server);
                }
            }
        } catch (final Exception e) {
            this.log.error("Problem in isExchangeWithConsumerOkayForComponent, returning " + test, e);
        }
        this.log.end();
        return test;
    }

    @Override
    public boolean isExchangeWithProviderOkayForComponent(final Location consumerLocation,
            final PetalsServiceEndpoint providerEndpoint, final MessageExchange exchange) {
        this.log.start();
        ParameterCheckHelper.isNullParameterWithLog(providerEndpoint, "providerEndpoint", this.log);
        boolean test = true;
        try {
            if (this.configurationService.getContainerConfiguration().getName()
                    .equals(consumerLocation.getContainerName())) {
                // we are in local, no need to use JMX
                try {
                    test = this.adminService.isExchangeWithProviderOkayForComponent(consumerLocation.getComponentName(),
                            providerEndpoint, exchange);
                } catch (final JBIException e) {
                    // there is this exception if the component does not exists on the container, so we simply return
                    // false
                    test = false;
                    this.log.error("Problem in isExchangeWithProviderOkayForComponent, returning " + test, e);
                }
            } else {
                // Retrieve a connection to the JMX server of the container were the
                // component is deployed
                final String containerName = consumerLocation.getContainerName();
                final JMXClient server = this.jmxService.getJMXClient(containerName);
                try {
                    // Retrieve the endpoint description
                    test = server.getAdminServiceClient().isExchangeWithProviderOkayForComponent(
                            consumerLocation.getComponentName(), providerEndpoint, exchange);
                } finally {
                    this.jmxService.returnJMXClient(containerName, server);
                }
            }
        } catch (final Exception e) {
            this.log.error("Problem in isExchangeWithProviderOkayForComponent", e);
        }
        this.log.end();
        return test;
    }

}
