/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.messaging.routing;

import static javax.jbi.messaging.MessageExchange.Role.CONSUMER;
import static javax.jbi.messaging.MessageExchange.Role.PROVIDER;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Logger;

import javax.jbi.messaging.MessageExchange.Role;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Cardinality;
import org.objectweb.fractal.fraclet.types.Contingency;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.jbi.component.PetalsComponentContext;
import org.ow2.petals.microkernel.api.jbi.messaging.RouterService;
import org.ow2.petals.microkernel.api.jbi.messaging.exception.RoutingException;
import org.ow2.petals.microkernel.api.jbi.messaging.exception.TrafficStoppedException;
import org.ow2.petals.microkernel.jbi.messaging.routing.module.InstallModule;
import org.ow2.petals.microkernel.jbi.messaging.routing.module.ReceiverModule;
import org.ow2.petals.microkernel.jbi.messaging.routing.module.SenderModule;
import org.ow2.petals.microkernel.transport.TransportException;
import org.ow2.petals.microkernel.transport.TransportListener;
import org.ow2.petals.microkernel.transport.Transporter;
import org.ow2.petals.microkernel.transport.util.TransportSendContext;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * Routes messages to their destinations. 
 * <p>
 * Performs an itinerary resolution,
 * adding for example a transformation service before delivering the message to
 * its recipient, and perform an address resolution, finding the endpoint
 * satisfying the the destination information (finding an endpoint for a
 * specified service). Resolutions are based on defined rules.
 * </p>
 * 
 * @author Adrien Louis - EBM WebSourcing
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Roland Naudin - EBM WebSourcing
 */
@Component(provides = {
 @Interface(name = "service", signature = RouterService.class),
        @Interface(name = "transportlistener", signature = TransportListener.class) })
public class RouterServiceImpl implements RouterService, TransportListener {

    private static final String INSTALLMODULE_FRACTAL_PREFIX = "installmodule-";

    private static final String RECEIVERMODULE_FRACTAL_PREFIX = "receivermodule-";

    private static final String SENDERMODULE_FRACTAL_PREFIX = "sendermodule-";

    private static final String TRANSPORTER_FRACTAL_PREFIX = "transporter-";

    /**
     * The map of the exchanges queues. One queue per installed component.
     */
    private final Map<String, PetalsComponentContext> componentContexts = new ConcurrentHashMap<String, PetalsComponentContext>();

    /**
     * Configuration Service
     */
    @Requires(name = "configuration")
    private ConfigurationService configurationService;

    /**
     * The list of installModule Fractal components
     * 
     * Note: the order of the modules is important, hence the use of a TreeMap with keys that are ordered
     */
    @Requires(name = INSTALLMODULE_FRACTAL_PREFIX, cardinality = Cardinality.COLLECTION, contingency = Contingency.OPTIONAL)
    private final Map<String, InstallModule> installModules = new TreeMap<String, InstallModule>();

    /**
     * Logger wrapper.
     */
    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(RouterService.COMPONENT_LOGGER_NAME));

    /**
     * The list of ReceiverModule Fractal components
     * 
     * Note: the order of the modules is important, hence the use of a TreeMap with keys that are ordered
     * 
     */
    @Requires(name = RECEIVERMODULE_FRACTAL_PREFIX, cardinality = Cardinality.COLLECTION, contingency = Contingency.OPTIONAL)
    private final Map<String, ReceiverModule> receiverModules = new TreeMap<String, ReceiverModule>();

    /**
     * The list of SenderModule Fractal components
     * 
     * Note: the order of the modules is important, hence the use of a TreeMap with keys that are ordered
     * 
     */
    @Requires(name = SENDERMODULE_FRACTAL_PREFIX, cardinality = Cardinality.COLLECTION, contingency = Contingency.OPTIONAL)
    private final Map<String, SenderModule> senderModules = new TreeMap<String, SenderModule>();

    /**
     * Traffic flag to handle stop of traffic
     * 
     * It is volatile to ensure that every thread has its real value
     */
    private volatile boolean trafficStopped = false;

    /**
     * the Transporter Fractal components
     */
    @Requires(name = TRANSPORTER_FRACTAL_PREFIX, cardinality = Cardinality.COLLECTION, contingency = Contingency.OPTIONAL)
    private final Map<String, Transporter> transporters = new HashMap<String, Transporter>();

    @Override
    public void addComponent(final PetalsComponentContext componentContext) throws RoutingException {
        this.log.start();

        for (final InstallModule installModule : this.installModules.values()) {
            installModule.addComponent(componentContext);
        }

        this.componentContexts.put(componentContext.getComponentName(), componentContext);

        this.log.end();
    }

    @Override
    public void exchangeSent(final PetalsMessageExchange exchange) {
        // do nothing
    }

    @Override
    public void onExchange(final PetalsMessageExchange exchange) throws TransportException {
        this.log.start();

        if (this.trafficStopped) {
            throw new TransportException("The traffic is stopped on the router.");
        }

        final String componentName;
        if (exchange.isConsumer()) {
            componentName = exchange.getConsumerLocation().getComponentName();
        } else {
            componentName = exchange.getElectedProviderEndpoint().getLocation().getComponentName();
        }

        final PetalsComponentContext componentContext = this.componentContexts.get(componentName);

        if (componentContext == null) {
            throw new TransportException("Unknown component '" + componentName + "' in router");
        }

        for (final ReceiverModule receiverModule : this.receiverModules.values()) {
            try {
                receiverModule.receiveExchange(exchange, componentContext);
            } catch (final RoutingException e) {
                throw new TransportException("Error while processing received exchange with receiver module", e);
            }
        }

        try {
            componentContext.offerExchange(exchange);
        } catch (final RoutingException e) {
            throw new TransportException(e);
        }

        this.log.end();
    }

    @Override
    public void removeComponent(final PetalsComponentContext componentContext) throws RoutingException {
        this.log.start();

        for (final InstallModule installModule : this.installModules.values()) {
            installModule.removeComponent(componentContext);
        }

        this.componentContexts.remove(componentContext.getComponentName());

        this.log.end();
    }

    @Override
    public void sendInNMR(final PetalsMessageExchange exchange) throws RoutingException {
        this.log.call();

        if (this.trafficStopped) {
            throw new TrafficStoppedException();
        }

        // switch the classloader to the container one
        // TODO document why it is needed...
        final ClassLoader componentClassLoader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());
        try {
            realSendInNMR(exchange);
        } finally {
            Thread.currentThread().setContextClassLoader(componentClassLoader);
        }
    }

    private void realSendInNMR(final PetalsMessageExchange exchange) throws RoutingException {

        // Get the list of elected destinations from the sender modules
        // using a LinkedList to keep the insertion order
        final List<TransportSendContext> electedDestinations = new LinkedList<TransportSendContext>();
        for (final SenderModule senderModule : this.senderModules.values()) {
            senderModule.electEndpoints(electedDestinations, exchange);
        }

        final Role originalRole = exchange.getRole();

        try {
            if (exchange.isConsumer()) {
                exchange.setRole(PROVIDER);
            } else {
                exchange.setRole(CONSUMER);
                // only one destination for a consumer
                this.log.assertOrLog(electedDestinations.size() == 1,
                        "There shouldn't be multiple elected destination if an exchange is sent back to a Consumer");
            }

            this.sendToTransporters(electedDestinations, exchange);

        } catch (final RoutingException e) {
            // an exception means the message wasn't routed where we wanted
            // reset the role
            exchange.setRole(originalRole);
            throw e;
        }
    }

    @Override
    public void stopTraffic() {
        this.log.start();

        // TODO could we maybe reinject all these exchanges to another node when it makes sense instead of waiting for
        // them to terminate? or for those that couldn't terminate
        if (!this.waitEndOfPendingExchange(this.configurationService.getContainerConfiguration()
                .getRouterStopTrafficDelay())) {
            this.log.warning("The timeout for stopping the Router traffic has been reached."
                    + " Some message have not beed accepted by the component, they are lost.");
        }

        this.trafficStopped = true;

        this.log.end();
    }

    @Override
    public boolean waitEndOfPendingExchange(final long waitingTime) {
        // Wait that all the pending exchanges are received or the timeout is
        // reached
        final long pauseTime = waitingTime / Math.max(1, this.componentContexts.size());
        final long timeout = waitingTime + System.currentTimeMillis();
        boolean done = false;
        try {
            while (System.currentTimeMillis() < timeout && !done) {
                for (final PetalsComponentContext context : this.componentContexts.values()) {
                    if (context.hasExchangesInDeliveryChannel()) {
                        // sleep a bit
                        Thread.sleep(pauseTime);
                        break;
                    }
                }
                done = true;
            }
        } catch (final InterruptedException e) {
            // do nothing
        }

        return done;
    }
    
    /**
     * Sends the specified message to the provider of the service.
     */
    private void sendToTransporters(final List<TransportSendContext> electedDestinations,
            final PetalsMessageExchange exchange) throws RoutingException {
        this.log.start();

        boolean retry = true;
        int attemptDelay = 0;

        // TODO improve that code, it is terribly hard to read
        while (retry) {
            if (attemptDelay != 0) {
                if (this.log.isDebugEnabled()) {
                    this.log.debug("Wait " + attemptDelay + " millisecond before the next send attempt");
                }
                try {
                    Thread.sleep(attemptDelay);
                } catch (InterruptedException e) {
                    // we got interrupted, this could happen when component's threads are stopped, so we should stop
                    // here and return. Since we haven't succeeded in sending the message, we should return an error.
                    throw new RoutingException("Send was aborted by interrupt, the exchange wasn't sent.");
                }
            }
            retry = false;
            final Iterator<TransportSendContext> iterator = electedDestinations.iterator();

            while (iterator.hasNext()) {
                final TransportSendContext context = iterator.next();
                if (context.attempt > 0) {
                    if (this.log.isDebugEnabled()) {
                        this.log.debug("Send attempt to '" + context + "'");
                    }

                    // if we are sending back to the consumer, then this shouldn't be touched
                    if (exchange.isProvider()) {
                        assert context.getEndpoint() != null;
                        exchange.setElectedProviderEndpoint(context.getEndpoint());
                    }

                    try {
                        final Transporter transporter = this.transporters.get(context.transport);

                        transporter.send(exchange, context);

                        this.log.end("Transporter's send succeeded");

                        // if it worked let's return
                        return;
                    } catch (final TransportException e) {
                        context.attempt -= 1;
                        if (context.attempt > 0) {
                            retry = true;
                            // get the max delay
                            if (context.delay > attemptDelay) {
                                attemptDelay = context.delay;
                            }
                        }

                        if (iterator.hasNext() || retry) {
                            this.log.warning("The send attempt to '" + context + "' failed", e);
                        } else {
                            throw new RoutingException(e);
                        }
                    }
                }
            }
        }
    }

    /**
     * start the router.
     * 
     * @throws RoutingException
     *             impossible to start the router
     */
    @Lifecycle(step = Step.START)
    public void start() {
        this.log.start();
        this.log.debug("The message router is started.");
        this.log.end();
    }

    /**
     * stop the router.
     */
    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.start();
        this.log.debug("The message router is stopped.");
        this.log.end();
    }
}