/**
 * Copyright (c) 2008-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.messaging.routing.module;

import java.io.File;
import java.net.URISyntaxException;
import java.net.URL;
import java.security.Principal;
import java.util.Set;
import java.util.logging.Logger;

import javax.security.auth.Subject;
import javax.security.auth.login.LoginContext;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.jaas.JAASCredentialCallbackHandler;
import org.ow2.petals.microkernel.api.jbi.component.PetalsComponentContext;
import org.ow2.petals.microkernel.api.jbi.messaging.exception.RoutingException;
import org.ow2.petals.microkernel.api.jbi.security.SecurityContext;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * A router module used to control access of add/remove component methods with JAAS. 
 * <p>
 * Not really useful, we can imagine to do such process when we will
 * implement a addConnection mechanism in the transport layer.
 * </p>
 * 
 * @author Christophe HAMERLING - EBM WebSourcing
 */
@Component(provides = @Interface(name = "service", signature = InstallModule.class))
public class JAASAuthenticationModule implements InstallModule {

    private static final String jaasConfiguration = "petals-domain";

    private static final String LOGIN_PROPS = "login.properties";

    /**
     * Logger wrapper
     */
    protected final LoggingUtil log = new LoggingUtil(Logger.getLogger(JAASAuthenticationModule.class.getName()));

    static class JaasSecurityContext extends SecurityContext {

        private final Subject subject;

        public JaasSecurityContext(String userName, Subject subject) {
            super(userName);
            this.subject = subject;
        }

        @Override
        public Set<Principal> getPrincipals() {
            return this.subject.getPrincipals();
        }
    }

    /**
     * Start the Fractal component
     */
    @Lifecycle(step = Step.START)
    public void start() throws URISyntaxException {
        this.log.start();

        this.initialiseJAAS();

        this.log.end();
    }

    /**
     * Stop the Fractal component
     */
    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.call();
    }

    @Override
    public void addComponent(final PetalsComponentContext componentContext) throws RoutingException {

        // TODO : get username and password from somewhere...
        String username = "petals";
        String password = "petals";

        ClassLoader original = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(
                JAASAuthenticationModule.class.getClassLoader());
        try {
            try {
                JAASCredentialCallbackHandler callback = new JAASCredentialCallbackHandler(
                        username, password);
                LoginContext lc = new LoginContext(jaasConfiguration, callback);
                lc.login();
                Subject subject = lc.getSubject();

                SecurityContext s = new JaasSecurityContext(username, subject);
                componentContext.setSecurityContext(s);
            } catch (Exception e) {
                throw (SecurityException) new SecurityException("User name or password is invalid")
                        .initCause(e);
            }
        } finally {
            Thread.currentThread().setContextClassLoader(original);
        }
    }

    @Override
    public void removeComponent(final PetalsComponentContext componentContext)
            throws RoutingException {
        componentContext.setSecurityContext(null);
    }

    /**
     * Initialize JAAS system properties
     */
    private void initialiseJAAS() throws URISyntaxException {
        String path = System.getProperty("java.security.auth.login.config");
        if (path == null) {
            // try to get the default resource
            URL resource = this.getClass().getClassLoader().getResource("/" + LOGIN_PROPS);
            if (resource != null) {
                path = new File(resource.toURI()).getAbsolutePath();
                System.setProperty("java.security.auth.login.config", path);
            }
        }
    }
}
