/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.messaging.routing.module.endpoint;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.List;
import java.util.Random;
import java.util.SortedSet;
import java.util.TreeSet;

import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint.EndpointType;
import org.ow2.petals.microkernel.api.communication.topology.TopologyService;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.jbi.messaging.exception.RoutingException;
import org.ow2.petals.microkernel.jbi.messaging.routing.module.EndpointResolverModule;
import org.ow2.petals.microkernel.jbi.messaging.routing.module.endpoint.WeightedEndpoint.Localization;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * The endpoint chooser strategy.
 * @author Nicolas Salatge - EBM WebSourcing
 */
public final class EndpointOrderer {

    /**
     * logger wrapper.
     */
    private LoggingUtil log;

    private ContainerConfiguration containerConfiguration;

    private TopologyService topologyService;

    /**
     * This random is used to configure the load balancing strategy.
     */
    private Random random;

    /**
     * Default constructor.
     * 
     */
    public EndpointOrderer(final LoggingUtil log,
            final ContainerConfiguration containerConfiguration,
            final TopologyService topologyService) {
        super();

        this.log = log;
        this.containerConfiguration = containerConfiguration;
        this.topologyService = topologyService;
        this.random = new Random();
    }

    /**
     * Elects the endpoints list of endpoints as the message recipient, with the
     * defined strategy. Ask the consumer and the provider if they are Okay for
     * the exchange.
     * 
     * @param network
     *            the container network
     * @param endpoints
     *            the list of endpoints to select an endpoint from
     * @param exchange
     *            the exchange
     * @param strategyParameters
     *            the list of parameters of the strategy. Must be non-null
     * 
     * @return the service endpoint selected or null
     * @throws RoutingException
     *             impossible to realize this strategy
     */
    public final Collection<PetalsServiceEndpoint> orderEndpoints(final Collection<PetalsServiceEndpoint> endpoints,
            final List<Object> strategyParameters) throws RoutingException {

        final String type;
        final float localWeighting;
        final float remoteActiveWeighting;
        final float remoteInactiveWeighting;
        try {
            final Object[] parameters = strategyParameters.toArray();
            type = ((String) parameters[0]);
            localWeighting = ((Float) parameters[1]).floatValue();
            remoteActiveWeighting = ((Float) parameters[2]).floatValue();
            remoteInactiveWeighting = ((Float) parameters[3]).floatValue();

        } catch (final ClassCastException e) {
            throw new RoutingException("Impossible to convert the "
                    + "parameters to realize the routing strategy");
        }

        final Collection<PetalsServiceEndpoint> orderedEndpoints;
        if (type.equals(EndpointResolverModule.HIGHEST)) {
            orderedEndpoints = this.loadHighestChoose(endpoints, localWeighting,
                    remoteActiveWeighting, remoteInactiveWeighting);
        } else {
            orderedEndpoints = this.loadRandomChoose(endpoints, localWeighting,
                    remoteActiveWeighting, remoteInactiveWeighting);
        }

        return orderedEndpoints;
    }

    /**
     * The random strategy. This strategy order endpoints in function of defined
     * weightings. The endpoints that have the strongest weighting can be more
     * easily choose in comparison with the others.
     * 
     * @param network
     *            the container network. Must be non-null
     * @param endpoints
     *            the list of valid endpoints. May be null
     * @param localPonderation
     *            the local weighting
     * @param remoteActivePonderation
     *            the remote active weighting
     * @param remoteInactivePonderation
     *            the remote inactive weighting
     * @param exchange
     *            the exchange. Must be non-null
     * @param r
     *            a random parameter. Must be non-null.
     * @return the service endpoint selected
     */
    private final Collection<PetalsServiceEndpoint> loadRandomChoose(final Collection<PetalsServiceEndpoint> endpoints,
            final float localWeight, final float remoteActiveWeight,
            final float remoteInactiveWeight) {

        // We create a sorted list with its comparator defining the algorithm
        // "Random"
        final SortedSet<WeightedEndpoint> orderedEndpoints = new TreeSet<WeightedEndpoint>(
                new Comparator<WeightedEndpoint>() {

                    public int compare(final WeightedEndpoint endpointOne,
                            final WeightedEndpoint endpointTwo) {

                        // Only the endpoint weight is take into account to
                        // order endpoints
                        return endpointOne.weight.compareTo(endpointTwo.weight);
                    }

                });

        // Finally, we add all endpoints in the list to get sorted endpoints
        this.log.debug("Apply the algorithm RANDOM");
        return this.fillSortedList(endpoints, localWeight, remoteActiveWeight,
                remoteInactiveWeight, orderedEndpoints);
    }

    /**
     * The highest strategy. This strategy creates a ordered endpoint list
     * according to the weight of their container type (local, remote active or
     * remote inactive). If two endpoints are hosted on two containers that have
     * the same type, the container weight is multiply by a randomized value.
     * 
     * @param network
     *            the container network. Must be non-null
     * @param endpoints
     *            the list of valid endpoints. May be null
     * @param localWeighting
     *            the local weighting
     * @param remoteActiveWeighting
     *            the remote active weighting
     * @param remoteInactiveWeighting
     *            the remote inactive weighting
     * @param exchange
     *            the exchange. Must be non-null
     * @return the ordered selected service endpoints according to the algorithm
     *         "Highest"
     */
    private final Collection<PetalsServiceEndpoint> loadHighestChoose(final Collection<PetalsServiceEndpoint> endpoints,
            final float localWeight, final float remoteActiveWeight,
            final float remoteInactiveWeight) {

        // We create an array containing weighted localization. Array index is
        // the localization (the ordinal numbers associated to the enumeration
        // constants defined in Localization)
        final Float[] weightedLocalization = { localWeight, remoteActiveWeight,
                remoteInactiveWeight };

        // We create a sorted list with its comparator defining the algorithm
        // "Highest": The first element of the list has the highest weight in
        // the highest localization.
        final SortedSet<WeightedEndpoint> orderedEndpoints = new TreeSet<WeightedEndpoint>(
                new Comparator<WeightedEndpoint>() {

                    public int compare(final WeightedEndpoint endpointOne,
                            final WeightedEndpoint endpointTwo) {

                        // The localization is the first sort criteria. If they
                        // are equal, the weight is used.
                        // Caution: the first elements of the list must be the
                        // endpoints
                        // associated to the localization with the highest
                        // weight.
                        int result = weightedLocalization[endpointTwo.localization.ordinal()]
                                .compareTo(weightedLocalization[endpointOne.localization.ordinal()]);
                        if (result != 0) {
                            return result;
                        } else {
                            return endpointOne.weight.compareTo(endpointTwo.weight);
                        }
                    }

                });

        // Finally, we add all endpoints in the list to get sorted endpoints
        this.log.debug("Apply the algorithm HIGHEST");
        return this.fillSortedList(endpoints, localWeight, remoteActiveWeight,
                remoteInactiveWeight, orderedEndpoints);

    }

    /**
     * Fill the ordered endpoint list with the weighted endpoints defined from
     * <code>endpoints</code> from which the weight is computed according to the
     * container type weight multiply by a random value.
     * 
     * @param network
     *            the container network. Must be non-null
     * @param endpoints
     *            the list of valid endpoints. May be null
     * @param localWeighting
     *            the local weighting
     * @param remoteActiveWeighting
     *            the remote active weighting
     * @param remoteInactiveWeighting
     *            the remote inactive weighting
     * @param r
     *            a randomized value
     * @param orderedEndpoint
     *            the ordered list that will be fill by this method with
     *            <code>endpoints</code>
     */
    private Collection<PetalsServiceEndpoint> fillSortedList(final Collection<PetalsServiceEndpoint> endpoints,
            final float localWeight, final float remoteActiveWeight,
            final float remoteInactiveWeight, final SortedSet<WeightedEndpoint> tempOrderedEndpoints) {
        final List<PetalsServiceEndpoint> orderedEndpoints = new ArrayList<PetalsServiceEndpoint>();

        for (PetalsServiceEndpoint providerEndpoint : endpoints) {
            assert providerEndpoint.getType() == EndpointType.INTERNAL;
            final String endpointContainerName = providerEndpoint.getLocation().getContainerName();
            if (endpointContainerName.equals(this.containerConfiguration.getName())) {
                tempOrderedEndpoints.add(new WeightedEndpoint(providerEndpoint,
                        localWeight * this.random.nextFloat(), Localization.LOCAL));
            } else if (this.topologyService.isContainerReachable(endpointContainerName)) {
                tempOrderedEndpoints.add(new WeightedEndpoint(providerEndpoint,
                        remoteActiveWeight * this.random.nextFloat(), Localization.REMOTE_ACTIVE));
            } else {
                tempOrderedEndpoints.add(new WeightedEndpoint(providerEndpoint,
                        remoteInactiveWeight * this.random.nextFloat(), Localization.REMOTE_INACTIVE));
            }
        }

        for (final WeightedEndpoint weightedEndpoint : tempOrderedEndpoints) {
            orderedEndpoints.add(weightedEndpoint.endpoint);
        }

        return orderedEndpoints;
    }
}
