/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.server;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Set;

import javax.management.Descriptor;
import javax.management.InstanceNotFoundException;
import javax.management.JMException;
import javax.management.MBeanException;
import javax.management.MBeanServer;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.management.RuntimeOperationsException;
import javax.management.StandardMBean;
import javax.management.modelmbean.DescriptorSupport;
import javax.management.modelmbean.InvalidTargetObjectTypeException;
import javax.management.modelmbean.ModelMBeanAttributeInfo;
import javax.management.modelmbean.ModelMBeanConstructorInfo;
import javax.management.modelmbean.ModelMBeanInfo;
import javax.management.modelmbean.ModelMBeanInfoSupport;
import javax.management.modelmbean.ModelMBeanNotificationInfo;
import javax.management.modelmbean.ModelMBeanOperationInfo;
import javax.management.modelmbean.RequiredModelMBean;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.jmx.agent.Introspector;
import org.objectweb.fractal.util.Fractal;
import org.ow2.petals.jmx.api.api.AdminServiceClient;
import org.ow2.petals.jmx.api.api.DeploymentServiceClient;
import org.ow2.petals.jmx.api.api.EndpointDirectoryClient;
import org.ow2.petals.jmx.api.api.InstallationServiceClient;
import org.ow2.petals.jmx.api.api.LoggerServiceClient;
import org.ow2.petals.jmx.api.api.PetalsAdminServiceClient;
import org.ow2.petals.jmx.api.api.TopologyServiceClient;
import org.ow2.petals.jmx.api.api.monitoring.container.transporter.local.LocalTransporterMonitoringServiceClient;
import org.ow2.petals.jmx.api.api.monitoring.container.transporter.remote.TcpTransporterMonitoringServiceClient;
import org.ow2.petals.microkernel.api.admin.PetalsAdminServiceMBean;
import org.ow2.petals.microkernel.api.communication.JMXService;
import org.ow2.petals.microkernel.api.communication.topology.TopologyServiceMBean;
import org.ow2.petals.microkernel.api.jbi.management.AdminServiceMBean;
import org.ow2.petals.microkernel.api.jbi.management.DeploymentServiceMBean;
import org.ow2.petals.microkernel.api.jbi.management.InstallationServiceMBean;
import org.ow2.petals.microkernel.api.jbi.messaging.endpoint.EndpointDirectoryMBean;
import org.ow2.petals.microkernel.api.jbi.messaging.endpoint.EndpointDirectoryService;
import org.ow2.petals.microkernel.api.server.FractalHelper;
import org.ow2.petals.microkernel.server.exception.InvalidCompBasedArchiException;
import org.ow2.petals.microkernel.system.logging.LoggingServiceMBean;
import org.ow2.petals.microkernel.transport.local.monitoring.LocalTransporterMonitoringMBean;
import org.ow2.petals.microkernel.transport.platform.nio.monitoring.TcpTransporterMonitoringMBean;

/**
 * This class helps the Launcher to register MBean.
 * @author Adrien Louis - EBM WebSourcing
 */
public class MBeanHelper {

    public static final String CURRENCY_TIME_LIMIT = "-1"; // !!!

    public static final String DOMAIN = "Petals";

    /**
     * Find the local JMX server in the petals component
     * 
     * @param petalsComposite
     *            the petals component
     * @return the local jmx server
     * @throws NoSuchInterfaceException
     * @throws ADLException
     */
    public static final MBeanServer findLocalJMXServer(Component petalsComposite) throws NoSuchInterfaceException {
        ContentController petalsContentController = Fractal.getContentController(petalsComposite);
        Component jmxService = FractalHelper.getRecursiveComponentByName(petalsContentController,
                FractalHelper.JMX_COMPONENT);

        MBeanServer jmxServer = ((JMXService) jmxService.getFcInterface("service"))
                .getLocalJMXServer();

        return jmxServer;
    }

    public static final void registerMBeans(Component petalsComposite) throws Exception {
        ContentController contentController = Fractal.getContentController(petalsComposite);

        // unactivate the method invokation cache for the Fractal MBeans
        Introspector.CURRENCY_TIME_LIMIT = "-1";

        MBeanServer jmxServer = findLocalJMXServer(petalsComposite);

        // register AdminService
        registerComponent(contentController, FractalHelper.ADMIN_COMPONENT, "service", jmxServer,
                AdminServiceMBean.class,
                DOMAIN + ":name=" + AdminServiceClient.ADMIN_SERVICE_MBEAN_NAME + ",type=service", true);

        // register DeploymentService
        registerComponent(contentController, FractalHelper.DEPLOYMENT_COMPONENT, "service", jmxServer,
                DeploymentServiceMBean.class,
                DOMAIN + ":name=" + DeploymentServiceClient.DEPLOYMENT_SERVICE_MBEAN_NAME + ",type=service", true);

        // register InstallationService
        registerComponent(contentController, FractalHelper.INSTALLATION_COMPONENT, "service", jmxServer,
                InstallationServiceMBean.class,
                DOMAIN + ":name=" + InstallationServiceClient.INSTALLATION_SERVICE_MBEAN_NAME + ",type=service", true);

        // register EndpointDirectory
        registerComponent(contentController, FractalHelper.ENDPOINT_DIRECTORY_COMPONENT,
                EndpointDirectoryService.FRACTAL_SRV_ITF_NAME, jmxServer, EndpointDirectoryMBean.class,
                DOMAIN + ":name=" + EndpointDirectoryClient.ENDPOINT_DIRECTORY_MBEAN_NAME + ",type=service", true);

        // register Petals Admin
        registerComponent(contentController, FractalHelper.PETALSADMIN_COMPONENT, "service", jmxServer,
                PetalsAdminServiceMBean.class,
                DOMAIN + ":name=" + PetalsAdminServiceClient.PETALS_ADMIN_MBEAN_NAME + ",type=service", true);

        // register Petals TransporterLocalMonitoring
        registerComponent(contentController, FractalHelper.LOCAL_TRANSPORTER_MONITORING_COMPONENT, "service", jmxServer,
                LocalTransporterMonitoringMBean.class,
                DOMAIN + ":name=" + LocalTransporterMonitoringServiceClient.LOCAL_TRANSPORTER_MONITORING_MBEAN_NAME
                        + ",type=service",
                false);

        // register Petals TransporterNIOMonitoring
        registerComponent(contentController, FractalHelper.TCP_TRANSPORTER_MONITORING_COMPONENT, "service", jmxServer,
                TcpTransporterMonitoringMBean.class, DOMAIN + ":name="
                        + TcpTransporterMonitoringServiceClient.TCP_TRANSPORTER_MONITORING_MBEAN_NAME + ",type=service",
                false);

        // register Logger service
        registerComponent(contentController, FractalHelper.LOGGER_COMPONENT, "service", jmxServer,
                LoggingServiceMBean.class,
                DOMAIN + ":name=" + LoggerServiceClient.LOGGER_SERVICE_MBEAN_NAME + ",type=service", true);

        // register Topology service
        registerComponent(contentController, FractalHelper.TOPOLOGY_COMPONENT, "service", jmxServer,
                TopologyServiceMBean.class,
                DOMAIN + ":name=" + TopologyServiceClient.TOPOLOGY_SERVICE_MBEAN_NAME + ",type=service", true);
    }

    /**
     * Return the MBean name from the jmx server with the specified pattern
     * 
     * @param serviceName
     *            The name of the service to retrieve, must be {@code null}
     * @param jmxServer
     *            The JMX server instance, must be non {@code null}
     * 
     * @return the objectname, null if no object match
     * @throws MalformedObjectNameException
     */
    public static final ObjectName retrieveServiceMBean(String serviceName, MBeanServer jmxServer)
            throws MalformedObjectNameException {

        ObjectName result = null;

        // crate the search filter
        final Hashtable<String, String> attributes = new Hashtable<String, String>();
        attributes.put("name", serviceName);
        attributes.put("type", "service");
        final ObjectName objName = new ObjectName("Petals", attributes);

        final Set<ObjectName> objNames = jmxServer.queryNames(objName, null);
        if ((objNames != null) && (objNames.size() == 1)) {
            result = objNames.iterator().next();
        }
        return result;
    }

    /**
     * Register a fractal component-service to the jmx server.
     * 
     * @param <T>
     * 
     * @param contentController
     *            contentcontroller where the fractal comp can be found
     * @param fractalName
     *            the name of the fractal comp
     * @param fractalItf
     *            the name of the fractal interface
     * @param jmxServer
     *            the jmx server to register to
     * @param mbeanClass
     *            the interface of the mbean
     * @param mbeanName
     *            the name of the mbean
     * @param createAsCustomMBean
     *            If <code>true</code> the MBean will be registered as custom MBean using
     *            {@link #createCustomMBean(Object, Class)}. If <code>false</code> the MBean will be registered as
     *            {@link StandardMBean}.
     * @throws Exception
     */
    public static final <T> void registerComponent(final ContentController contentController, final String fractalName,
            final String fractalItf, final MBeanServer jmxServer, final Class<T> mbeanClass, final String mbeanName,
            final boolean createAsCustomMBean)
            throws Exception {
        Component fractalComponent = FractalHelper.getRecursiveComponentByName(contentController, fractalName);
        // TODO for now we have to directly get the implementation instead of the component interface
        // this is completely WRONG and breaks the component-orientation of the architecture!
        // We should investigate if we really need to do this createCustomMBean and if a simpler
        // registerSimplyComponent is enough
        if (!mbeanClass.isAssignableFrom(fractalComponent.getFcInterface(fractalItf).getClass())) {
            throw new InvalidCompBasedArchiException("The " + fractalItf
                    + " fractal interface does not implement the MBean java interface " + mbeanClass.getName());
        }
        T object = (T) fractalComponent.getFcInterface("/content");

        if (createAsCustomMBean) {
            jmxServer.registerMBean(createCustomMBean(object, mbeanClass), new ObjectName(mbeanName));
        } else {
            jmxServer.registerMBean(new StandardMBean(object, mbeanClass), new ObjectName(mbeanName));
        }
    }

    /**
     * Create a custom MBean which process the method isXXX, setXXX, or getXXX
     * as getter only if an attribute is define in the class.
     * 
     * @param object
     * @param mbeanClass
     * @return
     * @throws RuntimeOperationsException
     * @throws MBeanException
     * @throws InstanceNotFoundException
     * @throws InvalidTargetObjectTypeException
     */
    public static final RequiredModelMBean createCustomMBean(Object object, Class<?> mbeanClass)
            throws RuntimeOperationsException, MBeanException, InstanceNotFoundException,
            InvalidTargetObjectTypeException {
        Method[] methods = mbeanClass.getMethods();
        List<ModelMBeanAttributeInfo> attributes = new ArrayList<ModelMBeanAttributeInfo>();
        List<ModelMBeanOperationInfo> operators = new ArrayList<ModelMBeanOperationInfo>();

        for (final Method method : methods) {
            // IF THE METHOD STARTS WITH "is," "get," or "set" THEN TRY TO ADD
            // AN ATTRIBUTE TO THE MODELMBEAN
            final String methodName = method.getName();
            if ((methodName.startsWith("get")) || (methodName.startsWith("set")) || (methodName.startsWith("is"))) {
                String fieldName;
                if (methodName.startsWith("is")) {
                    fieldName = methodName.substring(2);
                } else {
                    fieldName = methodName.substring(3);
                }
                fieldName = fieldName.substring(0, 1).toLowerCase().concat(fieldName.substring(1));
                try {
                    try {
                        object.getClass().getDeclaredField(fieldName);
                        ModelMBeanAttributeInfo mmai = doAttribute(methods, method);
                        if (mmai != null) {
                            attributes.add(mmai);
                        }
                    } catch (NoSuchFieldException e) {
                        // ADD THE METHOD AS AN OPERATION TO THE MODEL MBEAN
                        ModelMBeanOperationInfo mmoi = doOperation(methods, method, true);
                        if (mmoi != null) {
                            operators.add(mmoi);
                        }
                    }
                } catch (Exception e) {
                    // not possible
                }
            } else {
                // ADD THE METHOD AS AN OPERATION TO THE MODEL MBEAN
                try {
                    ModelMBeanOperationInfo mmoi = doOperation(methods, method, false);
                    if (mmoi != null) {
                        operators.add(mmoi);
                    }
                } catch (Exception e) {
                    // not possible
                }
            }
        }

        // ATTRIBUTES
        ModelMBeanAttributeInfo[] attributeArray = new ModelMBeanAttributeInfo[attributes.size()];
        attributes.toArray(attributeArray);

        // METHODS
        ModelMBeanOperationInfo[] operatorArray = new ModelMBeanOperationInfo[operators.size()];
        operators.toArray(operatorArray);

        // Descriptor
        Descriptor desc = new DescriptorSupport();
        desc.setField("name", object.getClass().getName());
        desc.setField("descriptorType", "MBean");

        // CREATE MODELMBEAN
        ModelMBeanInfo mminfo = new ModelMBeanInfoSupport(object.getClass().getName(), object
                .getClass().getName(), attributeArray, new ModelMBeanConstructorInfo[0],
                operatorArray, new ModelMBeanNotificationInfo[0], desc);
        RequiredModelMBean model = new RequiredModelMBean(mminfo);
        model.setManagedResource(object, "ObjectReference");
        return model;
    }

    /**
     * Build the ModelMBeanAttributeInfo for the method. Take the array of the objects methods and the index of the
     * method to work on. It needs the entire array of methods to determine matching sets of setters and getters.
     * 
     * @param methods
     *            the array of the objects methods to work on.
     * @param method
     *            the method to work on.
     * @return the ModelMBeanOperationInfo for the method.
     * @throws JMException
     *             if something goes wrong.
     */
    private static ModelMBeanAttributeInfo doAttribute(final Method methods[], final Method method) throws JMException {
        final String name;
        final Method getter;
        final Method setter;
        final String type;

        // THIS METHOD IS A GETTER
        final String methodName = method.getName();
        if ((methodName.startsWith("get")) && (method.getParameterTypes().length == 0)) {
            final String setName = "s" + methodName.substring(1);
            name = methodName.substring(3);
            getter = method;
            type = method.getReturnType().getName();
            // CHECK TO SEE IF THERE IS THE EQUVALENT WRITE OBJECT
            Method found = null;
            for (int i = 0; i < methods.length; i++) {
                if (methods[i].getName().equals(setName)) {
                    found = methods[i];
                    break;
                }
            }
            setter = found;

        } // THIS OBJECT IS A SETTER
        else if ((methodName.startsWith("set")) && (method.getParameterTypes().length == 1)) {
            final String getName = "g" + methodName.substring(1);
            name = methodName.substring(3);
            // CHECK TO SEE IF THERE IS THE EQUVALENT READ OBJECT
            // IF SO, THEN THE GET EQUIVALENT WOULD HAVE RETURNED
            // A ModelMBeanAttributeInfo, SO RETURN A NULL
            // SO THIS ATTRIBUTE DOESN'T SHOW UP TWICE
            for (int i_ = 0; i_ < methods.length; i_++) {
                if (methods[i_].getName().equals(getName)) {
                    return null;
                }
            }
            setter = method;
            getter = null;
            type = method.getParameterTypes()[0].getName();

        } // THIS OBJECT IS A GETTER WITH NO SETTER
        else if ((methodName.startsWith("is")) && (method.getParameterTypes() == null)) {
            name = methodName.substring(2);
            getter = method;
            setter = null;
            type = boolean.class.getName();
        } // NOT A COMPLIENT get, set, OR is ATTRIBUTE
        else {
            return null;
        }

        final Descriptor desc = new DescriptorSupport();
        desc.setField("name", name); // THE NAME THAT THE ATTRIBUTE WILL HAVE
        desc.setField("descriptorType", "attribute"); // THIS IS MANDATORY FOR
        // ANY ATTRIBUTES
        // DESCRIPTION
        desc.setField("displayName", name);
        // MAKE SURE THAT THE AGENT DOESN'T READ FROM CACHE (0 does not work
        // with JMX RI)
        desc.setField("currencyTimeLimit", CURRENCY_TIME_LIMIT);

        if (getter != null) {
            desc.setField("getMethod", getter.getName());
        }
        if (setter != null) {
            desc.setField("setMethod", setter.getName());
        }

        return new ModelMBeanAttributeInfo(name, type, name, (getter != null), (setter != null), false, desc);
    }

    /*
     * Build the ModelMBeanOperationInfo for the method. Take the array of the
     * objects methods and the index of the method to work on.
     * 
     * @param methods the array of the objects methods to work on.
     * 
     * @param i the index of the method to work on.
     * 
     * @return the ModelMBeanOperationInfo for the method.
     * 
     * @throws JMException if something goes wrong.
     */
    private static ModelMBeanOperationInfo doOperation(final Method methods[], final Method method,
            boolean isGetterOrSetter) throws JMException {
        final Descriptor desc = new DescriptorSupport();
        final String methodName = method.getName();
        desc.setField("name", methodName);
        desc.setField("descriptorType", "operation");
        // MAKE SURE THAT THE AGENT DOESN'T READ FROM CACHE (0 does not work
        // with JMX RI)
        // desc.setField("currencyTimeLimit", "1"); //!!!
        desc.setField("currencyTimeLimit", CURRENCY_TIME_LIMIT);

        // IF THE METHOD STARTS WITH "get" or "is" TREAT IT AS A GETTER.
        // IF IT STARTS WITH A "set" TREAT IT AS A SETTER.
        if (isGetterOrSetter) {
            if ((methodName.startsWith("get") || methodName.startsWith("is"))
                    && (method.getParameterTypes().length == 0)) {
                desc.setField("role", "getter");
            } else if (methodName.startsWith("set")
                    && (method.getParameterTypes().length == 1)) {
                desc.setField("role", "setter");
            }
        } else {
            desc.setField("role", "operation");
        }        

        return new ModelMBeanOperationInfo(methodName, method, desc);
    }

}
