/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.server;

import java.io.File;
import java.net.URL;
import java.net.URLStreamHandler;
import java.net.URLStreamHandlerFactory;
import java.util.Collections;
import java.util.Properties;
import java.util.logging.Logger;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.control.NameController;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.util.Fractal;
import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.launcher.api.server.PetalsServer;
import org.ow2.petals.launcher.api.server.conf.ConfigurationProperties;
import org.ow2.petals.launcher.api.server.exception.PetalsServerException;
import org.ow2.petals.microkernel.api.admin.PetalsAdminInterface;
import org.ow2.petals.microkernel.api.communication.sharedarea.SharedAreaService;
import org.ow2.petals.microkernel.api.communication.sharedarea.SharedAreaService.SavedEndpoint;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration.ContainerState;
import org.ow2.petals.microkernel.api.jbi.management.ManagementException;
import org.ow2.petals.microkernel.api.jbi.management.recovery.SystemRecoveryService;
import org.ow2.petals.microkernel.api.jbi.messaging.RouterService;
import org.ow2.petals.microkernel.api.server.FractalHelper;
import org.ow2.petals.topology.generated.Topology;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * This class allows to start and stop the Petals Server.
 * @author Roland Naudin - EBM WebSourcing
 */
public class PetalsServerImpl implements PetalsServer {

    /**
     * The local properties of the current container (ie: the content of '
     * <code>server.properties</code>'.
     */
    private Properties serverLocalProperties;

    /**
     * The initial topology definition (ie. topology.xml) in which the current container is a node.
     */
    private Topology topology;

    /**
     * The container configuration
     */
    private ContainerConfiguration containerConfiguration;

    /**
     * Petals composite
     */
    private Component petalsComposite;

    /**
     * Petals Content Controller
     */
    private ContentController petalsContentController;

    /**
     * Flag indicating the state of the Petals container. <code>true</code> if the container is starting, started;
     * <code>false</code> if the container is stopping or stopped.
     */
    private volatile boolean isRunning = false;

    /**
     * Logger wrapper.
     */
    private LoggingUtil log = new LoggingUtil(Logger.getLogger(PetalsServer.COMPONENT_LOGGER_NAME));

    /**
     * Creates a new instance of {@link PetalsServerImpl}
     * 
     * @throws PetalsException
     */
    public PetalsServerImpl() throws PetalsException {
        this(false);
    }

    /**
     * Creates a new instance of {@link PetalsServerImpl}
     * 
     * @throws PetalsException
     */
    public PetalsServerImpl(boolean observer) throws PetalsException {
        // Initialize an execution context for each child threads of the server
        PetalsExecutionContext.clear();
    }

    @Override
    public void init(final Properties serverLocalProperties, final Topology topology,
            final File dataRootPath) throws PetalsServerException {

        // Register the Maven URL handler because the class URL is loaded with the system classloader
        URL.setURLStreamHandlerFactory(new URLStreamHandlerFactory() {

            @Override
            public URLStreamHandler createURLStreamHandler(final String protocol) {
                if ("mvn".equals(protocol)) {
                    return new org.ops4j.pax.url.mvn.Handler();
                }

                return null;
            }
        });

        this.serverLocalProperties = serverLocalProperties;
        this.topology = topology;

        // The data root path MUST be set as a property to be resolved, if
        // needed, in other property values
        this.serverLocalProperties.setProperty(
                ConfigurationProperties.DATA_ROOT_DIRECTORY_PROPERTY_NAME,
                dataRootPath.getAbsolutePath());

        try {
            // initialize Petals Fractal composite
            this.initializePetalsComposite();
        } catch (final Exception exception) {
            // clean up the petals server
            throw new PetalsServerException("Problem while initializing Petals", exception);
        }
    }

    @Override
    public void start() throws PetalsServerException {

        this.isRunning = true;

        try {
            // start PEtALS Fractal composite
            this.startPetalsComposite();

            // register the PEtALS server in the Petals admin.
            this.registerPetalsServer();

            // expose PEtALS MBean object into the JMX server
            MBeanHelper.registerMBeans(this.petalsComposite);

            // TODO maybe factorise all of that in the same place in the adequate component such as {@link
            // TopologyService} or  {@link ConfigurationService}...

            final SharedAreaService sharedAreaService = (SharedAreaService) FractalHelper
                    .getRecursiveComponentByName(this.petalsContentController, FractalHelper.SHARED_AREA_COMPOSITE)
                    .getFcInterface(SharedAreaService.FRACTAL_SRV_ITF_NAME);

            sharedAreaService.joinTopology(false);

            // recover all JBI entities
            this.recoverSystem();

            sharedAreaService.restoreEndpoints(Collections.<SavedEndpoint> emptyList());

            // The Petals server is started. We set the local container state to started. The local container state is
            // in unknown state until all is really started.
            this.containerConfiguration.setState(ContainerState.REACHABLE);
        } catch (Exception exception) {
            // clean up the petals server
            if (this.petalsComposite != null) {
                this.log.info("Problem while starting PEtALS, try to stop PEtALS cleanly...");
                try {
                    FractalHelper.stopComponent(this.petalsComposite, this.log);
                } catch (final Exception e) {
                    this.log.error("Failed to stop PEtALS cleanly.", e);
                }
            }
            throw new PetalsServerException("Failed to start PEtALS", exception);
        }

        Runtime.getRuntime().addShutdownHook(new PetalsStopThread(this, this.log));

        this.log.info("Server STARTED");
    }

    @Override
    public void stop() throws PetalsServerException {

        this.isRunning = false;

        PetalsServerException exception = null;
        try {
            this.stopPetalsComposite();
        } catch (final Exception e) {
            exception = new PetalsServerException(e);
        }

        if (exception == null) {
            this.log.info("Server STOPPED");
        } else {
            this.log.error("ERROR: " + exception.getMessage(), exception);
        }

        synchronized (this) {
            this.notifyAll();
        }
        if (exception != null) {
            throw exception;
        }
    }

    /**
     * Initialize petals composite
     * 
     * @throws PetalsException
     * 
     */
    private void initializePetalsComposite() throws PetalsException {
        try {
            this.petalsComposite = FractalHelper.createCompositeComponent(FractalHelper.PETALS_COMPOSITE);
            this.petalsContentController = Fractal.getContentController(this.petalsComposite);
        } catch (NoSuchInterfaceException e) {
            throw new PetalsException("Error creating PEtALS Fractal Composite", e);
        } catch (InstantiationException e) {
            throw new PetalsException("Error creating PEtALS Fractal Composite", e);
        }
    }

    /**
     * Start the Petals Fractal composite. TODO create a Fractal component to
     * handle the starting and stopping of Fractal composites/components.
     * 
     * </br> <b>NOTE :</b> Container name is set to the PetalsFormatter
     * 
     * @throws NoSuchInterfaceException
     * @throws IllegalLifeCycleException
     * @throws IllegalBindingException
     * @throws PetalsException
     */
    private void startPetalsComposite() throws NoSuchInterfaceException, IllegalLifeCycleException, PetalsException {

        // start first the configuration component
        final Component configurationComponent = FractalHelper.getComponentByName(this.petalsContentController,
                FractalHelper.CONFIGURATION_COMPONENT);
        if (!FractalHelper.startComponent(configurationComponent, this.log)) {
            throw new PetalsException("Failed to start PEtALS Fractal component "
                    + FractalHelper.CONFIGURATION_COMPONENT);
        }

        // launch the configuration loading of the configuration service
        final ConfigurationService configurationService = (ConfigurationService) configurationComponent
                .getFcInterface(ConfigurationService.FRACTAL_SRV_ITF_NAME);
        configurationService.loadConfiguration(this.serverLocalProperties, this.topology);

        // retrieve the configuration from the Configuration Service
        this.containerConfiguration = configurationService.getContainerConfiguration();

        // finally, start the others components
        if (!FractalHelper.startComponent(this.petalsComposite, this.log)) {
            throw new PetalsException("Failed to start PEtALS Fractal components");
        }
    }

    /**
     * Log the component status (started or stopped) of a Fractal component.
     * 
     * @param component
     * @throws NoSuchInterfaceException
     */
    private void logComponentState(final Component component, final String space)
            throws NoSuchInterfaceException {

        final LifeCycleController componentLC = Fractal.getLifeCycleController(component);
        final NameController componentNC = Fractal.getNameController(component);
        this.log.debug(space + "- " + componentNC.getFcName() + " is " + componentLC.getFcState());

        ContentController componentCC = null;
        try {
            componentCC = Fractal.getContentController(component);
        } catch (final NoSuchInterfaceException e) {
            // No content controller (ie. no sub components)
            componentCC = null;
        }

        if (componentCC != null) {
            final Component[] componentSubComponents = componentCC.getFcSubComponents();
            for (final Component componentSubComponent : componentSubComponents) {
                this.logComponentState(componentSubComponent, space + "\t");
            }
        }
    }

    /**
     * Stop the Petals Fractal composite.
     * 
     * @throws Exception
     */
    private void stopPetalsComposite() throws Exception {

        this.log.debug("Fractal component status:");
        this.logComponentState(this.petalsComposite, "");

        final LifeCycleController lifeCycleController = Fractal
                .getLifeCycleController(this.petalsComposite);

        if (LifeCycleController.STARTED.equals(lifeCycleController.getFcState())) {

            // First stop and shutdown all the Fractal components associated to
            // the JBI components and SAs
            final Component petalsCompositeCtrlComponent = FractalHelper.getComponentByName(
                    this.petalsContentController, FractalHelper.PETALS_COMPOSITE_CTRL_COMPONENT);
            final PetalsCompositeController petalsCompositeController = (PetalsCompositeController) petalsCompositeCtrlComponent
                    .getFcInterface(PetalsCompositeController.FRACTAL_SRV_ITF_NAME);
            petalsCompositeController.stopJbiArtefacts();

            // After, prepare the stop of the router
            final Component routerComponent = FractalHelper.getRecursiveComponentByName(
                    this.petalsContentController, FractalHelper.ROUTER_COMPONENT);
            final RouterService router = (RouterService) routerComponent.getFcInterface("service");
            router.stopTraffic();

            // TODO and now we should stop listening for external messages, no? on the NIO transporter...

            // Let's first stop this one as it depends on other components being started
            // and when a fractal component is stopped, it blocks calls to its operations
            final Component topologyComponent = FractalHelper.getRecursiveComponentByName(
                    this.petalsContentController, FractalHelper.TOPOLOGY_COMPONENT);
            FractalHelper.stopComponent(topologyComponent, this.log);

            // and this one because it relies on the embedded shared area to be started if it is used
            final Component endpointDirectoryComponent = FractalHelper.getRecursiveComponentByName(
                    this.petalsContentController, FractalHelper.ENDPOINT_DIRECTORY_COMPONENT);
            FractalHelper.stopComponent(endpointDirectoryComponent, this.log);

            // and also this one because it relies on other components too
            final Component postExtensionManagerImpl = FractalHelper.getRecursiveComponentByName(
                    this.petalsContentController, FractalHelper.POST_EXTENSIONSMANAGER_COMPONENT);
            FractalHelper.stopComponent(postExtensionManagerImpl, this.log);

            // and this one because it needs configurationservice (but is needed by some post extensions)
            final Component sharedAreaComponent = FractalHelper
                    .getRecursiveComponentByName(this.petalsContentController, FractalHelper.SHARED_AREA_COMPOSITE);
            FractalHelper.stopComponent(sharedAreaComponent, this.log);

            // and this one is case of...
            final Component preExtensionManagerImpl = FractalHelper.getRecursiveComponentByName(
                    this.petalsContentController, FractalHelper.PRE_EXTENSIONSMANAGER_COMPONENT);
            FractalHelper.stopComponent(preExtensionManagerImpl, this.log);

            // Finally, stop the others components
            FractalHelper.stopComponent(this.petalsComposite, this.log);
        }
    }

    /**
     * Register to the PetalsAdmin service this Petals server instance
     * 
     * @throws NoSuchInterfaceException
     * @throws PetalsException
     */
    private void registerPetalsServer() throws NoSuchInterfaceException, PetalsException {
        Component petalsAdminComponent = FractalHelper.getRecursiveComponentByName(
                this.petalsContentController, FractalHelper.PETALSADMIN_COMPONENT);
        PetalsAdminInterface petalsAdminService = (PetalsAdminInterface) petalsAdminComponent
                .getFcInterface("service");

        petalsAdminService.setPetalsStopThread(new PetalsStopThread(this, this.log));
    }

    /**
     * Recover the system state. All state elements are recovered (components,
     * share libraries, service assemblies).
     * 
     * @throws NoSuchInterfaceException
     * @throws ManagementException
     * @throws PetalsException
     */
    private void recoverSystem() throws NoSuchInterfaceException, ManagementException, PetalsException {

        final Component systemRecoveryComponent = FractalHelper.getRecursiveComponentByName(
                this.petalsContentController, FractalHelper.SYSTEMRECOVERY_COMPONENT);
        final SystemRecoveryService systemRecoveryService = (SystemRecoveryService) systemRecoveryComponent
                .getFcInterface(SystemRecoveryService.FRACTAL_SRV_ITF_NAME);

        systemRecoveryService.recoverAllEntities();

        // Next we stop it without removing it because it can be reused when moving a container from a domain to
        // another one (TODO actually it's not...)
        final Component petalsCompositeCtrlComponent = FractalHelper.getRecursiveComponentByName(
                this.petalsContentController, FractalHelper.PETALS_COMPOSITE_CTRL_COMPONENT);
        final PetalsCompositeController petalsCompositeCtrl = (PetalsCompositeController) petalsCompositeCtrlComponent
                .getFcInterface(PetalsCompositeController.FRACTAL_SRV_ITF_NAME);
        petalsCompositeCtrl.stopSystemRecoveryService();
    }

    @Override
    public boolean isRunning() {
        return this.isRunning;
    }

    @Override
    public String getImplementationName() {
        return "Petals ESB µKernel";
    }

    @Override
    public String getImplementationVersion() {
        return PetalsServerImpl.class.getPackage().getImplementationVersion();
    }
}
