/**
 * Copyright (c) 2009-2012 Capgemini, 2009-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.transport.platform.nio;

import java.util.logging.Logger;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.jbi.messaging.RouterService;
import org.ow2.petals.microkernel.api.jbi.messaging.exception.RoutingException;
import org.ow2.petals.microkernel.transport.AbstractTransporterImpl;
import org.ow2.petals.microkernel.transport.TransportException;
import org.ow2.petals.microkernel.transport.Transporter;
import org.ow2.petals.microkernel.transport.platform.nio.client.NioClientObjectPool;
import org.ow2.petals.microkernel.transport.platform.nio.monitoring.TcpTransporterMonitoring;
import org.ow2.petals.microkernel.transport.platform.nio.selector.NioClient;
import org.ow2.petals.microkernel.transport.platform.nio.server.NioReceiver;
import org.ow2.petals.microkernel.transport.util.TransportSendContext;
import org.ow2.petals.probes.api.enums.ExecutionStatus;
import org.ow2.petals.probes.api.key.ProbeKey;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * NIO Implementation of the Transporter.
 * 
 * @author Christophe DENEUX (Capgemini Sud)
 * @author Victor Noel - Linagora
 */
@Component(provides = { @Interface(name = "service", signature = Transporter.class),
        @Interface(name = NioReceiver.NIO_MSG_RECEIVER_ITF, signature = NioReceiver.class) })
public class NioTransporterImpl extends AbstractTransporterImpl implements NioReceiver {

    @Requires(name = "router")
    private RouterService router;

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration")
    private ConfigurationService configurationService;

    /**
     * All connections to remote containers are pooled into a keyed-pool. The
     * key is the name of the remote container.
     */
    @Requires(name = NioClientObjectPool.NIO_CLIENT_OBJECT_ITF)
    private NioClientObjectPool clientConnections;

    /**
     * The TCP transporter monitoring service - Fractal component
     */
    @Requires(name = "transportermonitoring")
    private TcpTransporterMonitoring tcpTransporterMonitoring;

    /**
     * Timeout to send a message exchange.
     */
    private long sendTimeout;

    public NioTransporterImpl() {
        super(new LoggingUtil(Logger.getLogger(Constants.FRACTAL_COMPONENT_LOGGER_NAME)));
    }

    @Override
    public void onMessage(final Object object) {
        this.log.start();

        if (object instanceof PetalsMessageExchange) {

            final PetalsMessageExchange exchange = (PetalsMessageExchange) object;

            final String exchangeId = exchange.getExchangeId();

            if (this.log.isDebugEnabled()) {
                this.log.debug("Receiving exchange " + exchangeId + " on NIO");
            }

            try {
                deliverExchange(exchange);
            } catch (final TransportException e) {
                if (exchange.isActive()) {
                    // TODO it would be best to send back an error to the other side so that the router can re-route the
                    // exchange to another endpoint if it can!
                    this.log.error("Couldn't deliver an exchange from NIO, sending back the error.", e);
                    try {
                        exchange.setError(e);
                        exchange.setTermination("Delivery error" + (exchange.getTermination() != null
                                ? " (was: " + exchange.getTermination() + ")" : ""));
                        this.router.sendInNMR(exchange);
                    } catch (final RoutingException e1) {
                        this.log.warning(
                                "Couldn't send back an error: the message is lost. Nevertheless, the sender should "
                                        + "not be assuming the exchange was received until it gets a DONE acknowledgement",
                                e1);
                    }
                } else {
                    this.log.error("Couldn't deliver an exchange from NIO.", e);
                }
            }
        } else {
            // This case shouldn't happen...
            this.log.warning("Unexpected received data (shouldn't happen). Data is discarded.");
        }

        this.log.end();
    }

    /**
     * This method does following action :
     * <ol>
     * <li>Serialize the message exchange.</li>
     * <li>Get the destination of the message.</li>
     * <li>Add it to the message.</li>
     * <li>Send it to the transport agent.</li>
     * </ol>
     */
    @Override
    public void send(final PetalsMessageExchange exchange, final TransportSendContext transportContext)
            throws TransportException {
        this.log.start();

        if (this.log.isDebugEnabled()) {
            this.log.debug("Sending exchange " + exchange.getExchangeId());
        }

        // we then send it over NIO
        final String containerName = transportContext.getDestination().getContainerName();

        final ProbeKey probeKey = this.tcpTransporterMonitoring.incPendingOutgoingMessagesProbe(containerName);
        if (this.log.isDebugEnabled()) {
            this.log.debug("Sending over NIO exchange " + exchange.getExchangeId() + " to " + containerName);
        }

        try {
            final NioClient clientConnection = this.clientConnections.borrowObject(containerName);
            try {
                this.tcpTransporterMonitoring.pickOutgoingConnectionNumbers(containerName);

                clientConnection.send(exchange, this.sendTimeout);

                this.tcpTransporterMonitoring.endsPendingOutgoingMessagesProbe(probeKey, ExecutionStatus.SUCCEEDED);

                // let's notify that the message was sent (only if no error happened)
                notifySent(exchange);

            } finally {
                this.clientConnections.returnObject(containerName, clientConnection);

                // We adjust values of probes when the client connection previously used was returned to the pool
                this.tcpTransporterMonitoring.pickOutgoingConnectionNumbers(containerName);
            }
        } catch (final Exception e) {
            this.tcpTransporterMonitoring.endsPendingOutgoingMessagesProbe(probeKey, ExecutionStatus.ERROR);
            throw new TransportException(e);
        }

        this.log.end();
    }

    @Lifecycle(step = Step.START)
    public void startFractalComponent() {
        this.log.start();

        this.sendTimeout = this.configurationService.getContainerConfiguration().getTCPSendTimeout();

        this.log.end();
    }
}