/**
 * Copyright (c) 2009-2012 Capgemini, 2009-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.transport.platform.nio.client;

import java.util.logging.Logger;

import org.apache.commons.pool.KeyedObjectPool;
import org.apache.commons.pool.impl.GenericKeyedObjectPool;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.transport.platform.nio.monitoring.TcpTransporterMonitoring;
import org.ow2.petals.microkernel.transport.platform.nio.selector.NioClient;
import org.ow2.petals.microkernel.transport.platform.nio.selector.NioSelector;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * This class is only a Fractal Wrapper of the Apache generic keyed object pool.
 * @author Christophe DENEUX - Capgemini Sud
 */
@Component(provides = @Interface(name = "service", signature = NioClientObjectPool.class))
public class NioClientObjectPoolImpl implements NioClientObjectPool {

    /**
     * Interface used to access configuration
     */
    @Requires(name = "configuration")
    protected ConfigurationService configurationService;

    /**
     * <p>
     * The socket selector Fractal component to use to process network IO.
     * </p>
     */
    @Requires(name = NioSelector.NIO_SELECTOR_ITF)
    protected NioSelector selectorAgent;

    /**
     * The TCP transporter monitoring service - Fractal component
     */
    @Requires(name = "transportermonitoring")
    private TcpTransporterMonitoring tcpTransporterMonitoring;

    /**
     * All connections to remote containers are pooled into a keyed-pool. The key is the name of the remote container.
     */
    private KeyedObjectPool<String, NioClientAgent> clientConnections;

    /**
     * Logger Wrapper
     */
    protected final LoggingUtil log = new LoggingUtil(Logger.getLogger(Constants.FRACTAL_COMPONENT_LOGGER_NAME));

    public NioClient borrowObject(final String targetContainerName) throws Exception {
        return this.clientConnections.borrowObject(targetContainerName);
    }

    public void returnObject(final String targetContainerName, final NioClient clientConnection)
            throws Exception {
        this.clientConnections.returnObject(targetContainerName, (NioClientAgent) clientConnection);
    }

    @Override
    public int getNumActive(final String containerName) {
        if (this.clientConnections != null) {
            return this.clientConnections.getNumActive(containerName);
        } else {
            return 0;
        }
    }

    @Override
    public int getNumIdle(final String containerName) {
        if (this.clientConnections != null) {
            return this.clientConnections.getNumIdle(containerName);
        } else {
            return 0;
        }
    }

    @Override
    public int getNumMax() {
        if (this.clientConnections != null) {
            return ((GenericKeyedObjectPool) this.clientConnections).getMaxActive();
        } else {
            return 0;
        }
    }

    @Override
    public int getNumMinIdle() {
        if (this.clientConnections != null) {
            return ((GenericKeyedObjectPool) this.clientConnections).getMinIdle();
        } else {
            return 0;
        }
    }

    @Lifecycle(step = Step.START)
    public void start() throws IllegalLifeCycleException, NoSuchInterfaceException {
        this.log.start();

        // Connection pool initialization
        final GenericKeyedObjectPool.Config config = new GenericKeyedObjectPool.Config();
        config.minIdle = 0;
        config.maxActive = this.configurationService.getContainerConfiguration().getTCPSenders();
        config.testOnBorrow = true;
        config.minEvictableIdleTimeMillis = this.configurationService.getContainerConfiguration()
                .getTCPSenderEvictableDelay();
        config.timeBetweenEvictionRunsMillis = this.configurationService
                .getContainerConfiguration().getTCPSenderEvictorDelay();

        this.clientConnections = new GenericKeyedObjectPool<String, NioClientAgent>(
                new NioClientAgentPoolObjectFactory(this.configurationService, this.selectorAgent,
                        this.configurationService.getContainerConfiguration().getTCPConnectionTimeout(),
                        this.tcpTransporterMonitoring, this.log.getLogger()),
                config);

        this.log.end();
    }

    /**
     * Method called when the fractal component is stopped
     * 
     * @throws Exception
     */
    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.start();

        // Close the connection pool and disconnect all clients
        try {
            this.clientConnections.close();
        } catch (final Exception e) {
            this.log.warning("Can't close the connections pool", e);
        }

        this.log.end();
    }

}
