/**
 * Copyright (c) 2009-2012 Capgemini, 2009-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.transport.platform.nio.server;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.io.input.ClassLoaderObjectInputStream;
import org.ow2.petals.microkernel.transport.platform.nio.monitoring.TcpTransporterMonitoring;
import org.ow2.petals.microkernel.transport.platform.nio.selector.PipedDeserializer;
import org.ow2.petals.microkernel.transport.platform.nio.util.stream.queue.QueueInputStream;
import org.ow2.petals.microkernel.transport.platform.nio.util.stream.queue.QueueOutputStream;
import org.ow2.petals.probes.api.enums.ExecutionStatus;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * @author Christophe DENEUX - Capgemini Sud
 */
public final class PipedDeserializerImpl implements PipedDeserializer {

    private final Logger logger;

    private final LoggingUtil log;

    private final OutputStream outputStream;

    private final InputStream inputStream;

    private final Object objectStreamLock = new Object();

    private final Socket socket;

    private final NioReceiver messageReceiver;

    private final TcpTransporterMonitoring tcpTransporterMonitoring;

    public PipedDeserializerImpl(final Socket socket, final NioReceiver messageReceiver,
            final TcpTransporterMonitoring tcpTransporterMonitoring, final Logger logger, final LoggingUtil log)
                    throws IOException {
        this.logger = logger;
        this.log = log;
        // We can't use the following stream:
        // - Circular byte buffer piped stream from Ostermiller because they
        // seem to be not thread-safe,
        // - Standard piped streams because they need to be connected before
        // writing/reading and writes are blocking.
        final QueueOutputStream queueOutputStream = new QueueOutputStream();
        this.outputStream = queueOutputStream;
        this.inputStream = new QueueInputStream(queueOutputStream);
        this.socket = socket;
        this.messageReceiver = messageReceiver;
        this.tcpTransporterMonitoring = tcpTransporterMonitoring;
    }

    @Override
    public final OutputStream getOutputStream() {
        return this.outputStream;
    }

    @Override
    public final void run() {

        try {

            if (this.logger.isLoggable(Level.FINE)) {
                this.log.debug("A read occurs on socket " + this.socket.toString());
            }

            final Object obj;
            synchronized (this.objectStreamLock) {
                obj = new ClassLoaderObjectInputStream(Thread.currentThread().getContextClassLoader(), this.inputStream)
                        .readUnshared();
            }

            if (this.logger.isLoggable(Level.FINE)) {
                this.log.debug("Received object: " + obj.toString());
            }

            this.tcpTransporterMonitoring.endsPendingIncomingMessagesProbe(
                    this.socket.getInetAddress().getHostAddress(), ExecutionStatus.SUCCEEDED);

            this.messageReceiver.onMessage(obj);

        } catch (final IOException | ClassNotFoundException e) {
            this.log.warning("An error occurs during deserialization: (" + e.getMessage() + "). Data is discarded.", e);
            this.tcpTransporterMonitoring.endsPendingIncomingMessagesProbe(
                    this.socket.getInetAddress().getHostAddress(), ExecutionStatus.ERROR);
        }
    }

}
