/**
 * Copyright (c) 2014-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.communication.topology;

import java.util.Collection;
import java.util.Collections;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.easymock.classextension.EasyMock;
import org.junit.Assert;
import org.junit.Rule;
import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.clientserverapi.topology.RemoteContainer;
import org.ow2.petals.clientserverapi.topology.exception.AttachContainerException;
import org.ow2.petals.clientserverapi.topology.exception.DetachContainerException;
import org.ow2.petals.clientserverapi.topology.exception.DomainIsNotDynamicException;
import org.ow2.petals.clientserverapi.topology.exception.InconsistencyException;
import org.ow2.petals.clientserverapi.topology.exception.NoDomainConfigurationProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.NoRemoteContainerProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.NoSecurityPassPhraseProvidedException;
import org.ow2.petals.microkernel.api.communication.sharedarea.SharedAreaService;
import org.ow2.petals.microkernel.api.communication.topology.TopologyService;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.configuration.ContainerConfigurationListener;
import org.ow2.petals.microkernel.api.configuration.DomainConfiguration;
import org.ow2.petals.microkernel.api.configuration.exception.ConfigurationException;
import org.ow2.petals.microkernel.api.jbi.messaging.RouterService;
import org.ow2.petals.microkernel.configuration.ConfigurationServiceImpl;
import org.ow2.petals.microkernel.configuration.ConfigurationTestUtils;
import org.ow2.petals.microkernel.configuration.LocalServerPropertiesBuilder;
import org.ow2.petals.microkernel.server.PetalsCompositeController;
import org.ow2.petals.microkernel.utest.mock.system.SharedAreaImpl;
import org.ow2.petals.topology.generated.Topology;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

/**
 * Abstract class for unit tests of {@link TopologyServiceImpl}.
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public abstract class AbstractTopologyServiceImplTest extends Assert {

    @Rule
    public final LocalServerPropertiesBuilder localServerPropertiesBuilder = new LocalServerPropertiesBuilder();

    public interface TestTopologyService extends TopologyService, AutoCloseable {
        // make TopologyService AutoClosable
    }

    public interface TestConfigurationService extends ConfigurationService, AutoCloseable {
        // make ConfigurationService AutoClosable
    }

    protected TestConfigurationService createAndStartConfSvc() throws Exception {
        return createAndStartConfSvc(Collections.<String, String> emptyMap());
    }

    protected TestConfigurationService createAndStartConfSvc(final Topology topology) throws Exception {
        return createAndStartConfSvc(Collections.<String, String> emptyMap(), topology);
    }

    protected TestConfigurationService createAndStartConfSvc(final Map<String, String> extraProperties)
            throws Exception {
        return createAndStartConfSvc(extraProperties, createBasicSampleTopology());
    }

    protected TestConfigurationService createAndStartConfSvc(final Map<String, String> extraProperties,
            final Topology topology) throws Exception {
        final Properties properties = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        properties.putAll(extraProperties);

        return createAndStartConfSvc(properties, topology);
    }

    protected static TestConfigurationService createAndStartConfSvc(final Properties properties,
            final Topology topology) throws Exception {
        // Create and initialize the configuration service
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();

        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        
        cfgSvc.loadConfiguration(properties, topology);

        return new TestConfigurationService() {

            @Override
            public ContainerConfiguration getContainerConfiguration() {
                return cfgSvc.getContainerConfiguration();
            }

            @Override
            public Map<String, String> getAllServerProperties() {
                return cfgSvc.getAllServerProperties();
            }

            @Override
            public Properties getServerProperties() {
                return cfgSvc.getServerProperties();
            }

            @Override
            public Set<ContainerConfiguration> getContainersConfiguration() {
                return cfgSvc.getContainersConfiguration();
            }

            @Override
            public DomainConfiguration getDomainConfiguration() {
                return cfgSvc.getDomainConfiguration();
            }

            @Override
            public Topology getLocalTopology() {
                return cfgSvc.getLocalTopology();
            }

            @Override
            public ContainerConfiguration getContainerConfiguration(String containerName) {
                return cfgSvc.getContainerConfiguration(containerName);
            }

            @Override
            public void loadConfiguration(Properties serverLocalProperties, Topology topology)
                    throws ConfigurationException {
                cfgSvc.loadConfiguration(serverLocalProperties, topology);
            }

            @Override
            public void updateTopology(Topology topology, boolean validate) throws ConfigurationException {
                cfgSvc.updateTopology(topology, validate);
            }

            @Override
            public void reconfigureDomain(Topology topology) throws ConfigurationException {
                cfgSvc.reconfigureDomain(topology);
            }

            @Override
            public boolean isSecurityTopologyPassphraseOk(String securityPassPhrase) {
                return cfgSvc.isSecurityTopologyPassphraseOk(securityPassPhrase);
            }

            @Override
            public void registerContainerConfigurationListener(ContainerConfigurationListener listener) {
                cfgSvc.registerContainerConfigurationListener(listener);
            }

            @Override
            public void deregisterContainerConfigurationListener(ContainerConfigurationListener listener) {
                cfgSvc.deregisterContainerConfigurationListener(listener);
            }

            @Override
            public void close() throws Exception {
                ConfigurationTestUtils.stopConfigurationService(cfgSvc);
            }
        };
    }

    /**
     * Instantiate and start the topology service
     */
    @SuppressWarnings("deprecation")
    protected static TestTopologyService createAndStartTopologySvc(final ConfigurationService cfgSvc,
            final PetalsCompositeController petalsCompositeCtrlMock) throws Exception {

        final SharedAreaService sharedAreaMock = EasyMock.createMock(SharedAreaService.class);
        Collection<SharedAreaService.SavedEndpoint> endpoints = Collections.<SharedAreaService.SavedEndpoint> emptyList();
        EasyMock.expect(sharedAreaMock.leaveTopology()).andReturn(endpoints);
        sharedAreaMock.joinTopology(true);
        sharedAreaMock.restoreEndpoints(endpoints);
        EasyMock.replay(sharedAreaMock);

        final RouterService routerMock = EasyMock.createMock(RouterService.class);
        EasyMock.expect(routerMock.waitEndOfPendingExchange(EasyMock.anyLong())).andReturn(true);
        EasyMock.replay(routerMock);

        final TopologyService topoSvc = new TopologyServiceImpl();
        ReflectionHelper.setPrivateField(topoSvc, "configurationService", cfgSvc);
        ReflectionHelper.setPrivateField(topoSvc, "sharedArea", sharedAreaMock);
        ReflectionHelper.setPrivateField(topoSvc, "router", routerMock);

        final PetalsCompositeController compController;
        if (petalsCompositeCtrlMock != null) {
            compController = petalsCompositeCtrlMock;
        } else {
            compController = EasyMock.createMock(PetalsCompositeController.class);
            compController.replaceSharedArea(SharedAreaImpl.class);
            EasyMock.replay(compController);
        }
        ReflectionHelper.setPrivateField(topoSvc, "petalsCompositeController", compController);

        ReflectionHelper.invokePrivateMethod(topoSvc, "start", null, null);

        return new TestTopologyService() {

            @Override
            public boolean isContainerReachable(String containerName) {
                return topoSvc.isContainerReachable(containerName);
            }

            @Override
            public Set<Map<String, String>> retrieveTopology(String securityPassPhrase,
                    boolean isSensibleInformationExpected) throws PetalsException {
                return topoSvc.retrieveTopology(securityPassPhrase, isSensibleInformationExpected);
            }

            @Override
            public Object retrieveTopologyObject(String securityPassPhrase, boolean isSensibleInformationExpected)
                    throws PetalsException {
                return topoSvc.retrieveTopologyObject(securityPassPhrase, isSensibleInformationExpected);
            }

            @Override
            public void attachContainerTo(RemoteContainer container, String securityPassPhrase)
                    throws NoRemoteContainerProvidedException, NoDomainConfigurationProvidedException,
                    DomainIsNotDynamicException, InconsistencyException, NoSecurityPassPhraseProvidedException,
                    AttachContainerException {
                topoSvc.attachContainerTo(container, securityPassPhrase);
            }

            @Override
            public void detachContainer() throws DomainIsNotDynamicException, DetachContainerException {
                topoSvc.detachContainer();
            }

            @Override
            public void close() throws Exception {
                ReflectionHelper.invokePrivateMethod(topoSvc, "stop", null, null);
            }
        };
    }

    protected static Topology createBasicSampleTopology() {
        final Topology topology = ConfigurationTestUtils.createBasicSampleTopology();
        topology.getDomain().setRegistryImplementation(SharedAreaImpl.class.getName());
        return topology;
    }

}
