/**
 * Copyright (c) 2014-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.communication.topology;

import java.util.Properties;

import org.easymock.classextension.EasyMock;
import org.junit.Rule;
import org.junit.Test;
import org.ow2.petals.clientserverapi.topology.RemoteContainer;
import org.ow2.petals.clientserverapi.topology.exception.AttachContainerException;
import org.ow2.petals.clientserverapi.topology.exception.DomainIsNotDynamicException;
import org.ow2.petals.clientserverapi.topology.exception.InconsistentDomainNameException;
import org.ow2.petals.clientserverapi.topology.exception.InvalidPassphraseException;
import org.ow2.petals.clientserverapi.topology.exception.NoDomainConfigurationProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.NoRemoteContainerProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.NoSecurityPassPhraseProvidedException;
import org.ow2.petals.jmx.api.api.TopologyServiceClient;
import org.ow2.petals.jmx.api.api.exception.ConnectionErrorException;
import org.ow2.petals.jmx.api.mock.junit.PetalsJmxApiJunitRule;
import org.ow2.petals.launcher.api.server.conf.ConfigurationProperties;
import org.ow2.petals.microkernel.configuration.ConfigurationTestUtils;
import org.ow2.petals.microkernel.configuration.LocalServerPropertiesBuilder;
import org.ow2.petals.microkernel.server.PetalsCompositeController;
import org.ow2.petals.microkernel.utest.mock.system.SharedAreaImpl;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.Domain;
import org.ow2.petals.topology.generated.DomainMode;
import org.ow2.petals.topology.generated.JmxService;
import org.ow2.petals.topology.generated.Topology;
import org.ow2.petals.topology.generated.TransportService;

/**
 * Unit tests of {@link TopologyServiceImpl#attachContainerTo(RemoteContainer, String)}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class TopologyServiceImpl_AttachContainerToTest extends AbstractTopologyServiceImplTest {

    private final static String TARGET_DOMAIN_NAME = "target-domain-name";

    private final static String TARGET_CONTAINER_NAME = "target-container-name";

    /**
     * Port of the JMX server of the target container
     */
    private final static Integer TARGET_JMX_PORT = Integer.valueOf(7070);

    /**
     * Hostname of the target container
     */
    private final static String TARGET_JMX_HOST = "localhost";

    private final static String TARGET_JMX_USER = "petals-user";

    private final static String TARGET_JMX_PWD = "petals-pwd";

    private final static int TARGET_CONTAINER_TCP_PORT = ConfigurationTestUtils.CONTAINER_TCP_PORT_0 + 1;

    private final static String TARGET_PASS_PHRASE = "pass-phrase";

    @Rule
    public final PetalsJmxApiJunitRule jmxApiOnTargetContainer = new PetalsJmxApiJunitRule(TARGET_JMX_HOST,
            TARGET_JMX_PORT, null, null);

    @Rule
    public final LocalServerPropertiesBuilder targetServerPropertiesBuilder = new LocalServerPropertiesBuilder();

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(RemoteContainer, String)} against:
     * <ul>
     * <li>no container configuration is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs about the missing container configuration
     * </p>
     */
    @Test(expected = NoRemoteContainerProvidedException.class)
    public void testNoRemoteContainer() throws Exception {
        try (final TestConfigurationService cfgSvc = createAndStartConfSvc();
                final TestTopologyService topoSvc = createAndStartTopologySvc(cfgSvc, null)) {
            topoSvc.attachContainerTo(null, null);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(RemoteContainer, String)} against:
     * <ul>
     * <li>no security pass-phrase is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs about the missing security pass-phrase
     * </p>
     */
    @Test(expected = NoSecurityPassPhraseProvidedException.class)
    public void testNoSecurityPassPhrase() throws Exception {
        try (final TestConfigurationService cfgSvc = createAndStartConfSvc();
                final TestTopologyService topoSvc = createAndStartTopologySvc(cfgSvc, null)) {
            topoSvc.attachContainerTo(new RemoteContainer(), null);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(RemoteContainer, String)} against:
     * <ul>
     * <li>an invalid pass-phrase is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs about the invalid pass-phrase
     * </p>
     */
    @Test(expected = InvalidPassphraseException.class)
    public void testInvalidPassPhrase() throws Exception {

        // Create a basic target topology for the remote container
        final Topology basicTargetTopology = createBasicTargetTopology();
        basicTargetTopology.getDomain().setMode(DomainMode.DYNAMIC);

        final Properties targetServerProps = this.targetServerPropertiesBuilder
                .createMinimalServerProperties(TARGET_CONTAINER_NAME);
        targetServerProps.put(ConfigurationProperties.TOPOLOGY_PASSPHRASE, TARGET_PASS_PHRASE);

        try (final TestConfigurationService targetCfgSvc = createAndStartConfSvc(targetServerProps,
                basicTargetTopology)) {
            final PetalsCompositeController targetPetalsCompositeCtrl = EasyMock
                    .createMock(PetalsCompositeController.class);
            EasyMock.expect(targetPetalsCompositeCtrl.getSharedAreaCltConfiguration(TARGET_PASS_PHRASE))
                    .andReturn(new Properties()).anyTimes();
            targetPetalsCompositeCtrl.replaceSharedArea(SharedAreaImpl.class);
            EasyMock.replay(targetPetalsCompositeCtrl);

            try (final TestTopologyService targetTopoSvc = createAndStartTopologySvc(targetCfgSvc,
                    targetPetalsCompositeCtrl)) {

                // Create a basic sample topology for the current container
                final Topology basicCurrentTopology = createBasicSampleTopology();
                final Domain currentDomain = basicCurrentTopology.getDomain();
                currentDomain.setMode(DomainMode.DYNAMIC);

                try (final TestConfigurationService currentCfgSvc = createAndStartConfSvc(basicCurrentTopology);
                        final TestTopologyService currentTopoSvc = createAndStartTopologySvc(currentCfgSvc, null)) {

                    final RemoteContainer targetRemoteContainer = new RemoteContainer();
                    targetRemoteContainer.setDomainName(TARGET_DOMAIN_NAME);
                    targetRemoteContainer.setHost(TARGET_JMX_HOST);
                    targetRemoteContainer.setPort(TARGET_JMX_PORT);

                    final TopologyServiceClient jmxTopologyServiceClientMock = EasyMock
                            .createMock(TopologyServiceClient.class);
                    EasyMock.expect(jmxTopologyServiceClientMock.retrieveTopologyObject("invalid-pass-phrase", true))
                            .andReturn(targetTopoSvc.retrieveTopologyObject("invalid-pass-phrase", true)).anyTimes();

                    EasyMock.replay(jmxTopologyServiceClientMock);
                    this.jmxApiOnTargetContainer.setTopologyServiceClient(jmxTopologyServiceClientMock);

                    currentTopoSvc.attachContainerTo(targetRemoteContainer, "invalid-pass-phrase");
                }
            }
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(RemoteContainer, String)} against:
     * <ul>
     * <li>an empty security pass-phrase is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs about the missing security pass-phrase
     * </p>
     */
    @Test(expected = NoSecurityPassPhraseProvidedException.class)
    public void testEmptySecurityPassPhrase() throws Exception {

        try (final TestConfigurationService cfgSvc = createAndStartConfSvc();
                final TestTopologyService topoSvc = createAndStartTopologySvc(cfgSvc, null)) {
            topoSvc.attachContainerTo(new RemoteContainer(), "");
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(RemoteContainer, String)} against:
     * <ul>
     * <li>the domain of the given container configuration is not set</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs about the missing domain information of the given container
     * </p>
     */
    @Test(expected = NoDomainConfigurationProvidedException.class)
    public void testNoDomain() throws Exception {

        try (final TestConfigurationService cfgSvc = createAndStartConfSvc();
                final TestTopologyService topoSvc = createAndStartTopologySvc(cfgSvc, null)) {
            topoSvc.attachContainerTo(new RemoteContainer(), TARGET_PASS_PHRASE);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(RemoteContainer, String)} against:
     * <ul>
     * <li>the domain of the current container is static</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs because the domain mode is not <b>dynamic</b>
     * </p>
     */
    @Test(expected = DomainIsNotDynamicException.class)
    public void testCurrentDomainNotDynamic() throws Exception {

        // Create a basic sample topology
        final Topology basicCurrentTopology = createBasicSampleTopology();
        basicCurrentTopology.getDomain().setMode(DomainMode.STATIC);

        try (final TestConfigurationService cfgSvc = createAndStartConfSvc(basicCurrentTopology);
                final TestTopologyService topoSvc = createAndStartTopologySvc(cfgSvc, null)) {
            final RemoteContainer targetRemoteContainer = new RemoteContainer();
            targetRemoteContainer.setDomainName(TARGET_DOMAIN_NAME);
            targetRemoteContainer.setHost(TARGET_JMX_HOST);
            targetRemoteContainer.setPort(TARGET_JMX_PORT);

            topoSvc.attachContainerTo(targetRemoteContainer, TARGET_PASS_PHRASE);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(RemoteContainer, String)} against:
     * <ul>
     * <li>the target container is not running</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs because the container can't be contacted
     * </p>
     */
    @Test(expected = ConnectionErrorException.class)
    public void testTargetContainerNotRunning() throws Throwable {

        // Create a basic sample topology
        final Topology basicCurrentTopology = createBasicSampleTopology();
        basicCurrentTopology.getDomain().setMode(DomainMode.DYNAMIC);

        try (final TestConfigurationService cfgSvc = createAndStartConfSvc(basicCurrentTopology);
                final TestTopologyService topoSvc = createAndStartTopologySvc(cfgSvc, null)) {
            final RemoteContainer targetRemoteContainer = new RemoteContainer();
            targetRemoteContainer.setDomainName(TARGET_DOMAIN_NAME);
            targetRemoteContainer.setHost("host-not-running");
            targetRemoteContainer.setPort(TARGET_JMX_PORT);

            topoSvc.attachContainerTo(targetRemoteContainer, TARGET_PASS_PHRASE);
        } catch (final AttachContainerException e) {
            // It's the expected direct exception. The error connection exception is the cause of this exception
            throw e.getCause();
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(RemoteContainer, String)} against:
     * <ul>
     * <li>the domain in which the container must be attached is not the expected one</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs about the inconsistency about domain name
     * </p>
     */
    @Test(expected = InconsistentDomainNameException.class)
    public void testTargetDomainNotFound() throws Exception {

        // Create a basic target topology for the remote container
        final Topology basicTargetTopology = createBasicTargetTopology();

        final Properties targetServerProps = this.targetServerPropertiesBuilder
                .createMinimalServerProperties(TARGET_CONTAINER_NAME);
        targetServerProps.put(ConfigurationProperties.TOPOLOGY_PASSPHRASE, TARGET_PASS_PHRASE);

        try (final TestConfigurationService targetCfgSvc = createAndStartConfSvc(targetServerProps,
                basicTargetTopology)) {

            final PetalsCompositeController targetPetalsCompositeCtrl = EasyMock
                    .createMock(PetalsCompositeController.class);
            EasyMock.expect(targetPetalsCompositeCtrl.getSharedAreaCltConfiguration(TARGET_PASS_PHRASE))
                    .andReturn(new Properties()).anyTimes();
            targetPetalsCompositeCtrl.replaceSharedArea(SharedAreaImpl.class);
            EasyMock.replay(targetPetalsCompositeCtrl);

            try (final TestTopologyService targetTopoSvc = createAndStartTopologySvc(targetCfgSvc,
                    targetPetalsCompositeCtrl)) {
                // Create a basic sample topology for the current container
                final Topology basicCurrentTopology = createBasicSampleTopology();
                basicCurrentTopology.getDomain().setMode(DomainMode.DYNAMIC);

                try (final TestConfigurationService currentCfgSvc = createAndStartConfSvc(basicCurrentTopology);
                        final TestTopologyService currentTopoSvc = createAndStartTopologySvc(currentCfgSvc, null)) {
                    final RemoteContainer targetRemoteContainer = new RemoteContainer();
                    targetRemoteContainer.setDomainName("inconsistent-domain-name");
                    targetRemoteContainer.setHost(TARGET_JMX_HOST);
                    targetRemoteContainer.setPort(TARGET_JMX_PORT);

                    final TopologyServiceClient jmxTopologyServiceClientMock = EasyMock
                            .createMock(TopologyServiceClient.class);
                    EasyMock.expect(jmxTopologyServiceClientMock.retrieveTopologyObject(TARGET_PASS_PHRASE, true))
                            .andReturn(targetTopoSvc.retrieveTopologyObject(TARGET_PASS_PHRASE, true)).anyTimes();
                    EasyMock.replay(jmxTopologyServiceClientMock);
                    this.jmxApiOnTargetContainer.setTopologyServiceClient(jmxTopologyServiceClientMock);

                    currentTopoSvc.attachContainerTo(targetRemoteContainer, TARGET_PASS_PHRASE);
                }
            }
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(RemoteContainer, String)} against:
     * <ul>
     * <li>the domain of the remote container is static</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs because the remote domain mode is not <b>dynamic</b>
     * </p>
     */
    @Test(expected = DomainIsNotDynamicException.class)
    public void testTargetDomainNotDynamic() throws Exception {
        
        // Create a basic target topology for the remote container
        final Topology basicTargetTopology = createBasicTargetTopology();
        basicTargetTopology.getDomain().setMode(DomainMode.STATIC);

        final Properties targetServerProps = this.targetServerPropertiesBuilder
                .createMinimalServerProperties(TARGET_CONTAINER_NAME);
        targetServerProps.put(ConfigurationProperties.TOPOLOGY_PASSPHRASE, TARGET_PASS_PHRASE);

        try (final TestConfigurationService targetCfgSvc = createAndStartConfSvc(targetServerProps,
                basicTargetTopology)) {

            final PetalsCompositeController targetPetalsCompositeCtrl = EasyMock
                    .createMock(PetalsCompositeController.class);
            EasyMock.expect(targetPetalsCompositeCtrl.getSharedAreaCltConfiguration(TARGET_PASS_PHRASE))
                    .andReturn(new Properties()).anyTimes();
            targetPetalsCompositeCtrl.replaceSharedArea(SharedAreaImpl.class);
            EasyMock.replay(targetPetalsCompositeCtrl);

            try (final TestTopologyService targetTopoSvc = createAndStartTopologySvc(targetCfgSvc,
                    targetPetalsCompositeCtrl)) {
                // Create a basic sample topology for the current container
                final Topology basicCurrentTopology = createBasicSampleTopology();
                basicCurrentTopology.getDomain().setMode(DomainMode.DYNAMIC);

                try (final TestConfigurationService currentCfgSvc = createAndStartConfSvc(basicCurrentTopology);
                        final TestTopologyService currentTopoSvc = createAndStartTopologySvc(currentCfgSvc, null)) {
                    final RemoteContainer targetRemoteContainer = new RemoteContainer();
                    targetRemoteContainer.setDomainName(TARGET_DOMAIN_NAME);
                    targetRemoteContainer.setHost(TARGET_JMX_HOST);
                    targetRemoteContainer.setPort(TARGET_JMX_PORT);

                    final TopologyServiceClient jmxTopologyServiceClientMock = EasyMock
                            .createMock(TopologyServiceClient.class);
                    EasyMock.expect(jmxTopologyServiceClientMock.retrieveTopologyObject(TARGET_PASS_PHRASE, true))
                            .andReturn(targetTopoSvc.retrieveTopologyObject(TARGET_PASS_PHRASE, true)).anyTimes();
                    EasyMock.replay(jmxTopologyServiceClientMock);
                    this.jmxApiOnTargetContainer.setTopologyServiceClient(jmxTopologyServiceClientMock);

                    currentTopoSvc.attachContainerTo(targetRemoteContainer, TARGET_PASS_PHRASE);
                }
            }
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(RemoteContainer, String)} against:
     * <ul>
     * <li>all is ok</li>
     * </ul>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>No error occurs</li>
     * <li>...
     * <li>
     * </ul>
     * </p>
     */
    @Test
    public void testAllIsOk() throws Exception {

        // Create a basic target topology for the remote container
        final Topology basicTargetTopology = createBasicTargetTopology();
        final Domain targetDomain = basicTargetTopology.getDomain();
        targetDomain.setMode(DomainMode.DYNAMIC);

        final Properties targetServerProps = this.targetServerPropertiesBuilder
                .createMinimalServerProperties(TARGET_CONTAINER_NAME);
        targetServerProps.put(ConfigurationProperties.TOPOLOGY_PASSPHRASE, TARGET_PASS_PHRASE);

        try (final TestConfigurationService targetCfgSvc = createAndStartConfSvc(targetServerProps,
                basicTargetTopology)) {
            targetCfgSvc.loadConfiguration(targetServerProps, basicTargetTopology);
            final PetalsCompositeController targetPetalsCompositeCtrl = EasyMock
                    .createMock(PetalsCompositeController.class);
            EasyMock.expect(targetPetalsCompositeCtrl.getSharedAreaCltConfiguration(TARGET_PASS_PHRASE))
                    .andReturn(new Properties()).anyTimes();
            targetPetalsCompositeCtrl.replaceSharedArea(SharedAreaImpl.class);
            EasyMock.replay(targetPetalsCompositeCtrl);
            try (final TestTopologyService targetTopoSvc = createAndStartTopologySvc(targetCfgSvc,
                    targetPetalsCompositeCtrl)) {

                // Create a basic sample topology for the current container
                final Topology basicCurrentTopology = createBasicSampleTopology();
                final Domain currentDomain = basicCurrentTopology.getDomain();
                currentDomain.setMode(DomainMode.DYNAMIC);

                try (final TestConfigurationService currentCfgSvc = createAndStartConfSvc(basicCurrentTopology)) {

                    final PetalsCompositeController petalsCompositeCtrlMock = EasyMock
                            .createMock(PetalsCompositeController.class);
                    @SuppressWarnings("resource")
                    final AutoCloseable msPause = EasyMock.createMock(AutoCloseable.class);
                    msPause.close();
                    @SuppressWarnings("resource")
                    final AutoCloseable edPause = EasyMock.createMock(AutoCloseable.class);
                    edPause.close();
                    @SuppressWarnings("resource")
                    final AutoCloseable saPause = EasyMock.createMock(AutoCloseable.class);
                    saPause.close();
                    EasyMock.expect(petalsCompositeCtrlMock.pauseMessagingSubsystem()).andReturn(msPause);
                    EasyMock.expect(petalsCompositeCtrlMock.pauseEndpointDirectory()).andReturn(edPause);
                    EasyMock.expect(petalsCompositeCtrlMock.pauseSharedArea()).andReturn(saPause);
                    petalsCompositeCtrlMock.replaceSharedArea(SharedAreaImpl.class);
                    EasyMock.expectLastCall().times(2);
                    petalsCompositeCtrlMock.onConfigurationUpdated();
                    EasyMock.replay(petalsCompositeCtrlMock, msPause, edPause, saPause);

                    try (final TestTopologyService currentTopoSvc = createAndStartTopologySvc(currentCfgSvc,
                            petalsCompositeCtrlMock)) {
                        final RemoteContainer targetRemoteContainer = new RemoteContainer();
                        targetRemoteContainer.setDomainName(TARGET_DOMAIN_NAME);
                        targetRemoteContainer.setHost(TARGET_JMX_HOST);
                        targetRemoteContainer.setPort(TARGET_JMX_PORT);

                        final TopologyServiceClient jmxTopologyServiceClientMock = EasyMock
                                .createMock(TopologyServiceClient.class);
                        EasyMock.expect(jmxTopologyServiceClientMock.retrieveTopologyObject(TARGET_PASS_PHRASE, true))
                                .andReturn(targetTopoSvc.retrieveTopologyObject(TARGET_PASS_PHRASE, true)).anyTimes();
                        EasyMock.replay(jmxTopologyServiceClientMock);
                        this.jmxApiOnTargetContainer.setTopologyServiceClient(jmxTopologyServiceClientMock);

                        currentTopoSvc.attachContainerTo(targetRemoteContainer, TARGET_PASS_PHRASE);
                    }
                }
            }
        }
    }

    /**
     * Create a sample topology with only required parameters
     * 
     * @return A sample topology with only required parameters
     */
    private static Topology createBasicTargetTopology() {
        final Topology topology = new Topology();
        final Domain domain = new Domain();
        topology.setDomain(domain);

        domain.setName(TARGET_DOMAIN_NAME);
        domain.setMode(DomainMode.DYNAMIC);
        domain.setRegistryImplementation(SharedAreaImpl.class.getName());

        final Container container0 = new Container();
        domain.getContainer().add(container0);
        container0.setName(TARGET_CONTAINER_NAME);
        container0.setHost(TARGET_JMX_HOST);
        container0.setUser(TARGET_JMX_USER);
        container0.setPassword(TARGET_JMX_PWD);

        final JmxService jmxService0 = new JmxService();
        container0.setJmxService(jmxService0);

        final TransportService transportService0 = new TransportService();
        transportService0.setTcpPort(TARGET_CONTAINER_TCP_PORT);
        container0.setTransportService(transportService0);

        return topology;
    }
}
