/**
 * Copyright (c) 2014-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.configuration;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.util.Properties;

import org.ow2.petals.launcher.api.server.conf.ConfigurationProperties;
import org.ow2.petals.microkernel.api.communication.topology.TopologyService;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.configuration.DomainConfiguration;
import org.ow2.petals.microkernel.api.jbi.messaging.RouterService;
import org.ow2.petals.microkernel.api.transport.FastRemoteTransportService;
import org.ow2.petals.topology.generated.Topology;

import com.ebmwebsourcing.easycommons.properties.PropertiesException;
import com.ebmwebsourcing.easycommons.properties.PropertiesHelper;

/**
 * Abstract class for unit tests of {@link ConfigurationServiceImpl}.
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public abstract class AbstractConfigurationServiceImplTest {

    protected final Topology topology = ConfigurationTestUtils.createBasicSampleTopology();

    /**
     * Assert to check the configuration loaded from the sample topology created with {@link #getBasicSampleTopology()}
     * 
     * @param actualCfgSvc
     *            The configuration service containing the configuration loaded.
     */
    protected void assertBasicSampleTopology(final ConfigurationService actualCfgSvc) {

        final DomainConfiguration domainConfig = actualCfgSvc.getDomainConfiguration();
        assertEquals("Unexpected domain name", ConfigurationTestUtils.DOMAIN_NAME, domainConfig.getName());
        assertEquals("Unexpected domain description", ConfigurationTestUtils.DOMAIN_DESCRIPTION,
                domainConfig.getDescription());
        assertEquals("Unexpected domain mode", DomainConfiguration.DomainMode.STATIC, domainConfig.getMode());
        assertEquals("Unexpected number of container", 1, actualCfgSvc.getContainersConfiguration().size());

        final ContainerConfiguration containerConfig = actualCfgSvc.getContainerConfiguration();
        assertEquals("Unexpected container name", ConfigurationTestUtils.CONTAINER_NAME_0, containerConfig.getName());
        assertEquals("Unexpected container description", ConfigurationTestUtils.CONTAINER_DESCRIPTION_0,
                containerConfig.getDescription());
        assertEquals("Unexpected container host", ConfigurationTestUtils.CONTAINER_HOST, containerConfig.getHost());
        assertEquals("Unexpected container jmx user", ConfigurationTestUtils.CONTAINER_JMX_USER,
                containerConfig.getUser());
        assertEquals("Unexpected container jmx pwd", ConfigurationTestUtils.CONTAINER_JMX_PWD,
                containerConfig.getPassword());
    }

    /**
     * Assert to check the configuration loaded against default server property values
     * 
     * @param containerConfig
     *            The container configuration containing server properties loaded
     * @param lspb
     *            The {@link LocalServerPropertiesBuilder} used to create local server properties
     */
    protected void assertDefaultServerProperties(final ContainerConfiguration containerConfig,
            final LocalServerPropertiesBuilder lspb)
            throws PropertiesException {
        final Properties resolvingPropertyValues = new Properties();
        resolvingPropertyValues.put(ConfigurationProperties.DATA_ROOT_DIRECTORY_PROPERTY_NAME, lspb.getDataArea()
                .getAbsolutePath());

        assertEquals("unexpected default value for repository path", containerConfig.getRepositoryDirectoryPath(),
                PropertiesHelper
                        .resolveString(ConfigurationProperties.DEFAULT_REPOSITORY_PATH, resolvingPropertyValues));
        assertEquals("unexpected default value for working path", containerConfig.getWorkDirectoryPath(),
                PropertiesHelper.resolveString(ConfigurationProperties.DEFAULT_WORKING_AREA_PATH,
                        resolvingPropertyValues));
        assertEquals("unexpected default value for the maximum duration of a life-cycle operation processing",
                containerConfig.getTaskTimeout(), ContainerConfiguration.DEFAULT_INSTALLATION_DEPLOYMENT_TASK_TIMEOUT);
        assertFalse("unexpected default value for the control of exchange acceptance by target component",
                containerConfig.isExchangeValidation());
        assertTrue("unexpected default value for the classloader isolation",
                containerConfig.isIsolateJBIClassLoaders());
        assertTrue("non null and non empty default value for topology passphrase",
                containerConfig.getTopologyPassPhrase() == null || containerConfig.getTopologyPassPhrase().isEmpty());
        assertEquals("unexpected default value for the maximum waiting time to acquire the topology lock",
                containerConfig.getTopologyLocksMaxWaitTime(),
                TopologyService.DEFAULT_TOPOLOGY_LOCKS_MAX_WAIT_TIME);

        // Router asserts
        assertEquals("unexpected default value for the router strategy", containerConfig.getRouterStrategy(),
                RouterService.DEFAULT_PLATFORM_STRATEGY);
        assertEquals("unexpected default value for the router send attempts", containerConfig.getRouterSendAttempt(),
                RouterService.DEFAULT_SEND_ATTEMPTS);
        assertEquals("unexpected default value for the router send delay", containerConfig.getRouterSendDelay(),
                RouterService.DEFAULT_SEND_DELAY);

        // Transporter
        assertEquals("unexpected default value for the transport queue max size",
                containerConfig.getTransporterQueueMaxSize(), RouterService.DEFAULT_TRANSPORT_QUEUE_MAX_SIZE);
        assertEquals("unexpected default value for the transport queue offering timeout",
                containerConfig.getTransporterQueueOfferingTimeout(),
                RouterService.DEFAULT_TRANSPORT_QUEUE_OFFERING_TIMEOUT);

        // TCP Transporter
        assertEquals("unexpected default value for the TCP transport sender number", containerConfig.getTCPSenders(),
                FastRemoteTransportService.DEFAULT_SENDERS);
        assertEquals("unexpected default value for the connection timeout of the fast remote transporter",
                containerConfig.getTCPConnectionTimeout(), FastRemoteTransportService.DEFAULT_CONNECTION_TIMEOUT);
        assertEquals("unexpected default value for the send timeout of the fast remote transporter",
                containerConfig.getTCPSendTimeout(), FastRemoteTransportService.DEFAULT_SEND_TIMEOUT);
        assertEquals(
                "unexpected default value for the delay before running the 'sender' eviction thread of the fast remote transporter",
                containerConfig.getTCPSenderEvictorDelay(), FastRemoteTransportService.DEFAULT_SEND_EVICTOR_DELAY);
        assertEquals(
                "unexpected default value for the delay before an idle 'sender' of the fast remote transporter is set evictable",
                containerConfig.getTCPSenderEvictableDelay(), FastRemoteTransportService.DEFAULT_SEND_EVICTABLE_DELAY);
    }

}
