/**
 * Copyright (c) 2014-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.configuration;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Properties;

import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;
import org.ow2.petals.launcher.api.server.conf.ConfigurationProperties;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.exception.ConfigurationException;
import org.ow2.petals.microkernel.api.configuration.exception.ContainerNameDuplicatedException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerDuplicatedPortException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerHostException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerNameException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerTransportSvcException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentDomainJndiFactoryException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentDomainJndiProviderUrlException;
import org.ow2.petals.microkernel.api.configuration.exception.MissingServerPropertyException;
import org.ow2.petals.topology.TopologyBuilder;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.Domain;
import org.ow2.petals.topology.generated.JmxService;
import org.ow2.petals.topology.generated.Jndi;
import org.ow2.petals.topology.generated.Topology;
import org.ow2.petals.topology.generated.TransportService;

import com.ebmwebsourcing.easycommons.properties.PropertiesException;

/**
 * Unit tests of
 * {@link ConfigurationServiceImpl#loadConfiguration(java.util.Properties, org.ow2.petals.topology.generated.Topology)}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ConfigurationServiceImpl_LoadConfigurationTest extends AbstractConfigurationServiceImplTest {

    private final static String DOMAIN_JNDI_FACTORY = "domain.jndi.factory";

    private final static String DOMAIN_JNDI_PROVIDER_URL = "http://domain.jndi.provider.url";

    private final static String PERSISTED_DOMAIN_NAME = "domain-name-of-the-persisted-topology-definition";

    @Rule
    public final TemporaryFolder temporaryFolder = new TemporaryFolder();

    @Rule
    public final LocalServerPropertiesBuilder localServerPropertiesBuilder = new LocalServerPropertiesBuilder();

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set except
     * {@value ConfigurationProperties#DATA_ROOT_DIRECTORY_PROPERTY_NAME} that is not set</li>
     * <li>a sample topology is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: The error {@link MissingServerPropertyException} occurs because the property
     * {@value ConfigurationProperties#DATA_ROOT_DIRECTORY_PROPERTY_NAME} is missing.
     * </p>
     */
    @Test(expected = MissingServerPropertyException.class)
    public void testServerPropWithMissingRequiredPropsAndTopology_001() throws ConfigurationException,
            InvocationTargetException {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = new Properties();
        serverProps.put(ConfigurationProperties.WORKING_AREA_PATH_PROPERTY_NAME, "working.area.path.value");
        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);
        } catch (final MissingServerPropertyException e) {
            assertEquals("unexpected missing property", ConfigurationProperties.DATA_ROOT_DIRECTORY_PROPERTY_NAME,
                    e.getMissingPropertyName());
            throw e;
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set except
     * {@value ConfigurationProperties#DATA_ROOT_DIRECTORY_PROPERTY_NAME} that is empty</li>
     * <li>a sample topology is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: The error {@link MissingServerPropertyException} occurs because the property
     * {@value ConfigurationProperties#DATA_ROOT_DIRECTORY_PROPERTY_NAME} is missing.
     * </p>
     */
    @Test(expected = MissingServerPropertyException.class)
    public void testServerPropWithMissingRequiredPropsAndTopology_002() throws ConfigurationException,
            InvocationTargetException {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = new Properties();
        serverProps.put(ConfigurationProperties.WORKING_AREA_PATH_PROPERTY_NAME, "working.area.path.value");
        serverProps.put(ConfigurationProperties.DATA_ROOT_DIRECTORY_PROPERTY_NAME, "");
        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);
        } catch (final MissingServerPropertyException e) {
            assertEquals("unexpected missing property", ConfigurationProperties.DATA_ROOT_DIRECTORY_PROPERTY_NAME,
                    e.getMissingPropertyName());
            throw e;
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set except
     * {@value ConfigurationProperties#WORKING_AREA_PATH_PROPERTY_NAME} that is empty</li>
     * <li>a sample topology is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: The error {@link MissingServerPropertyException} occurs because the property
     * {@value ConfigurationProperties#WORKING_AREA_PATH_PROPERTY_NAME} is missing.
     * </p>
     */
    @Test(expected = MissingServerPropertyException.class)
    public void testServerPropWithMissingRequiredPropsAndTopology_003() throws ConfigurationException,
            InvocationTargetException, IOException {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        serverProps.put(ConfigurationProperties.WORKING_AREA_PATH_PROPERTY_NAME, "");
        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);
        } catch (final MissingServerPropertyException e) {
            assertEquals("unexpected missing property", ConfigurationProperties.WORKING_AREA_PATH_PROPERTY_NAME,
                    e.getMissingPropertyName());
            throw e;
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set except
     * {@value ConfigurationProperties#CONTAINER_NAME} that is not set (not set in server properties and not set as
     * system property)</li>
     * <li>a sample topology is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: The error {@link MissingServerPropertyException} occurs because the property
     * {@value ConfigurationProperties#WORKING_AREA_PATH_PROPERTY_NAME} is missing.
     * </p>
     */
    @Test(expected = MissingServerPropertyException.class)
    public void testServerPropWithMissingRequiredPropsAndTopology_004() throws ConfigurationException,
            InvocationTargetException, IOException {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        serverProps.remove(ConfigurationProperties.CONTAINER_NAME);
        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);
        } catch (final MissingServerPropertyException e) {
            assertEquals("unexpected missing property", ConfigurationProperties.CONTAINER_NAME,
                    e.getMissingPropertyName());
            throw e;
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set except
     * {@value ConfigurationProperties#CONTAINER_NAME} that is empty (set empty in server properties)</li>
     * <li>a sample topology is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: The error {@link MissingServerPropertyException} occurs because the property
     * {@value ConfigurationProperties#WORKING_AREA_PATH_PROPERTY_NAME} is missing.
     * </p>
     */
    @Test(expected = MissingServerPropertyException.class)
    public void testServerPropWithMissingRequiredPropsAndTopology_005() throws ConfigurationException,
            InvocationTargetException, IOException {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        serverProps.put(ConfigurationProperties.CONTAINER_NAME, "");
        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);
        } catch (final MissingServerPropertyException e) {
            assertEquals("unexpected missing property", ConfigurationProperties.CONTAINER_NAME,
                    e.getMissingPropertyName());
            throw e;
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set except
     * {@value ConfigurationProperties#CONTAINER_NAME} that is set to a valid value (set at system property level, no
     * set in server properties)</li>
     * <li>a sample topology is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: No error occurs. The container name is the one defined by the system property.
     * </p>
     */
    @Test
    public void testServerPropWithMissingRequiredPropsAndTopology_006() throws ConfigurationException,
            InvocationTargetException, IOException {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        serverProps.remove(ConfigurationProperties.CONTAINER_NAME);
        System.setProperty(ConfigurationProperties.CONTAINER_NAME, ConfigurationTestUtils.CONTAINER_NAME_0);
        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            assertEquals("Unexpected container name", ConfigurationTestUtils.CONTAINER_NAME_0, cfgSvc
                    .getContainerConfiguration().getName());
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set except
     * {@value ConfigurationProperties#CONTAINER_NAME} that is set to a valid value (set through the server properties,
     * a different value is set at system property level)</li>
     * <li>a sample topology is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: No error occurs. The container name is the one defined by server properties.
     * </p>
     */
    @Test
    public void testServerPropWithMissingRequiredPropsAndTopology_007() throws ConfigurationException,
            InvocationTargetException, IOException {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        System.setProperty(ConfigurationProperties.CONTAINER_NAME, "another-container-name");
        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            assertEquals("Unexpected container name", ConfigurationTestUtils.CONTAINER_NAME_0, cfgSvc
                    .getContainerConfiguration().getName());
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set</li>
     * <li>a sample topology is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>No error occurs</li>.
     * <li>Default values are correctly set,</li>
     * <li>Topology correctly set.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testServerPropSampleTopology() throws ConfigurationException, InvocationTargetException,
            PropertiesException, IOException {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            assertNotNull("Initial local topology null", cfgSvc.getLocalTopology());
            assertBasicSampleTopology(cfgSvc);
            assertDefaultServerProperties(cfgSvc.getContainerConfiguration(), this.localServerPropertiesBuilder);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set</li>
     * <li>a sample based topology is provided where:
     * <ul>
     * <li>A JNDI configuration is provided</li>
     * <li>with only the JNDI factory set</li>
     * </ul>
     * </li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs about the missing JNDI Provider URL
     * </p>
     */
    @Test(expected = InconsistentDomainJndiProviderUrlException.class)
    public void testMissingJndiProviderUrl() throws ConfigurationException, InvocationTargetException,
            PropertiesException, IOException {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        try {
            final Topology topology = ConfigurationTestUtils.createBasicSampleTopology();
            final Jndi jndi = new Jndi();
            topology.getDomain().setJndi(jndi);
            jndi.setFactory(DOMAIN_JNDI_FACTORY);

            cfgSvc.loadConfiguration(serverProps, topology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set</li>
     * <li>a sample based topology is provided where:
     * <ul>
     * <li>A JNDI configuration is provided</li>
     * <li>with only the JNDI provider URL set</li>
     * </ul>
     * </li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs about the missing JNDI Provider URL
     * </p>
     */
    @Test(expected = InconsistentDomainJndiFactoryException.class)
    public void testMissingJndiFactory() throws ConfigurationException, InvocationTargetException, PropertiesException,
            IOException {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        try {
            final Topology topology = ConfigurationTestUtils.createBasicSampleTopology();
            final Jndi jndi = new Jndi();
            topology.getDomain().setJndi(jndi);
            jndi.setProviderUrl(DOMAIN_JNDI_PROVIDER_URL);

            cfgSvc.loadConfiguration(serverProps, topology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set</li>
     * <li>a sample based topology is provided where:
     * <ul>
     * <li>a container is defined,</li>
     * <li>but does not match the expected container name</li>
     * </ul>
     * </li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs about the container name not found
     * </p>
     */
    @Test(expected = InconsistentContainerNameException.class)
    public void testContainerNameDoesNotMatch() throws ConfigurationException, InvocationTargetException,
            PropertiesException, IOException {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties("unexisting-container-name-in-topology");

        try {
            final Topology topology = ConfigurationTestUtils.createBasicSampleTopology();

            cfgSvc.loadConfiguration(serverProps, topology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set</li>
     * <li>a sample based topology is provided where:
     * <ul>
     * <li>a second container with the same name is defined</li>
     * </ul>
     * </li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs about the duplicated container name
     * </p>
     */
    @Test(expected = ContainerNameDuplicatedException.class)
    public void testDuplicatedContainerName() throws Exception {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        try {
            final Topology topology = ConfigurationTestUtils.createBasicSampleTopology();

            final Container container0 = new Container();
            topology.getDomain().getContainer().add(container0);
            container0.setName(ConfigurationTestUtils.CONTAINER_NAME_0);
            container0.setDescription(ConfigurationTestUtils.CONTAINER_DESCRIPTION_0);
            container0.setHost(ConfigurationTestUtils.CONTAINER_HOST);
            container0.setUser(ConfigurationTestUtils.CONTAINER_JMX_USER);
            container0.setPassword(ConfigurationTestUtils.CONTAINER_JMX_PWD);

            cfgSvc.loadConfiguration(serverProps, topology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set</li>
     * <li>a sample based topology is provided where:
     * <ul>
     * <li>a network port is duplicated inside a container</li>
     * </ul>
     * </li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs about the duplicated network port
     * </p>
     */
    @Test(expected = InconsistentContainerDuplicatedPortException.class)
    public void testDuplicatedContainerPortWithOnlyOneContainer() throws Exception {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        try {
            final Topology topology = ConfigurationTestUtils.createBasicSampleTopology();

            final Container container0 = topology.getDomain().getContainer().get(0);
            container0.getJmxService().setRmiPort(ConfigurationTestUtils.CONTAINER_TCP_PORT_0);
            container0.getTransportService().setTcpPort(ConfigurationTestUtils.CONTAINER_TCP_PORT_0);

            cfgSvc.loadConfiguration(serverProps, topology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set</li>
     * <li>a sample based topology is provided where:
     * <ul>
     * <li>a network port is duplicated over two container</li>
     * </ul>
     * </li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs about the duplicated network port
     * </p>
     */
    @Test(expected = InconsistentContainerDuplicatedPortException.class)
    public void testDuplicatedContainerPortOverTwoContainers() throws Exception {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        try {
            final Topology topology = ConfigurationTestUtils.createBasicSampleTopology();

            final Container container1 = new Container();
            topology.getDomain().getContainer().add(container1);
            container1.setName("another-container-name");
            container1.setDescription(ConfigurationTestUtils.CONTAINER_DESCRIPTION_0);
            container1.setHost(ConfigurationTestUtils.CONTAINER_HOST);
            container1.setUser(ConfigurationTestUtils.CONTAINER_JMX_USER);
            container1.setPassword(ConfigurationTestUtils.CONTAINER_JMX_PWD);

            final JmxService jmxService1 = new JmxService();
            container1.setJmxService(jmxService1);

            final TransportService transportService1 = new TransportService();
            transportService1.setTcpPort(ConfigurationTestUtils.CONTAINER_TCP_PORT_0);
            container1.setTransportService(transportService1);

            cfgSvc.loadConfiguration(serverProps, topology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set</li>
     * <li>a sample based topology is provided where:
     * <ul>
     * <li>a second container without transport service is defined</li>
     * </ul>
     * </li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs about the missing transport service
     * </p>
     */
    @Test(expected = InconsistentContainerTransportSvcException.class)
    public void testMissingTransportService() throws Exception {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        try {
            final Topology topology = ConfigurationTestUtils.createBasicSampleTopology();

            final Container container1 = new Container();
            topology.getDomain().getContainer().add(container1);
            container1.setName("another-container-name");
            container1.setDescription(ConfigurationTestUtils.CONTAINER_DESCRIPTION_0);
            container1.setHost(ConfigurationTestUtils.CONTAINER_HOST);
            container1.setUser(ConfigurationTestUtils.CONTAINER_JMX_USER);
            container1.setPassword(ConfigurationTestUtils.CONTAINER_JMX_PWD);

            final JmxService jmxService1 = new JmxService();
            container1.setJmxService(jmxService1);

            cfgSvc.loadConfiguration(serverProps, topology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set</li>
     * <li>a sample based topology is provided where:
     * <ul>
     * <li>the container host is missing</li>
     * </ul>
     * </li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs about the missing container host
     * </p>
     */
    @Test(expected = InconsistentContainerHostException.class)
    public void testContainerWithoutHost() throws ConfigurationException, InvocationTargetException,
            PropertiesException, IOException {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        try {
            final Topology topology = ConfigurationTestUtils.createBasicSampleTopology();

            final Container container = topology.getDomain().getContainer().get(0);
            container.setHost(null);

            cfgSvc.loadConfiguration(serverProps, topology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set, except the
     * {@value ConfigurationProperties#WORKING_AREA_PATH_PROPERTY_NAME} that is set to an invalid value (it is a file),</li>
     * <li>a sample topology is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs about the invalid working area detected checking the existence of a persisted
     * topology definition.
     * </p>
     */
    @Test(expected = ConfigurationException.class)
    public void testInvalidWorkingArea() throws Exception {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        final File invalidWorkingArea = this.temporaryFolder.newFile("invalid-working-area");
        serverProps.setProperty(ConfigurationProperties.WORKING_AREA_PATH_PROPERTY_NAME,
                invalidWorkingArea.getAbsolutePath());

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

        } catch (final ConfigurationException e) {
            assertTrue("Unexpected configuration exception",
                    e.getMessage().contains(invalidWorkingArea.getAbsolutePath()));
            throw e;
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set, except the
     * {@value ConfigurationProperties#WORKING_AREA_PATH_PROPERTY_NAME} that is set to an unexisting directory,</li>
     * <li>a sample topology is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>No error occurs</li>
     * <li>Topology correctly set.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testUnexistingWorkingArea() throws Exception {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        final File unexistingWorkingAreaRoot = this.temporaryFolder.newFolder("unexisting-working-area-root");
        final File unexistingWorkingArea = new File(unexistingWorkingAreaRoot, "unexisting-working-area");
        serverProps.setProperty(ConfigurationProperties.WORKING_AREA_PATH_PROPERTY_NAME,
                unexistingWorkingArea.getAbsolutePath());

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            assertNotNull("Initial local topology null", cfgSvc.getLocalTopology());
            assertBasicSampleTopology(cfgSvc);
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set,</li>
     * <li>a sample topology is provided</li>
     * <li>a persisted topology definition exists and is invalid (it is a directory)</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs about the invalid persisted topology definition detected checking the existence
     * of a persisted topology definition.
     * </p>
     */
    @Test(expected = ConfigurationException.class)
    public void testInvalidPersistedTopologyFile() throws Exception {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        final File workingArea = this.temporaryFolder.newFolder("working-area");
        final File invalidPersistedTopologyFile = new File(workingArea, "topology.xml");
        assertTrue("Unable to create the invalid persisted topology file", invalidPersistedTopologyFile.mkdir());
        serverProps.setProperty(ConfigurationProperties.WORKING_AREA_PATH_PROPERTY_NAME, workingArea.getAbsolutePath());

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

        } catch (final ConfigurationException e) {
            assertTrue("Unexpected configuration exception",
                    e.getMessage().contains(invalidPersistedTopologyFile.getAbsolutePath()));
            throw e;
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the loading of configuration against default values:
     * <ul>
     * <li>minimal server properties are provided, all required properties correctly set,</li>
     * <li>a sample topology is provided</li>
     * <li>a persisted topology definition exists, containing a different topology definition</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>No error occurs</li>
     * <li>The topology definition correctly set is the persisted topology definition.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testPersistedTopologyFile() throws Exception {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        final File workingArea = this.temporaryFolder.newFolder("working-area");
        final File persistedTopologyFile = new File(workingArea, "topology.xml");
        assertTrue("Unable to create the persisted topology file", persistedTopologyFile.createNewFile());
        final Topology persistedTopology = ConfigurationTestUtils.createBasicSampleTopology();
        final Domain persisteDomain = persistedTopology.getDomain();
        assertNotNull("No domain in persisted topology definition", persisteDomain);
        persisteDomain.setName(PERSISTED_DOMAIN_NAME);
        TopologyBuilder.marshallTopologyFile(persistedTopologyFile, persistedTopology);

        serverProps.setProperty(ConfigurationProperties.WORKING_AREA_PATH_PROPERTY_NAME, workingArea.getAbsolutePath());

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            final Topology loadedTopology = cfgSvc.getLocalTopology();
            assertNotNull("Initial local topology null", loadedTopology);
            assertEquals("Unexpected domain name", PERSISTED_DOMAIN_NAME, loadedTopology.getDomain().getName());
            // The current container should always exists in the persisted topology.
            assertEquals("The current container name not found", ConfigurationTestUtils.CONTAINER_NAME_0,
                    loadedTopology.getDomain().getContainer().get(0).getName());

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

}
