/**
 * Copyright (c) 2014-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.configuration;

import static org.junit.Assert.assertEquals;

import java.io.File;
import java.util.Properties;

import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.configuration.DomainConfiguration;
import org.ow2.petals.microkernel.api.configuration.DomainConfiguration.DomainMode;
import org.ow2.petals.topology.TopologyBuilder;
import org.ow2.petals.topology.generated.Domain;
import org.ow2.petals.topology.generated.Topology;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

/**
 * Unit tests of {@link ConfigurationServiceImpl#reconfigureDomain(DomainConfiguration)}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ConfigurationServiceImpl_ReconfigureDomainTest {

    private final static String CONTAINER_CONFIG_FIELD = "containerConfiguration";

    private final static String DOMAIN_NAME_1 = "domain-1";

    private final static String DOMAIN_DESCRIPTION_1 = "A sample domain 1 configuration";

    @Rule
    public final LocalServerPropertiesBuilder localServerPropertiesBuilder = new LocalServerPropertiesBuilder();

    @Rule
    public final TemporaryFolder temporaryFolder = new TemporaryFolder();

    /**
     * <p>
     * Check the reconfiguration of a domain against:
     * <ul>
     * <li>all is ok</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>No error occurs.</li>
     * <li>the new domain name is set in the current domain configuration</li>
     * <li>the new domain description is set in the current domain configuration</li>
     * </ul>
     * </p>
     */
    @Test
    public void testAllIsOk() throws Exception {

        final File workingArea = this.temporaryFolder.newFolder("working-area");

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        final ContainerConfiguration containerConfig = new ContainerConfiguration();
        containerConfig.setWorkDirectoryPath(workingArea.getAbsolutePath());
        ReflectionHelper.setPrivateField(cfgSvc, CONTAINER_CONFIG_FIELD, containerConfig);
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        try {
            // let's make it different from the default basic one
            final Topology topology = ConfigurationTestUtils.createBasicSampleTopology();
            final Domain oldDomain = topology.getDomain();
            oldDomain.setName(DOMAIN_NAME_1);
            oldDomain.setDescription(DOMAIN_DESCRIPTION_1);
            oldDomain.setMode(org.ow2.petals.topology.generated.DomainMode.DYNAMIC);

            cfgSvc.loadConfiguration(serverProps, topology);

            final Topology newTopology = ConfigurationTestUtils.createBasicSampleTopology();

            cfgSvc.reconfigureDomain(newTopology);
            
            final DomainConfiguration reconfiguredDomain = cfgSvc.getDomainConfiguration();
            assertEquals("Unreconfigured domain name", ConfigurationTestUtils.DOMAIN_NAME,
                    reconfiguredDomain.getName());
            assertEquals("Unreconfigured domain description", ConfigurationTestUtils.DOMAIN_DESCRIPTION,
                    reconfiguredDomain.getDescription());
            assertEquals("Unreconfigured domain mode", DomainMode.STATIC, reconfiguredDomain.getMode());

            final File persistedTopologyFile = new File(workingArea, "topology.xml");
            final Topology readTopology = TopologyBuilder.createTopology(persistedTopologyFile);
            ConfigurationTestUtils.assertBasicSampleTopology(readTopology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }
}
