/**
 * Copyright (c) 2014-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.configuration;

import static org.junit.Assert.assertEquals;

import java.util.Properties;

import org.junit.Rule;
import org.junit.Test;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.configuration.DomainConfiguration;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentDomainJndiFactoryException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentDomainJndiProviderUrlException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentDomainModeException;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.Domain;
import org.ow2.petals.topology.generated.DomainMode;
import org.ow2.petals.topology.generated.JmxService;
import org.ow2.petals.topology.generated.Jndi;
import org.ow2.petals.topology.generated.Topology;
import org.ow2.petals.topology.generated.TransportService;

/**
 * Unit tests of {@link ConfigurationServiceImpl#updateTopology(Topology)}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ConfigurationServiceImpl_UpdateTopologyTest extends AbstractConfigurationServiceImplTest {

    private final static String DOMAIN_JNDI_FACTORY = "domain.jndi.factory";

    private final static String DOMAIN_JNDI_PROVIDER_URL = "http://domain.jndi.provider.url";

    private final static String CONTAINER_NAME_1 = "sample-1";

    private final static String CONTAINER_DESCRIPTION_1 = "description of the container 1";

    @Rule
    public final LocalServerPropertiesBuilder localServerPropertiesBuilder = new LocalServerPropertiesBuilder();

    /**
     * <p>
     * Check the update of the current topology against:
     * <ul>
     * <li>the new topology is the same than the current one</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>No error occurs</li>.
     * <li>The current topology is always the same.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testSameTopology() throws Exception {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);


        final Topology expectedTopology = ConfigurationTestUtils.createBasicSampleTopology();

        try {
            cfgSvc.loadConfiguration(serverProps, expectedTopology);
            assertBasicSampleTopology(cfgSvc);

            cfgSvc.updateTopology(expectedTopology, true);
            assertBasicSampleTopology(cfgSvc);
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the update of the current topology against:
     * <ul>
     * <li>the domain mode is changed by the new topology</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs because the domain mode is changed.
     * </p>
     */
    @Test(expected = InconsistentDomainModeException.class)
    public void testDomainModeChanged() throws Exception {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();

            cfgSvc.loadConfiguration(serverProps, initialTopology);
            assertBasicSampleTopology(cfgSvc);

            final Topology finalTopology = ConfigurationTestUtils.createBasicSampleTopology();
            finalTopology.getDomain().setMode(DomainMode.DYNAMIC);

            cfgSvc.updateTopology(finalTopology, true);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the update of the current topology against:
     * <ul>
     * <li>the jndi factory of the domain is changed by the new topology</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs because the jndi factory of the domain is changed.
     * </p>
     */
    @Test(expected = InconsistentDomainJndiFactoryException.class)
    public void testDomainJndiFactoryChanged() throws Exception {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();
            final Jndi initialJndi = new Jndi();
            initialTopology.getDomain().setJndi(initialJndi);
            initialJndi.setFactory(DOMAIN_JNDI_FACTORY);
            initialJndi.setProviderUrl(DOMAIN_JNDI_PROVIDER_URL);

            cfgSvc.loadConfiguration(serverProps, initialTopology);
            assertBasicSampleTopology(cfgSvc);

            final Topology finalTopology = ConfigurationTestUtils.createBasicSampleTopology();
            final Jndi jndi = new Jndi();
            finalTopology.getDomain().setJndi(jndi);
            jndi.setFactory("another." + DOMAIN_JNDI_FACTORY);
            jndi.setProviderUrl(DOMAIN_JNDI_PROVIDER_URL);

            cfgSvc.updateTopology(finalTopology, true);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the update of the current topology against:
     * <ul>
     * <li>the jndi provider url of the domain is changed by the new topology</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs because the jndi provider url of the domain is changed.
     * </p>
     */
    @Test(expected = InconsistentDomainJndiProviderUrlException.class)
    public void testDomainJndiProviderUrlChanged() throws Exception {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();
            final Jndi initialJndi = new Jndi();
            initialTopology.getDomain().setJndi(initialJndi);
            initialJndi.setFactory(DOMAIN_JNDI_FACTORY);
            initialJndi.setProviderUrl(DOMAIN_JNDI_PROVIDER_URL);

            cfgSvc.loadConfiguration(serverProps, initialTopology);
            assertBasicSampleTopology(cfgSvc);

            final Topology finalTopology = ConfigurationTestUtils.createBasicSampleTopology();
            final Jndi jndi = new Jndi();
            finalTopology.getDomain().setJndi(jndi);
            jndi.setFactory(DOMAIN_JNDI_FACTORY);
            jndi.setProviderUrl(DOMAIN_JNDI_PROVIDER_URL + "/another");

            cfgSvc.updateTopology(finalTopology, true);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the update of the current topology against:
     * <ul>
     * <li>the topology that will be updated contains only one container located in one domain,</li>
     * <li>the new topology will:
     * <ul>
     * <li>add a new container in the existing domain,</li>
     * </ul>
     * </li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>No error occurs,</li>
     * <li>the new configuration (ie. topology) is composed of:
     * <ul>
     * <li>one domain (the initial one) with 2 containers (the initial one completed by a new one)</li>
     * </ul>
     * </li>
     * </p>
     */
    @Test
    public void test1NewContainerPlus1DomainWith1Container() throws Exception {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();

            cfgSvc.loadConfiguration(serverProps, initialTopology);
            assertBasicSampleTopology(cfgSvc);

            final Topology finalTopology = ConfigurationTestUtils.createBasicSampleTopology();
            final Domain finalDomain = finalTopology.getDomain();

            final Container container0_1 = new Container();
            finalDomain.getContainer().add(container0_1);
            container0_1.setName(CONTAINER_NAME_1);
            container0_1.setDescription(CONTAINER_DESCRIPTION_1);
            container0_1.setHost(ConfigurationTestUtils.CONTAINER_HOST);
            container0_1.setUser(ConfigurationTestUtils.CONTAINER_JMX_USER);
            container0_1.setPassword(ConfigurationTestUtils.CONTAINER_JMX_PWD);

            final JmxService jmxService0_1 = new JmxService();
            container0_1.setJmxService(jmxService0_1);

            final TransportService transportService0_1 = new TransportService();
            transportService0_1.setTcpPort(7781);
            container0_1.setTransportService(transportService0_1);

            cfgSvc.updateTopology(finalTopology, false);

            final DomainConfiguration domainConfig = cfgSvc.getDomainConfiguration();
            assertEquals("Unexpected domain name", ConfigurationTestUtils.DOMAIN_NAME, domainConfig.getName());
            assertEquals("Unexpected domain description", ConfigurationTestUtils.DOMAIN_DESCRIPTION,
                    domainConfig.getDescription());
            assertEquals("Unexpected domain mode", DomainConfiguration.DomainMode.STATIC, domainConfig.getMode());

            assertEquals("Unexpected number of container", 2, cfgSvc.getContainersConfiguration().size());
            final ContainerConfiguration containerConfig = cfgSvc.getContainerConfiguration();
            assertEquals("Unexpected container name", ConfigurationTestUtils.CONTAINER_NAME_0, containerConfig.getName());
            assertEquals("Unexpected container description", ConfigurationTestUtils.CONTAINER_DESCRIPTION_0,
                    containerConfig.getDescription());
            assertEquals("Unexpected container host", ConfigurationTestUtils.CONTAINER_HOST, containerConfig.getHost());
            assertEquals("Unexpected container jmx user", ConfigurationTestUtils.CONTAINER_JMX_USER,
                    containerConfig.getUser());
            assertEquals("Unexpected container jmx pwd", ConfigurationTestUtils.CONTAINER_JMX_PWD,
                    containerConfig.getPassword());

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

}
