/**
 * Copyright (c) 2013-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.extension;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;
import org.objectweb.fractal.api.Component;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.extension.PetalsExtensionController;
import org.ow2.petals.microkernel.api.extension.exception.CreationExtensionException;
import org.ow2.petals.microkernel.api.extension.exception.NotSatisfiedDependenciesException;
import org.ow2.petals.microkernel.api.extension.exception.PetalsExtensionException;

/**
 * Unit tests of {@link PostExtensionsManagerImpl}.
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ExtensionManagerUtilsTest {

    private class ExtensionController implements PetalsExtensionController {

        private String name;

        private String[] dependencies;

        private boolean isActivated;

        public ExtensionController(final String name, final String[] dependencies) {
            this(name, dependencies, false);
        }

        public ExtensionController(final String name, final String[] dependencies,
                final boolean isActivated) {
            super();
            this.name = name;
            this.dependencies = dependencies;
            this.isActivated = isActivated;
        }

        @Override
        public Component createFractalComponent(Component extensionComposite)
                throws CreationExtensionException {
            return null;
        }

        @Override
        public void removeFractalComponent() {
            // NOP
        }

        @Override
        public String getExtensionName() {
            return this.name;
        }

        @Override
        public String getExtensionVersion() {
            return null;
        }

        @Override
        public boolean isActivated(final ConfigurationService configurationService) throws PetalsExtensionException {
            return this.isActivated;
        }

        @Override
        public String[] getDependencies() {
            return this.dependencies;
        }

        @Override
        public boolean failsPetalsStartup() {
            return false;
        }

    }

    @Rule
    public ExpectedException expectedEx = ExpectedException.none();

    /**
     * Check the processing of
     * {@link PostExtensionsManagerImpl#sortExtensionControllers} against a simple
     * list of extensions already sorted.
     */
    @Test
    public void testSortExtensionControllers_00() throws NotSatisfiedDependenciesException {
        
        final List<PetalsExtensionController> extensionControllers = new ArrayList<PetalsExtensionController>();
        final String unitTest1 = "Unit Test 1";
        final String unitTest2 = "Unit Test 2";
        extensionControllers.add(new ExtensionController(unitTest1, null));
        extensionControllers.add(new ExtensionController(unitTest2, new String[] { unitTest1 } ));

        final List<PetalsExtensionController> sortedExtensionController = ExtensionManagerUtils
                .sortExtensionControllers(extensionControllers);
        assertNotNull("Null result", sortedExtensionController);
        assertEquals("Unexpected size", 2, sortedExtensionController.size());
        assertEquals("Unexpected first element", unitTest1, sortedExtensionController.get(0)
                .getExtensionName());
        assertEquals("Unexpected 2nd element", unitTest2, sortedExtensionController.get(1)
                .getExtensionName());
    }

    /**
     * Check the processing of {@link PostExtensionsManagerImpl#sortExtensionControllers} against a simple list of
     * extensions not sorted.
     */
    @Test
    public void testSortExtensionControllers_01() throws NotSatisfiedDependenciesException {

        final List<PetalsExtensionController> extensionControllers = new ArrayList<PetalsExtensionController>();
        final String unitTest1 = "Unit Test 1";
        final String unitTest2 = "Unit Test 2";
        extensionControllers.add(new ExtensionController(unitTest2, new String[] { unitTest1 }));
        extensionControllers.add(new ExtensionController(unitTest1, null));

        final List<PetalsExtensionController> sortedExtensionController = ExtensionManagerUtils
                .sortExtensionControllers(extensionControllers);
        assertNotNull("Null result", sortedExtensionController);
        assertEquals("Unexpected size", 2, sortedExtensionController.size());
        assertEquals("Unexpected first element", unitTest1, sortedExtensionController.get(0)
                .getExtensionName());
        assertEquals("Unexpected 2nd element", unitTest2, sortedExtensionController.get(1)
                .getExtensionName());
    }

    /**
     * Check the processing of
     * {@link PostExtensionsManagerImpl#sortExtensionControllers} against a list of
     * extensions not sorted.
     */
    @Test
    public void testSortExtensionControllers_02() throws NotSatisfiedDependenciesException {

        final List<PetalsExtensionController> extensionControllers = new ArrayList<PetalsExtensionController>();
        final String unitTest1 = "Unit Test 1";
        final String unitTest2 = "Unit Test 2";
        final String unitTest3 = "Unit Test 3";
        extensionControllers.add(new ExtensionController(unitTest2, new String[] { unitTest1 }));
        extensionControllers.add(new ExtensionController(unitTest1, null));
        extensionControllers.add(new ExtensionController(unitTest3, new String[] { unitTest1,
                unitTest2 }));

        final List<PetalsExtensionController> sortedExtensionController = ExtensionManagerUtils
                .sortExtensionControllers(extensionControllers);
        assertNotNull("Null result", sortedExtensionController);
        assertEquals("Unexpected size", 3, sortedExtensionController.size());
        assertEquals("Unexpected first element", unitTest1, sortedExtensionController.get(0)
                .getExtensionName());
        assertEquals("Unexpected 2nd element", unitTest2, sortedExtensionController.get(1)
                .getExtensionName());
        assertEquals("Unexpected 3rd element", unitTest3, sortedExtensionController.get(2)
                .getExtensionName());
    }

    /**
     * Check the processing of
     * {@link PostExtensionsManagerImpl#sortExtensionControllers} against a list of
     * extensions not sorted with a not satisfied dependency.
     * 
     */
    @Test
    public void testSortExtensionControllers_03() throws NotSatisfiedDependenciesException {

        this.expectedEx.expect(NotSatisfiedDependenciesException.class);
        // TODO: Add a check on the content of
        // NotSatisfiedDependenciesException.extensionControllers

        final List<PetalsExtensionController> extensionControllers = new ArrayList<PetalsExtensionController>();
        final String unitTest1 = "Unit Test 1";
        final String unitTest2 = "Unit Test 2";
        final String unitTest3 = "Unit Test 3";
        final String unitTest4 = "Unit Test 4";
        final String unitTest5 = "Unit Test 5";
        extensionControllers.add(new ExtensionController(unitTest2, new String[] { unitTest1 }));
        extensionControllers.add(new ExtensionController(unitTest1, null));
        extensionControllers.add(new ExtensionController(unitTest3, new String[] { unitTest1,
                unitTest2 }));
        extensionControllers.add(new ExtensionController(unitTest5, new String[] { unitTest4 }));

        ExtensionManagerUtils.sortExtensionControllers(extensionControllers);
    }

    /**
     * <p>
     * Check the processing of
     * {@link PostExtensionsManagerImpl#isAllDependenciesLoaded} against a extension
     * having no dependency.
     * </p>
     * <p>
     * Expected result: <code>true</code>
     * </p>
     * 
     */
    @Test
    public void testIsAllDependenciesLoaded_00() {

        final String unitTest1 = "Unit Test 1";
        final PetalsExtensionController pec1 = new ExtensionController(unitTest1, null, true);

        assertEquals("The extension should be activated", true, ExtensionManagerUtils.isAllDependenciesLoaded(pec1,
                new HashMap<PetalsExtensionController, Component>()));
    }

    /**
     * <p>
     * Check the processing of
     * {@link PostExtensionsManagerImpl#isAllDependenciesLoaded} against a extension
     * having a dependency not activated.
     * </p>
     * <p>
     * Expected result: <code>false</code>
     * </p>
     * 
     */
    @Test
    public void testIsAllDependenciesLoaded_01() {

        final String unitTest1 = "Unit Test 1";
        final String unitTest2 = "Unit Test 2";
        final PetalsExtensionController pec2 = new ExtensionController(unitTest2,
                new String[] { unitTest1 }, true);

        // No extension controller is loaded when creating the extension
        // manager, so the extension controller of 'unitTest1' is not loaded.

        assertEquals("The extension should not be activated", false, ExtensionManagerUtils.isAllDependenciesLoaded(
                pec2, new HashMap<PetalsExtensionController, Component>()));
    }

    /**
     * <p>
     * Check the processing of
     * {@link PostExtensionsManagerImpl#isAllDependenciesLoaded} against a extension
     * having a dependency not activated and another activated.
     * </p>
     * <p>
     * Expected result: <code>false</code>
     * </p>
     * 
     */
    @Test
    public void testIsAllDependenciesLoaded_02() throws InvocationTargetException {

        final String unitTest1 = "Unit Test 1";
        final String unitTest2 = "Unit Test 2";
        final PetalsExtensionController pec2 = new ExtensionController(unitTest2, new String[] {},
                true);
        final String unitTest3 = "Unit Test 3";
        final PetalsExtensionController pec3 = new ExtensionController(unitTest3, new String[] {
                unitTest1, unitTest2 }, true);

        final Map<PetalsExtensionController, Component> extensionControllers = new HashMap<PetalsExtensionController, Component>();
        // The extension 'unitTest1' is not loaded so not activated
        extensionControllers.put(pec2, null);

        assertEquals("The extension should not be activated", false,
                ExtensionManagerUtils.isAllDependenciesLoaded(pec3, extensionControllers));
    }

}
