/**
 * Copyright (c) 2014-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.admin;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.io.IOException;
import java.net.URL;
import java.util.Enumeration;

import org.junit.Test;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.util.Fractal;
import org.ow2.petals.jbi.descriptor.original.generated.ComponentType;
import org.ow2.petals.jbi.descriptor.original.generated.Identification;
import org.ow2.petals.microkernel.api.container.ContainerService;
import org.ow2.petals.microkernel.api.container.exception.ContainerServiceException;
import org.ow2.petals.microkernel.api.extension.PetalsPostExtensionController;
import org.ow2.petals.microkernel.api.extension.PetalsPreExtensionController;
import org.ow2.petals.microkernel.api.jbi.management.AdminService;
import org.ow2.petals.microkernel.api.server.FractalHelper;
import org.ow2.petals.microkernel.extension.utest.InstallationExtensionCtrl;
import org.ow2.petals.microkernel.extension.utest.PostExtensionController;
import org.ow2.petals.microkernel.extension.utest.PreExtensionController;

/**
 * Unit tests of {@link AdminServiceImpl}.
 * 
 * @author Christophe DENEUX - Linagora
 */
public class AdminServiceImplTest {

    /**
     * Name of the Fractal root composite
     */
    private final static String FRACTAL_ROOT_COMPOSITE_NAME = "org.ow2.petals.microkernel.jbi.management.admin.AdminServiceRoot";

    private Component getAdminCompositeStarted() throws NoSuchInterfaceException, IllegalLifeCycleException,
            InstantiationException {
        final Component adminComposite = FractalHelper.createCompositeComponent(FRACTAL_ROOT_COMPOSITE_NAME);
        FractalHelper.startComponent(adminComposite);

        return adminComposite;
    }

    private ContainerService getContainerService(final Component adminComposite) throws NoSuchInterfaceException {
        final ContentController adminCompositeCC = Fractal.getContentController(adminComposite);
        final Component containerServiceComp = FractalHelper.getRecursiveComponentByName(adminCompositeCC,
                FractalHelper.CONTAINER_SERVICE_COMPONENT);
        assertNotNull("Fractal component 'ContainerService' not found", containerServiceComp);
        final ContainerService containerService = (ContainerService) containerServiceComp
                .getFcInterface(ContainerService.FRACTAL_SRV_ITF_NAME);
        assertNotNull("Server interface of Fractal component 'ContainerService' not found", containerService);
        return containerService;
    }

    private AdminService getAdminService(final Component adminComposite) throws NoSuchInterfaceException {
        final ContentController adminCompositeCC = Fractal.getContentController(adminComposite);
        final Component adminServiceComp = FractalHelper.getRecursiveComponentByName(adminCompositeCC,
                FractalHelper.ADMIN_COMPONENT);
        assertNotNull("Fractal component 'AdminService' not found", adminServiceComp);
        final AdminService adminService = (AdminService) adminServiceComp.getFcInterface("service");
        assertNotNull("Server interface of Fractal component 'ContainerService' not found", adminService);
        return adminService;
    }

    /**
     * <p>
     * Check {@link ContainerServiceImpl#isBindingComponent(String)} when no installer was loaded.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occurs,</li>
     * <li>the value <code>false</code> is returned.</li>
     * </ul>
     * </p>
     */
    @Test
    public void IsBinding_NoInstallerLoaded() throws NoSuchInterfaceException, IllegalLifeCycleException,
            InstantiationException {

        final AdminService adminService = this.getAdminService(this.getAdminCompositeStarted());
        assertFalse(adminService.isBinding("no-installer-loaded"));
    }


    /**
     * <p>
     * Check {@link ContainerServiceImpl#isBindingComponent(String)} when no installer was loaded for the requested
     * component, but other component installers are loaded.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occurs,</li>
     * <li>the value <code>false</code> is returned.</li>
     * </ul>
     * </p>
     */
    @Test
    public void IsBinding_NoInstallerLoadedForRequestedComponent() throws NoSuchInterfaceException,
            IllegalLifeCycleException, InstantiationException, ContainerServiceException {

        final Component adminComposite = this.getAdminCompositeStarted();
        final org.ow2.petals.jbi.descriptor.original.generated.Component jbiComponent = new org.ow2.petals.jbi.descriptor.original.generated.Component();
        final Identification jbiIdentification = new Identification();
        jbiComponent.setIdentification(jbiIdentification);
        final String componentName = "a-component-loaded";
        jbiIdentification.setName(componentName);
        jbiComponent.setType(ComponentType.BINDING_COMPONENT);
        this.getContainerService(adminComposite).createInstaller(jbiComponent);

        final AdminService adminService = this.getAdminService(adminComposite);
        assertFalse(adminService.isBinding("no-installer-loaded"));
    }

    /**
     * <p>
     * Check {@link ContainerServiceImpl#isBindingComponent(String)} when:
     * <ul>
     * <li>only the component installer is loaded,</li>
     * <li>the component is a binding component.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occurs,</li>
     * <li>the value <code>true</code> is returned.</li>
     * </ul>
     * </p>
     */
    @Test
    public void IsBinding_InstallerLoaded() throws NoSuchInterfaceException, IllegalLifeCycleException,
            InstantiationException, ContainerServiceException {

        final Component adminComposite = this.getAdminCompositeStarted();
        final org.ow2.petals.jbi.descriptor.original.generated.Component jbiComponent = new org.ow2.petals.jbi.descriptor.original.generated.Component();
        final Identification jbiIdentification = new Identification();
        jbiComponent.setIdentification(jbiIdentification);
        final String componentName = "a-component-loaded";
        jbiIdentification.setName(componentName);
        jbiComponent.setType(ComponentType.BINDING_COMPONENT);
        this.getContainerService(adminComposite).createInstaller(jbiComponent);

        final AdminService adminService = this.getAdminService(adminComposite);
        assertTrue(adminService.isBinding(componentName));
    }

    /**
     * <p>
     * Check {@link ContainerServiceImpl#isBindingComponent(String)} when:
     * <ul>
     * <li>only the component installer is loaded,</li>
     * <li>the component is a service engine.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occurs,</li>
     * <li>the value <code>false</code> is returned.</li>
     * </ul>
     * </p>
     */
    @Test
    public void IsBinding_InstallerLoadedButServiceEngine() throws NoSuchInterfaceException, IllegalLifeCycleException,
            InstantiationException, ContainerServiceException {

        final Component adminComposite = this.getAdminCompositeStarted();
        final org.ow2.petals.jbi.descriptor.original.generated.Component jbiComponent = new org.ow2.petals.jbi.descriptor.original.generated.Component();
        final Identification jbiIdentification = new Identification();
        jbiComponent.setIdentification(jbiIdentification);
        final String componentName = "a-component-loaded";
        jbiIdentification.setName(componentName);
        jbiComponent.setType(ComponentType.SERVICE_ENGINE);
        this.getContainerService(adminComposite).createInstaller(jbiComponent);

        final AdminService adminService = this.getAdminService(adminComposite);
        assertFalse(adminService.isBinding(componentName));
    }

    /**
     * <p>
     * Check {@link ContainerServiceImpl#isServiceEngine(String)} when no installer was loaded.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occurs,</li>
     * <li>the value <code>false</code> is returned.</li>
     * </ul>
     * </p>
     */
    @Test
    public void IsEngine_NoInstallerLoaded() throws NoSuchInterfaceException, IllegalLifeCycleException,
            InstantiationException {

        final AdminService adminService = this.getAdminService(this.getAdminCompositeStarted());
        assertFalse(adminService.isEngine("no-installer-loaded"));
    }

    /**
     * <p>
     * Check {@link ContainerServiceImpl#isServiceEngine(String)} when no installer was loaded for the requested
     * component, but other component installers are loaded.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occurs,</li>
     * <li>the value <code>false</code> is returned.</li>
     * </ul>
     * </p>
     */
    @Test
    public void IsEngine_NoInstallerLoadedForRequestedComponent() throws NoSuchInterfaceException,
            IllegalLifeCycleException, InstantiationException, ContainerServiceException {

        final Component adminComposite = this.getAdminCompositeStarted();
        final org.ow2.petals.jbi.descriptor.original.generated.Component jbiComponent = new org.ow2.petals.jbi.descriptor.original.generated.Component();
        final Identification jbiIdentification = new Identification();
        jbiComponent.setIdentification(jbiIdentification);
        final String componentName = "a-component-loaded";
        jbiIdentification.setName(componentName);
        jbiComponent.setType(ComponentType.BINDING_COMPONENT);
        this.getContainerService(adminComposite).createInstaller(jbiComponent);

        final AdminService adminService = this.getAdminService(adminComposite);
        assertFalse(adminService.isEngine("no-installer-loaded"));
    }

    /**
     * <p>
     * Check {@link ContainerServiceImpl#isServiceEngine(String)} when:
     * <ul>
     * <li>only the component installer is loaded,</li>
     * <li>the component is a service engine.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occurs,</li>
     * <li>the value <code>true</code> is returned.</li>
     * </ul>
     * </p>
     */
    @Test
    public void IsEngine_InstallerLoaded() throws NoSuchInterfaceException, IllegalLifeCycleException,
            InstantiationException, ContainerServiceException {

        final Component adminComposite = this.getAdminCompositeStarted();
        final org.ow2.petals.jbi.descriptor.original.generated.Component jbiComponent = new org.ow2.petals.jbi.descriptor.original.generated.Component();
        final Identification jbiIdentification = new Identification();
        jbiComponent.setIdentification(jbiIdentification);
        final String componentName = "a-component-loaded";
        jbiIdentification.setName(componentName);
        jbiComponent.setType(ComponentType.SERVICE_ENGINE);
        this.getContainerService(adminComposite).createInstaller(jbiComponent);

        final AdminService adminService = this.getAdminService(adminComposite);
        assertTrue(adminService.isEngine(componentName));
    }

    /**
     * <p>
     * Check {@link ContainerServiceImpl#isServiceEngine(String)} when:
     * <ul>
     * <li>only the component installer is loaded,</li>
     * <li>the component is a service engine.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occurs,</li>
     * <li>the value <code>false</code> is returned.</li>
     * </ul>
     * </p>
     */
    @Test
    public void IsEngine_InstallerLoadedButBindingComponent() throws NoSuchInterfaceException,
            IllegalLifeCycleException, InstantiationException, ContainerServiceException {

        final Component adminComposite = this.getAdminCompositeStarted();
        final org.ow2.petals.jbi.descriptor.original.generated.Component jbiComponent = new org.ow2.petals.jbi.descriptor.original.generated.Component();
        final Identification jbiIdentification = new Identification();
        jbiComponent.setIdentification(jbiIdentification);
        final String componentName = "a-component-loaded";
        jbiIdentification.setName(componentName);
        jbiComponent.setType(ComponentType.BINDING_COMPONENT);
        this.getContainerService(adminComposite).createInstaller(jbiComponent);

        final AdminService adminService = this.getAdminService(adminComposite);
        assertFalse(adminService.isEngine(componentName));
    }

    /**
     * <p>
     * Check system information returned when:
     * <ul>
     * <li>a pre-extension is installed,</li>
     * <li>a post-extension is instammed.</li>
     * </ul>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the following information are displayed:
     * <ul>
     * <li>µKernel version,</li>
     * <li>registry client implementation and version,</li>
     * <li>pre-extension implementation and version,</li>
     * <li>post-extension implementation and version,</li>
     * <li>JVM runtime name, vendor and version,</li>
     * <li>operation system name, version and architecture.</li>
     * </ul>
     * </li>
     * <li>no information about extension are displayed,</li>
     * <li>no error occurs.</li>
     * </ul>
     * </p>
     * TODO: Check the versions displayed
     */
    @Test
    public void getSystemInfo() throws NoSuchInterfaceException, IllegalLifeCycleException,
            InstantiationException {

        final ClassLoader oldClassloader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(new ClassLoader(oldClassloader) {

            @Override
            public Enumeration<URL> getResources(final String name) throws IOException {
                if (name.equals("META-INF/services/" + PetalsPreExtensionController.class.getName())) {
                    return super.getResources("services/utest/" + PetalsPreExtensionController.class.getName());
                } else if (name.equals("META-INF/services/" + PetalsPostExtensionController.class.getName())) {
                    return super.getResources("services/utest/" + PetalsPostExtensionController.class.getName());
                } else {
                    return super.getResources(name);
                }
            }
        });

        try {
            final Component adminComposite = this.getAdminCompositeStarted();
            final AdminService adminService = this.getAdminService(adminComposite);

            final String versionsStr = adminService.getSystemInfo();
            assertNotNull(versionsStr);
            assertFalse("Empty information", versionsStr.isEmpty());
            final String[] versions = versionsStr.split(System.getProperty("line.separator"));
            System.out.println(versionsStr);
            assertEquals(7, versions.length);

            assertFalse(versions[0].isEmpty());
            assertFalse(versions[1].isEmpty());

            assertFalse(versions[2].isEmpty());
            assertTrue(versions[2].contains(PreExtensionController.class.getName()));

            assertFalse(versions[3].isEmpty());
            assertTrue(versions[3].contains(InstallationExtensionCtrl.class.getName()));

            assertFalse(versions[4].isEmpty());
            assertTrue(versions[4].contains(PostExtensionController.class.getName()));

            assertTrue(versions[5].contains(System.getProperty("java.runtime.name")));
            assertTrue(versions[5].contains(System.getProperty("java.runtime.version")));
            assertTrue(versions[5].contains(System.getProperty("java.vm.vendor")));

            assertTrue(versions[6].contains(System.getProperty("os.name")));
            assertTrue(versions[6].contains(System.getProperty("os.version")));
            assertTrue(versions[6].contains(System.getProperty("os.arch")));
        } finally {
            Thread.currentThread().setContextClassLoader(oldClassloader);
        }

    }

}
