/**
 * Copyright (c) 2011-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.task.download;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.logging.Logger;
import java.util.regex.Pattern;

import org.junit.After;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.contrib.java.lang.system.ProvideSystemProperty;
import org.junit.rules.TemporaryFolder;
import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.jbi.management.Context;
import org.ow2.petals.microkernel.jbi.management.task.DownloadTask;
import org.ow2.petals.microkernel.jbi.management.task.download.jetty.server.JettyServer;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * @author Christophe Deneux - EBM WebSourcing
 */
public class DownloadTaskTest {

    private static final String componentName = "test-contentType";

    private static final String DOWNLOADED_FILENAME_SUFFIX_REGEXP = "-downloaded-[0-9]{2}-[0-9]{2}-[0-9]{4}-[0-9]{6}-[0-9]*\\.zip";

    private JettyServer jettyServer = null;

    @Rule
    public TemporaryFolder testFolder = new TemporaryFolder();

    /**
     * Registration of PAX URL handler
     */
    @Rule
    public final ProvideSystemProperty provideSystemProperty = new ProvideSystemProperty("java.protocol.handler.pkgs",
            "org.ops4j.pax.url");

    @Before
    public void setUp() throws Exception {
        try {
            this.jettyServer = JettyServer.getInstance();
        } catch (final Exception e) {
            e.printStackTrace();
        }
    }

    @After
    public void tearDown() throws Exception {
        try {
            this.jettyServer.stop();
        } catch (final Exception e) {
            e.printStackTrace();
        }
    }

    @Test
    public void testDownloadEntityPackageWithCorrectContentType() throws Exception {

        final String baseFilename = "test-contentType";
        final String urlTest = "http://localhost:8085/CorrectContentType/" + baseFilename + ".zip";
        final Context ctx = createContext(urlTest);
        assertEquals(urlTest, ctx.getArchiveUrl().toString());

        final ContainerConfiguration containerConfigurationMock = new ContainerConfiguration();
        containerConfigurationMock.setWorkDirectoryPath(this.testFolder.newFolder("work").getAbsolutePath());

        final DownloadTask d = new DownloadTask(new LoggingUtil(Logger.getLogger("testLogger")),
                containerConfigurationMock);
        d.execute(ctx);

        // Assertions
        final File downloadedFile = ctx.getArchiveFile();
        assertTrue(Pattern.matches(baseFilename + DOWNLOADED_FILENAME_SUFFIX_REGEXP, downloadedFile.getName()));
        assertTrue(downloadedFile.exists());

    }

    @Test
    public void testDownloadEntityPackageWithWrongContentType() throws Exception {

        final String urlTest = "http://localhost:8085/WrongContentType/test-contentType.zip";
        final Context ctx = createContext(urlTest);
        assertEquals(urlTest, ctx.getArchiveUrl().toString());

        final ContainerConfiguration containerConfigurationMock = new ContainerConfiguration();
        containerConfigurationMock.setWorkDirectoryPath(this.testFolder.newFolder("work").getAbsolutePath());

        try {
            final DownloadTask d = new DownloadTask(new LoggingUtil(Logger.getLogger("testLogger")),
                    containerConfigurationMock);
            d.execute(ctx);
        } catch (PetalsException pE) {
            assertTrue(pE.getMessage().contains("http://localhost:8085/WrongContentType/test-contentType.zip"));
            assertTrue(pE.getMessage().contains("application/toto"));
        }

    }

    @Test
    public void testDownloadEntityPackageWithCorrectContentTypeWithCharset() throws Exception {

        final String baseFilename = "test-contentType";
        final String urlTest = "http://localhost:8085/CorrectContentTypeWithCharset/" + baseFilename + ".zip";
        final Context ctx = createContext(urlTest);
        assertEquals(urlTest, ctx.getArchiveUrl().toString());

        final ContainerConfiguration containerConfigurationMock = new ContainerConfiguration();
        containerConfigurationMock.setWorkDirectoryPath(this.testFolder.newFolder("work").getAbsolutePath());

        final DownloadTask d = new DownloadTask(new LoggingUtil(Logger.getLogger("testLogger")),
                containerConfigurationMock);
        d.execute(ctx);

        // Assertions
        final File downloadedFile = ctx.getArchiveFile();
        assertTrue(Pattern.matches(baseFilename + DOWNLOADED_FILENAME_SUFFIX_REGEXP, downloadedFile.getName()));
        assertTrue(downloadedFile.exists());

    }

    @Test
    public void testDownloadEntityPackageWithoutContentType() throws Exception {

        final String urlTest = "http://localhost:8085/WithoutContentType/test-contentType.zip";
        final Context ctx = createContext(urlTest);
        assertEquals(urlTest, ctx.getArchiveUrl().toString());

        final ContainerConfiguration containerConfigurationMock = new ContainerConfiguration();
        containerConfigurationMock.setWorkDirectoryPath(this.testFolder.newFolder("work").getAbsolutePath());

        try {
            final DownloadTask d = new DownloadTask(new LoggingUtil(Logger.getLogger("testLogger")),
                    containerConfigurationMock);

            d.execute(ctx);
        } catch (final PetalsException pE) {
            assertTrue(pE.getMessage().contains("http://localhost:8085/WithoutContentType/test-contentType.zip"));
            assertTrue(pE.getMessage().contains("It's content-type is undefined."));
        }

    }

    @Test
    public void testDownloadEntityPackageWithWrongURL() throws Exception {

        final String urlTest = "http://localhost:8085/WrongURL/test-contentType.zip";
        final Context ctx = createContext(urlTest);
        assertEquals(urlTest, ctx.getArchiveUrl().toString());

        final ContainerConfiguration containerConfigurationMock = new ContainerConfiguration();
        containerConfigurationMock.setWorkDirectoryPath(this.testFolder.newFolder("work").getAbsolutePath());

        try {
            final DownloadTask d = new DownloadTask(new LoggingUtil(Logger.getLogger("testLogger")),
                    containerConfigurationMock);
            d.execute(ctx);
        } catch (final PetalsException pE) {
            assertTrue(pE.getMessage().contains("http://localhost:8085/WrongURL/test-contentType.zip"));
            assertTrue(pE.getMessage().contains("HTTP 404"));
        }

    }

    @Test
    public void testDownloadMavenEntityPackage() throws Exception {

        final String artifactId = "petals-bc-soap";
        final String urlTest = "mvn:org.ow2.petals/petals-bc-soap/4.4.2/zip";
        final Context ctx = createContext(urlTest);
        assertEquals(urlTest, ctx.getArchiveUrl().toString());

        final ContainerConfiguration containerConfigurationMock = new ContainerConfiguration();
        containerConfigurationMock.setWorkDirectoryPath(this.testFolder.newFolder("work").getAbsolutePath());

        final DownloadTask d = new DownloadTask(new LoggingUtil(Logger.getLogger("testLogger")),
                containerConfigurationMock);
        d.execute(ctx);

        // Assertions
        final File downloadedFile = ctx.getArchiveFile();
        assertTrue(Pattern.matches(artifactId + DOWNLOADED_FILENAME_SUFFIX_REGEXP, downloadedFile.getName()));
        assertTrue(downloadedFile.exists());

    }

    @Test
    public void testDownloadMavenEntityPackageWithRepository() throws Exception {

        final String artifactId = "petals-bc-soap";
        final String urlTest = "mvn:http://repo.maven.apache.org/maven2!org.ow2.petals/petals-bc-soap/4.4.2/zip";
        final Context ctx = createContext(urlTest);
        assertEquals(urlTest, ctx.getArchiveUrl().toString());

        final ContainerConfiguration containerConfigurationMock = new ContainerConfiguration();
        containerConfigurationMock.setWorkDirectoryPath(this.testFolder.newFolder("work").getAbsolutePath());

        final DownloadTask d = new DownloadTask(new LoggingUtil(Logger.getLogger("testLogger")),
                containerConfigurationMock);
        d.execute(ctx);

        // Assertions
        final File downloadedFile = ctx.getArchiveFile();
        assertTrue(Pattern.matches(artifactId + DOWNLOADED_FILENAME_SUFFIX_REGEXP, downloadedFile.getName()));
        assertTrue(downloadedFile.exists());

    }

    @Test
    public void testDownloadMavenUnexistingEntityPackage() throws Exception {

        final String urlTest = "mvn:org.ow2.petals/unexisting-comp/1/zip";
        final Context ctx = createContext(urlTest);
        assertEquals(urlTest, ctx.getArchiveUrl().toString());

        final ContainerConfiguration containerConfigurationMock = new ContainerConfiguration();
        containerConfigurationMock.setWorkDirectoryPath(this.testFolder.newFolder("work").getAbsolutePath());

        try {

            final DownloadTask d = new DownloadTask(new LoggingUtil(Logger.getLogger("testLogger")),
                    containerConfigurationMock);
            d.execute(ctx);

        } catch (final PetalsException pE) {
            assertTrue(pE.getMessage().contains(urlTest));
            assertTrue(pE.getMessage().contains("Unable to retrieve the remote resource"));
            assertTrue(pE.getMessage().contains("Maven repositories"));
        }
    }

    private Context createContext(final String test) throws URISyntaxException, MalformedURLException {
        final Context ctx = new Context();
        ctx.setEntityName(componentName);

        final URI u = new URI(test);
        ctx.setArchiveUrl(u.toURL());

        return ctx;
    }

}
